/* Egoboo - game.c
*/

/*
This file is part of Egoboo.

Egoboo is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Egoboo is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with Egoboo.  If not, see <http://www.gnu.org/licenses/>.
*/


//#define MENU_DEMO   // Uncomment this to build just the menu demo
#define DECLARE_GLOBALS
#include "egoboo.h"
#include "Ui.h"
#include "Font.h"
#include "Clock.h"
#include "Log.h"
#include "Client.h"
#include "Server.h"
#include "System.h"
#include <SDL_endian.h>

#include <stdlib.h>
#include <stdio.h>
#include <time.h>


#define INITGUID
#define NAME "Boo"
#define TITLE "Boo"

#define RELEASE(x) if (x) {x->Release(); x=NULL;}

#define PROFILE_DECLARE(XX) ClockState * clkstate_##XX = NULL; double clkcount_##XX = 0.0; double clktime_##XX = 0.0;
#define PROFILE_INIT(XX)    { clkstate_##XX  = clock_create_state(); clock_init(clkstate_##XX); }
#define PROFILE_FREE(XX)    { clock_free_state(clkstate_##XX); clkstate_##XX = NULL; }
#define PROFILE_QUERY(XX)   ( (double)clktime_##XX / (double)clkcount_##XX )

#define PROFILE_BEGIN(XX) clock_frameStep(clkstate_##XX); {
#define PROFILE_END(XX)   clock_frameStep(clkstate_##XX);  clkcount_##XX = clkcount_##XX*0.9 + 0.1*1.0; clktime_##XX = clktime_##XX*0.9 + 0.1*clock_getFrameDuration(clkstate_##XX); }

ClockState * g_clk_state = NULL;

PROFILE_DECLARE(resize_characters);
PROFILE_DECLARE(keep_weapons_with_holders);
PROFILE_DECLARE(let_ai_think);
PROFILE_DECLARE(do_weather_spawn);
PROFILE_DECLARE(do_enchant_spawn);
PROFILE_DECLARE(cl_unbufferLatches);
PROFILE_DECLARE(sv_unbufferLatches);
PROFILE_DECLARE(check_respawn);
PROFILE_DECLARE(move_characters);
PROFILE_DECLARE(move_particles);
PROFILE_DECLARE(make_character_matrices);
PROFILE_DECLARE(attach_particles);
PROFILE_DECLARE(make_onwhichfan);
PROFILE_DECLARE(bump_characters);
PROFILE_DECLARE(stat_return);
PROFILE_DECLARE(pit_kill);
PROFILE_DECLARE(flash_select);
PROFILE_DECLARE(animate_tiles);
PROFILE_DECLARE(move_water);
PROFILE_DECLARE(figure_out_what_to_draw);
PROFILE_DECLARE(draw_main);

//---------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------
int what_action(char cTmp)
{
  // ZZ> This function changes a letter into an action code
  int action;
  action = ACTIONDA;
  if (cTmp == 'U' || cTmp == 'u')  action = ACTIONUA;
  if (cTmp == 'T' || cTmp == 't')  action = ACTIONTA;
  if (cTmp == 'S' || cTmp == 's')  action = ACTIONSA;
  if (cTmp == 'C' || cTmp == 'c')  action = ACTIONCA;
  if (cTmp == 'B' || cTmp == 'b')  action = ACTIONBA;
  if (cTmp == 'L' || cTmp == 'l')  action = ACTIONLA;
  if (cTmp == 'X' || cTmp == 'x')  action = ACTIONXA;
  if (cTmp == 'F' || cTmp == 'f')  action = ACTIONFA;
  if (cTmp == 'P' || cTmp == 'p')  action = ACTIONPA;
  if (cTmp == 'Z' || cTmp == 'z')  action = ACTIONZA;
  return action;
}

//------------------------------------------------------------------------------
bool_t memory_cleanUp()
{
  //ZF> This function releases all loaded things in memory and cleans up everything properly
  bool_t result = bfalse;

  close_session();					  //Turn off networking
  if (moduleActive) release_module(); //Remove memory loaded by a module
  if (mixeron) Mix_CloseAudio();      //Close audio systems
  ui_shutdown();					  //Shut down support systems
  net_shutDown();
  clock_shutdown(g_clk_state);
  sys_shutdown();
  SDL_Quit();						  //Quit SDL

  result = btrue;
  return result;  //Todo: Need to do a check here if the functions above worked
}

//------------------------------------------------------------------------------
//Random Things-----------------------------------------------------------------
//------------------------------------------------------------------------------
void make_newloadname(char *modname, char *appendname, char *newloadname)
{
  // ZZ> This function takes some names and puts 'em together
  int cnt, tnc;
  char ctmp;

  cnt = 0;
  ctmp = modname[cnt];
  while (ctmp != 0)
  {
    newloadname[cnt] = ctmp;
    cnt++;
    ctmp = modname[cnt];
  }
  tnc = 0;
  ctmp = appendname[tnc];
  while (ctmp != 0)
  {
    newloadname[cnt] = ctmp;
    cnt++;
    tnc++;
    ctmp = appendname[tnc];
  }
  newloadname[cnt] = 0;
}

//--------------------------------------------------------------------------------------------
void load_global_waves(char *modname)
{
  // ZZ> This function loads the global waves
  STRING tmploadname, newloadname;
  int cnt;

  if (CData.soundvalid)
  {
    // load in the sounds local to this module
    snprintf(tmploadname, sizeof(tmploadname), "%s%s/", modname, CData.gamedat_dir);
    for (cnt = 0; cnt < MAXWAVE; cnt++)
    {
      snprintf(newloadname, sizeof(newloadname), "%ssound%d.wav", tmploadname, cnt);
      globalwave[cnt] = Mix_LoadWAV(newloadname);
    };

    //These sounds are always standard, but DO NOT override sounds that were loaded local to this module
    if (NULL == globalwave[0])
    {
      snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s/%s", CData.basicdat_dir, CData.globalparticles_dir, CData.coinget_sound);
      globalwave[0] = Mix_LoadWAV(CStringTmp1);
    };

    if (NULL == globalwave[1])
    {
      snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s/%s", CData.basicdat_dir, CData.globalparticles_dir, CData.defend_sound);
      globalwave[1] = Mix_LoadWAV(CStringTmp1);
    }

    if (NULL == globalwave[4])
    {
      snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s/%s", CData.basicdat_dir, CData.globalparticles_dir, CData.coinfall_sound);
      globalwave[4] = Mix_LoadWAV(CStringTmp1);
    };
  }

  /*  The Global Sounds
  * 0 - Pick up coin
  * 1 - Defend clank
  * 2 - Weather Effect
  * 3 - Hit Water tile (Splash)
  * 4 - Coin falls on ground

  //These new values todo should determine sound and particle effects (examples below)
  Weather Type: DROPS, RAIN, SNOW, LAVABUBBLE (Which weather effect to spawn)
  Water Type: LAVA, WATER, DARK (To determine sound and particle effects)

  //We shold also add standard particles that can be used everywhere (Located and
  //loaded in globalparticles folder) such as these below.
  Particle Effect: REDBLOOD, SMOKE, HEALCLOUD
  */
}


//---------------------------------------------------------------------------------------------
void export_one_character(Uint16 character, Uint16 owner, int number)
{
  // ZZ> This function exports a character
  int tnc, profile;
  char letter;
  char fromdir[128];
  char todir[128];
  char fromfile[128];
  char tofile[128];
  char todirname[16];
  char todirfullname[64];

  // Don't export enchants
  disenchant_character(character);

  profile = chrmodel[character];
  if ((capcancarrytonextmodule[profile] || !capisitem[profile]) && exportvalid)
  {
    // TWINK_BO.OBJ
    snprintf(todirname, sizeof(todirname), "badname.obj");//"BADNAME.OBJ");
    tnc = 0;
    letter = chrname[owner][tnc];
    while (tnc < 8 && letter != 0)
    {
      letter = chrname[owner][tnc];
      if (letter >= 'A' && letter <= 'Z')  letter -= 'A' - 'a';
      if (letter != 0)
      {
        if (letter < 'a' || letter > 'z')  letter = '_';
        todirname[tnc] = letter;
        tnc++;
      }
    }
    todirname[tnc] = '.'; tnc++;
    todirname[tnc] = 'o'; tnc++;
    todirname[tnc] = 'b'; tnc++;
    todirname[tnc] = 'j'; tnc++;
    todirname[tnc] = 0;



    // Is it a character or an item?
    if (owner != character)
    {
      // Item is a subdirectory of the owner directory...
      snprintf(todirfullname, sizeof(todirfullname), "%s/%d.obj", todirname, number);
    }
    else
    {
      // Character directory
      strncpy(todirfullname, todirname, sizeof(todirfullname));
    }


    // players/twink.obj or players/twink.obj/sword.obj
    snprintf(todir, sizeof(todir), "players/%s", todirfullname);
    // modules/advent.mod/objects/advent.obj
    strncpy(fromdir, madname[profile], sizeof(fromdir));


    // Delete all the old items
    if (owner == character)
    {
      tnc = 0;
      while (tnc < 8)
      {
        snprintf(tofile, sizeof(tofile), "%s/%d.obj", todir, tnc); /*.OBJ*/
        fs_removeDirectoryAndContents(tofile);
        tnc++;
      }
    }


    // Make the directory
    fs_createDirectory(todir);


    // Build the DATA.TXT file
    snprintf(tofile, sizeof(tofile), "%s/%s", todir, CData.data_file);  /*DATA.TXT*/
    export_one_character_profile(tofile, character);


    // Build the SKIN.TXT file
    snprintf(tofile, sizeof(tofile), "%s/%s", todir, CData.skin_file);  /*SKIN.TXT*/
    export_one_character_skin(tofile, character);


    // Build the NAMING.TXT file
    snprintf(tofile, sizeof(tofile), "%s/%s", todir, CData.naming_file);  /*NAMING.TXT*/
    export_one_character_name(tofile, character);


    // Copy all of the misc. data files
    snprintf(fromfile, sizeof(fromfile), "%s/%s", fromdir, CData.message_file); /*MESSAGE.TXT*/
    snprintf(tofile, sizeof(tofile), "%s/%s", todir, CData.message_file); /*MESSAGE.TXT*/
    fs_copyFile(fromfile, tofile);

    snprintf(fromfile, sizeof(fromfile), "%s/tris.md2", fromdir);  /*TRIS.MD2*/
    snprintf(tofile,   sizeof(tofile), "%s/tris.md2", todir);  /*TRIS.MD2*/
    fs_copyFile(fromfile, tofile);

    snprintf(fromfile, sizeof(fromfile), "%s/%s", fromdir, CData.copy_file);  /*COPY.TXT*/
    snprintf(tofile,   sizeof(tofile), "%s/%s", todir, CData.copy_file);  /*COPY.TXT*/
    fs_copyFile(fromfile, tofile);

    snprintf(fromfile, sizeof(fromfile), "%s/%s", fromdir, CData.script_file);
    snprintf(tofile,   sizeof(tofile), "%s/%s", todir, CData.script_file);
    fs_copyFile(fromfile, tofile);

    snprintf(fromfile, sizeof(fromfile), "%s/%s", fromdir, CData.enchant_file);
    snprintf(tofile,   sizeof(tofile), "%s/%s", todir, CData.enchant_file);
    fs_copyFile(fromfile, tofile);

    snprintf(fromfile, sizeof(fromfile), "%s/%s", fromdir, CData.credits_file);
    snprintf(tofile,   sizeof(tofile), "%s/%s", todir, CData.credits_file);
    fs_copyFile(fromfile, tofile);


    // Copy all of the particle files
    tnc = 0;
    while (tnc < PRTPIP_PEROBJECT_COUNT)
    {
      snprintf(fromfile, sizeof(fromfile), "%s/part%d.txt", fromdir, tnc);
      snprintf(tofile,   sizeof(tofile), "%s/part%d.txt", todir,   tnc);
      fs_copyFile(fromfile, tofile);
      tnc++;
    }


    // Copy all of the sound files
    tnc = 0;
    while (tnc < MAXWAVE)
    {
      snprintf(fromfile, sizeof(fromfile), "%s/sound%d.wav", fromdir, tnc);
      snprintf(tofile,   sizeof(tofile), "%s/sound%d.wav", todir,   tnc);
      fs_copyFile(fromfile, tofile);
      tnc++;
    }


    // Copy all of the image files
    tnc = 0;
    while (tnc < 4)
    {
      snprintf(fromfile, sizeof(fromfile), "%s/tris%d.bmp", fromdir, tnc);
      snprintf(tofile,   sizeof(tofile), "%s/tris%d.bmp", todir,   tnc);
      fs_copyFile(fromfile, tofile);

      snprintf(fromfile, sizeof(fromfile), "%s/icon%d.bmp", fromdir, tnc);
      snprintf(tofile,   sizeof(tofile), "%s/icon%d.bmp", todir,   tnc);
      fs_copyFile(fromfile, tofile);
      tnc++;
    }
  }
}

//---------------------------------------------------------------------------------------------
void export_all_local_players(void)
{
  // ZZ> This function saves all the local players in the
  //     PLAYERS directory
  int cnt, character, item, number;

  // Check each player
  if (exportvalid)
  {
    cnt = 0;
    while (cnt < MAXPLAYER)
    {
      if (plavalid[cnt] && pladevice[cnt])
      {
        // Is it alive?
        character = plaindex[cnt];
        if (chron[character] && chralive[character])
        {
          // Export the character
          export_one_character(character, character, 0);


          // Export the left hand item
          item = chrholdingwhich[character][0];
          if (item != MAXCHR && chrisitem[item])  export_one_character(item, character, 0);

          // Export the right hand item
          item = chrholdingwhich[character][1];
          if (item != MAXCHR && chrisitem[item])  export_one_character(item, character, 1);

          // Export the inventory
          number = 2;
          item = chrnextinpack[character];
          while (item != MAXCHR)
          {
            if (chrisitem[item]) export_one_character(item, character, number);
            item = chrnextinpack[item];
            number++;
          }
        }
      }
      cnt++;
    }
  }
}

//---------------------------------------------------------------------------------------------
void quit_module(void)
{
  // ZZ> This function forces a return to the menu
  moduleActive = bfalse;
  hostactive = bfalse;
  export_all_local_players();
  gamepaused = bfalse;
  if (CData.soundvalid) Mix_FadeOutChannel(-1, 500);        //Stop all sounds that are playing
}

//--------------------------------------------------------------------------------------------
void quit_game(void)
{
  log_info("Exiting Egoboo %s the good way...\n", VERSION);
  // ZZ> This function exits the game entirely
  if (gameActive)
  {
    gameActive = bfalse;
  }
  if (moduleActive)
  {
    quit_module();
  }
  if (floatmemory != NULL)
  {
    free(floatmemory);
    floatmemory = NULL;
  }
}

//--------------------------------------------------------------------------------------------
void goto_colon(FILE* fileread)
{
  // ZZ> This function moves a file read pointer to the next colon
  //    char cTmp;
  Uint32 ch = fgetc(fileread);

  //    fscanf(fileread, "%c", &cTmp);
  while (ch != ':')
  {
    if (ch == EOF)
    {
      // not enough colons in file!
      log_error("There are not enough colons in file! (%s)\n", globalname);
    }

    ch = fgetc(fileread);
  }
}

//--------------------------------------------------------------------------------------------
Uint8 goto_colon_yesno(FILE* fileread)
{
  // ZZ> This function moves a file read pointer to the next colon, or it returns
  //     bfalse if there are no more
  char cTmp;

  do
  {
    if (fscanf(fileread, "%c", &cTmp) == EOF)
    {
      return bfalse;
    }
  }
  while (cTmp != ':');
  return btrue;
}

//--------------------------------------------------------------------------------------------
char get_first_letter(FILE* fileread)
{
  // ZZ> This function returns the next non-whitespace character
  char cTmp;
  fscanf(fileread, "%c", &cTmp);
  while (isspace(cTmp))
  {
    fscanf(fileread, "%c", &cTmp);
  }
  return cTmp;
}

//--------------------------------------------------------------------------------------------
//Tag Reading---------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------
void reset_tags()
{
  // ZZ> This function resets the tags
  numscantag = 0;
}

//--------------------------------------------------------------------------------------------
int read_tag(FILE *fileread)
{
  // ZZ> This function finds the next tag, returning btrue if it found one
  if (goto_colon_yesno(fileread))
  {
    if (numscantag < MAXTAG)
    {
      fscanf(fileread, "%s%d", tagname[numscantag], &tagvalue[numscantag]);
      numscantag++;
      return btrue;
    }
  }
  return bfalse;
}

//--------------------------------------------------------------------------------------------
void read_all_tags(char *szFilename)
{
  // ZZ> This function reads the scancode.txt file
  FILE* fileread;


  reset_tags();
  fileread = fs_fileOpen(PRI_NONE, NULL, szFilename, "r");
  if (NULL!=fileread)
  {
    while (read_tag(fileread));
    fs_fileClose(fileread);
  }
}

//--------------------------------------------------------------------------------------------
int tag_value(char *string)
{
  // ZZ> This function matches the string with its tag, and returns the value...
  //     It will return 255 if there are no matches.
  int cnt;

  cnt = 0;
  while (cnt < numscantag)
  {
    if (strcmp(string, tagname[cnt]) == 0)
    {
      // They match
      return tagvalue[cnt];
    }
    cnt++;
  }
  // No matches
  return 255;
}

//--------------------------------------------------------------------------------------------
void read_controls(char *szFilename)
{
  // ZZ> This function reads the controls.txt file
  FILE* fileread;
  char currenttag[TAGSIZE];
  int cnt;


  fileread = fs_fileOpen(PRI_NONE, NULL, szFilename, "r");
  if (NULL != fileread)
  {
    cnt = 0;
    while (goto_colon_yesno(fileread) && cnt < MAXCONTROL)
    {
      fscanf(fileread, "%s", currenttag);
      controlvalue[cnt] = tag_value(currenttag);
      //printf("CTRL: %i, %s\n", controlvalue[cnt], currenttag);
      controliskey[cnt] = (currenttag[0] == 'K');
      cnt++;
    }
    fs_fileClose(fileread);
  }
  else log_warning("Could not load input settings (%s)\n", szFilename);
}

//--------------------------------------------------------------------------------------------
bool_t control_key_is_pressed(CONTROL control)
{
  // ZZ> This function returns btrue if the given control is pressed...
  if (netmessagemode)  return bfalse;

  if (sdlkeybuffer)
    return SDLKEYDOWN(controlvalue[control]);
  else
    return bfalse;
}

//--------------------------------------------------------------------------------------------
bool_t control_mouse_is_pressed(CONTROL control)
{
  // ZZ> This function returns btrue if the given control is pressed...
  if (controliskey[control])
  {
    return control_key_is_pressed(control);
  }
  else
  {
    return (msb == controlvalue[control]);
  }

  return bfalse;
}

//--------------------------------------------------------------------------------------------
bool_t control_joya_is_pressed(CONTROL control)
{
  // ZZ> This function returns btrue if the given control is pressed...
  if (controliskey[control])
  {
    return control_key_is_pressed(control);
  }
  else
  {
    return (jab == controlvalue[control]);
  }

  return bfalse;
}

//--------------------------------------------------------------------------------------------
bool_t control_joyb_is_pressed(CONTROL control)
{
  // ZZ> This function returns btrue if the given control is pressed...

  if (controliskey[control])
  {
    return control_key_is_pressed(control);
  }
  else
  {
    return (jbb == controlvalue[control]);
  }

  return bfalse;
}

//--------------------------------------------------------------------------------------------
char * undo_idsz(IDSZ idsz)
{
  // ZZ> This function takes an integer and makes an text IDSZ out of it.
  //     It will set valueidsz to "NONE" if the idsz is 0
  static char value_string[5] = {"NONE"};

  if (idsz == IDSZNONE)
  {
    snprintf(value_string, sizeof(value_string), "NONE");
  }
  else
  {
    value_string[0] = ((idsz >> 15) & 31) + 'A';
    value_string[1] = ((idsz >> 10) & 31) + 'A';
    value_string[2] = ((idsz >> 5) & 31) + 'A';
    value_string[3] = ((idsz) & 31) + 'A';
    value_string[4] = 0;
  }

  return value_string;
}

//--------------------------------------------------------------------------------------------
IDSZ get_idsz(FILE* fileread)
{
  // ZZ> This function reads and returns an IDSZ tag, or IDSZNONE if there wasn't one
  char sTemp[4];
  IDSZ idsz = IDSZNONE;

  char cTmp = get_first_letter(fileread);
  if (cTmp == '[')
  {
    fscanf(fileread, "%c", &sTemp[0]);
    fscanf(fileread, "%c", &sTemp[1]);
    fscanf(fileread, "%c", &sTemp[2]);
    fscanf(fileread, "%c", &sTemp[3]);
    idsz = MAKE_IDSZ(sTemp);
  }

  return idsz;
}

//--------------------------------------------------------------------------------------------
int get_free_message(void)
{
  // This function finds the best message to use
  int cnt, tnc = msgstart, found = msgstart, mintime = 0;

  for(cnt=0; cnt<CData.maxmessage; cnt++)
  {
    if(0 == msgtime[tnc])
    {
      found = tnc;
      break;
    }

    if(msgtime[tnc] > 0 && (mintime==0 || mintime>msgtime[tnc]))
    {
      found = tnc;
      mintime = msgtime[tnc];
    }

    tnc = (tnc + 1) % CData.maxmessage;
  };

  msgstart = (found + 1) % CData.maxmessage;

  return found;
}

//--------------------------------------------------------------------------------------------
void display_message(int message, Uint16 character)
{
  // ZZ> This function sticks a message in the display queue and sets its timer
  int slot, read, write, cnt;
  char *eread;
  STRING szTmp;
  char cTmp, lTmp;

  Uint16 target = chraitarget[character];
  Uint16 owner = chraiowner[character];
  if (message < msgtotal)
  {
    slot = get_free_message();
    msgtime[slot] = MESSAGETIME;
    // Copy the message
    read = msgindex[message];
    cnt = 0;
    write = 0;
    cTmp = msgtext[read];  read++;
    while (cTmp != 0)
    {
      if (cTmp == '%')
      {
        // Escape sequence
        eread = szTmp;
        szTmp[0] = 0;
        cTmp = msgtext[read];  read++;
        if (cTmp == 'n') // Name
        {
          if (chrnameknown[character])
            strncpy(szTmp, chrname[character], sizeof(STRING));
          else
          {
            lTmp = capclassname[chrmodel[character]][0];
            if (lTmp == 'A' || lTmp == 'E' || lTmp == 'I' || lTmp == 'O' || lTmp == 'U')
              snprintf(szTmp, sizeof(szTmp), "an %s", capclassname[chrmodel[character]]);
            else
              snprintf(szTmp, sizeof(szTmp), "a %s", capclassname[chrmodel[character]]);
          }
          if (cnt == 0 && szTmp[0] == 'a')  szTmp[0] = 'A';
        }
        if (cTmp == 'c') // Class name
        {
          eread = capclassname[chrmodel[character]];
        }
        if (cTmp == 't') // Target name
        {
          if (chrnameknown[target])
            strncpy(szTmp, chrname[target], sizeof(STRING));
          else
          {
            lTmp = capclassname[chrmodel[target]][0];
            if (lTmp == 'A' || lTmp == 'E' || lTmp == 'I' || lTmp == 'O' || lTmp == 'U')
              snprintf(szTmp, sizeof(szTmp), "an %s", capclassname[chrmodel[target]]);
            else
              snprintf(szTmp, sizeof(szTmp), "a %s", capclassname[chrmodel[target]]);
          }
          if (cnt == 0 && szTmp[0] == 'a')  szTmp[0] = 'A';
        }
        if (cTmp == 'o') // Owner name
        {
          if (chrnameknown[owner])
            strncpy(szTmp, chrname[owner], sizeof(STRING));
          else
          {
            lTmp = capclassname[chrmodel[owner]][0];
            if (lTmp == 'A' || lTmp == 'E' || lTmp == 'I' || lTmp == 'O' || lTmp == 'U')
              snprintf(szTmp, sizeof(szTmp), "an %s", capclassname[chrmodel[owner]]);
            else
              snprintf(szTmp, sizeof(szTmp), "a %s", capclassname[chrmodel[owner]]);
          }
          if (cnt == 0 && szTmp[0] == 'a')  szTmp[0] = 'A';
        }
        if (cTmp == 's') // Target class name
        {
          eread = capclassname[chrmodel[target]];
        }
        if (cTmp >= '0' && cTmp <= '0' + (MAXSKIN - 1))  // Target's skin name
        {
          eread = capskinname[chrmodel[target]][cTmp-'0'];
        }
        if (cTmp == 'd') // tmpdistance value
        {
          snprintf(szTmp, sizeof(szTmp), "%d", valuetmpdistance);
        }
        if (cTmp == 'x') // tmpx value
        {
          snprintf(szTmp, sizeof(szTmp), "%d", valuetmpx);
        }
        if (cTmp == 'y') // tmpy value
        {
          snprintf(szTmp, sizeof(szTmp), "%d", valuetmpy);
        }
        if (cTmp == 'D') // tmpdistance value
        {
          snprintf(szTmp, sizeof(szTmp), "%2d", valuetmpdistance);
        }
        if (cTmp == 'X') // tmpx value
        {
          snprintf(szTmp, sizeof(szTmp), "%2d", valuetmpx);
        }
        if (cTmp == 'Y') // tmpy value
        {
          snprintf(szTmp, sizeof(szTmp), "%2d", valuetmpy);
        }
        if (cTmp == 'a') // Character's ammo
        {
          if (chrammoknown[character])
            snprintf(szTmp, sizeof(szTmp), "%d", chrammo[character]);
          else
            snprintf(szTmp, sizeof(szTmp), "?");
        }
        if (cTmp == 'k') // Kurse state
        {
          if (chriskursed[character])
            snprintf(szTmp, sizeof(szTmp), "kursed");
          else
            snprintf(szTmp, sizeof(szTmp), "unkursed");
        }
        if (cTmp == 'p') // Character's possessive
        {
          if (chrgender[character] == GENFEMALE)
          {
            snprintf(szTmp, sizeof(szTmp), "her");
          }
          else
          {
            if (chrgender[character] == GENMALE)
            {
              snprintf(szTmp, sizeof(szTmp), "his");
            }
            else
            {
              snprintf(szTmp, sizeof(szTmp), "its");
            }
          }
        }
        if (cTmp == 'm') // Character's gender
        {
          if (chrgender[character] == GENFEMALE)
          {
            snprintf(szTmp, sizeof(szTmp), "female ");
          }
          else
          {
            if (chrgender[character] == GENMALE)
            {
              snprintf(szTmp, sizeof(szTmp), "male ");
            }
            else
            {
              snprintf(szTmp, sizeof(szTmp), " ");
            }
          }
        }
        if (cTmp == 'g') // Target's possessive
        {
          if (chrgender[target] == GENFEMALE)
          {
            snprintf(szTmp, sizeof(szTmp), "her");
          }
          else
          {
            if (chrgender[target] == GENMALE)
            {
              snprintf(szTmp, sizeof(szTmp), "his");
            }
            else
            {
              snprintf(szTmp, sizeof(szTmp), "its");
            }
          }
        }
        cTmp = *eread;  eread++;
        while (cTmp != 0 && write < MESSAGESIZE - 1)
        {
          msgtextdisplay[slot][write] = cTmp;
          cTmp = *eread;  eread++;
          write++;
        }
      }
      else
      {
        // Copy the letter
        if (write < MESSAGESIZE - 1)
        {
          msgtextdisplay[slot][write] = cTmp;
          write++;
        }
      }
      cTmp = msgtext[read];  read++;
      cnt++;
    }
    msgtextdisplay[slot][write] = 0;
  }
}

//--------------------------------------------------------------------------------------------
void remove_enchant(Uint16 enchantindex)
{
  // ZZ> This function removes a specific enchantment and adds it to the unused list
  Uint16 character, overlay;
  Uint16 lastenchant, currentenchant;
  int add, cnt;


  if (enchantindex < MAXENCHANT)
  {
    if (encon[enchantindex])
    {
      // Unsparkle the spellbook
      character = encspawner[enchantindex];
      if (character < MAXCHR)
      {
        chrsparkle[character] = NOSPARKLE;
        // Make the spawner unable to undo the enchantment
        if (chrundoenchant[character] == enchantindex)
        {
          chrundoenchant[character] = MAXENCHANT;
        }
      }


      // Play the end sound
      character = enctarget[enchantindex];
      if (NULL != evewaveindex[enceve[enchantindex]])
        play_sound(chroldx[character], chroldy[character], evewaveindex[enceve[enchantindex]]);

	  

      // Unset enchant values, doing morph last (opposite order to spawn_enchant)
      unset_enchant_value(enchantindex, SETCHANNEL);
      for(cnt=SETCOSTFOREACHMISSILE; cnt>=SETCOSTFOREACHMISSILE; cnt--)
      {
        unset_enchant_value(enchantindex, cnt);
      }
      unset_enchant_value(enchantindex, SETMORPH);


      // Remove all of the cumulative values
      add = 0;
      while (add < EVE_ADD_COUNT)
      {
        remove_enchant_value(enchantindex, add);
        add++;
      }


      // Unlink it
      if (chrfirstenchant[character] == enchantindex)
      {
        // It was the first in the list
        chrfirstenchant[character] = encnextenchant[enchantindex];
      }
      else
      {
        // Search until we find it
        lastenchant    = MAXENCHANT;
        currentenchant = chrfirstenchant[character];
        while (currentenchant != enchantindex)
        {
          lastenchant = currentenchant;
          currentenchant = encnextenchant[currentenchant];
        }

        // Relink the last enchantment
        encnextenchant[lastenchant] = encnextenchant[enchantindex];
      }



      // See if we spit out an end message
      if (eveendmessage[enceve[enchantindex]] >= 0)
      {
        display_message(madmsgstart[enceve[enchantindex]] + eveendmessage[enceve[enchantindex]], enctarget[enchantindex]);
      }
      // Check to see if we spawn a poof
      if (evepoofonend[enceve[enchantindex]])
      {
        spawn_poof(enctarget[enchantindex], enceve[enchantindex]);
      }
      // Check to see if the character dies
      if (evekillonend[enceve[enchantindex]])
      {
        if (chrinvictus[character])  teammorale[chrbaseteam[character]]++;
        chrinvictus[character] = bfalse;
        kill_character(character, MAXCHR);
      }
      // Kill overlay too...
      overlay = encoverlay[enchantindex];
      if (overlay < MAXCHR)
      {
        if (chrinvictus[overlay])  teammorale[chrbaseteam[overlay]]++;
        chrinvictus[overlay] = bfalse;
        kill_character(overlay, MAXCHR);
      }





      // Now get rid of it
      encon[enchantindex] = bfalse;
      freeenchant[numfreeenchant] = enchantindex;
      numfreeenchant++;


      // Now fix dem weapons
      reset_character_alpha(chrholdingwhich[character][0]);
      reset_character_alpha(chrholdingwhich[character][1]);
    }
  }
}

//--------------------------------------------------------------------------------------------
Uint16 enchant_value_filled(Uint16 enchantindex, Uint8 valueindex)
{
  // ZZ> This function returns MAXENCHANT if the enchantment's target has no conflicting
  //     set values in its other enchantments.  Otherwise it returns the enchantindex
  //     of the conflicting enchantment
  Uint16 character, currenchant;

  character = enctarget[enchantindex];
  currenchant = chrfirstenchant[character];
  while (currenchant != MAXENCHANT)
  {
    if (encsetyesno[currenchant][valueindex])
    {
      return currenchant;
    }
    currenchant = encnextenchant[currenchant];
  }
  return MAXENCHANT;
}

//--------------------------------------------------------------------------------------------
void set_enchant_value(Uint16 enchantindex, Uint8 valueindex,
                       Uint16 enchanttype)
{
  // ZZ> This function sets and saves one of the character's stats
  Uint16 conflict, character;


  encsetyesno[enchantindex][valueindex] = bfalse;
  if (evesetyesno[enchanttype][valueindex])
  {
    conflict = enchant_value_filled(enchantindex, valueindex);
    if (conflict == MAXENCHANT || eveoverride[enchanttype])
    {
      // Check for multiple enchantments
      if (conflict < MAXENCHANT)
      {
        // Multiple enchantments aren't allowed for sets
        if (everemoveoverridden[enchanttype])
        {
          // Kill the old enchantment
          remove_enchant(conflict);
        }
        else
        {
          // Just unset the old enchantment's value
          unset_enchant_value(conflict, valueindex);
        }
      }
      // Set the value, and save the character's real stat
      character = enctarget[enchantindex];
      encsetyesno[enchantindex][valueindex] = btrue;
      switch (valueindex)
      {
      case SETDAMAGETYPE:
        encsetsave[enchantindex][valueindex] = chrdamagetargettype[character];
        chrdamagetargettype[character] = evesetvalue[enchanttype][valueindex];
        break;

      case SETNUMBEROFJUMPS:
        encsetsave[enchantindex][valueindex] = chrjumpnumberreset[character];
        chrjumpnumberreset[character] = evesetvalue[enchanttype][valueindex];
        break;

      case SETLIFEBARCOLOR:
        encsetsave[enchantindex][valueindex] = chrlifecolor[character];
        chrlifecolor[character] = evesetvalue[enchanttype][valueindex];
        break;

      case SETMANABARCOLOR:
        encsetsave[enchantindex][valueindex] = chrmanacolor[character];
        chrmanacolor[character] = evesetvalue[enchanttype][valueindex];
        break;

      case SETSLASHMODIFIER:
        encsetsave[enchantindex][valueindex] = chrdamagemodifier[character][DAMAGESLASH];
        chrdamagemodifier[character][DAMAGESLASH] = evesetvalue[enchanttype][valueindex];
        break;

      case SETCRUSHMODIFIER:
        encsetsave[enchantindex][valueindex] = chrdamagemodifier[character][DAMAGECRUSH];
        chrdamagemodifier[character][DAMAGECRUSH] = evesetvalue[enchanttype][valueindex];
        break;

      case SETPOKEMODIFIER:
        encsetsave[enchantindex][valueindex] = chrdamagemodifier[character][DAMAGEPOKE];
        chrdamagemodifier[character][DAMAGEPOKE] = evesetvalue[enchanttype][valueindex];
        break;

      case SETHOLYMODIFIER:
        encsetsave[enchantindex][valueindex] = chrdamagemodifier[character][DAMAGEHOLY];
        chrdamagemodifier[character][DAMAGEHOLY] = evesetvalue[enchanttype][valueindex];
        break;

      case SETEVILMODIFIER:
        encsetsave[enchantindex][valueindex] = chrdamagemodifier[character][DAMAGEEVIL];
        chrdamagemodifier[character][DAMAGEEVIL] = evesetvalue[enchanttype][valueindex];
        break;

      case SETFIREMODIFIER:
        encsetsave[enchantindex][valueindex] = chrdamagemodifier[character][DAMAGEFIRE];
        chrdamagemodifier[character][DAMAGEFIRE] = evesetvalue[enchanttype][valueindex];
        break;

      case SETICEMODIFIER:
        encsetsave[enchantindex][valueindex] = chrdamagemodifier[character][DAMAGEICE];
        chrdamagemodifier[character][DAMAGEICE] = evesetvalue[enchanttype][valueindex];
        break;

      case SETZAPMODIFIER:
        encsetsave[enchantindex][valueindex] = chrdamagemodifier[character][DAMAGEZAP];
        chrdamagemodifier[character][DAMAGEZAP] = evesetvalue[enchanttype][valueindex];
        break;

      case SETFLASHINGAND:
        encsetsave[enchantindex][valueindex] = chrflashand[character];
        chrflashand[character] = evesetvalue[enchanttype][valueindex];
        break;

      case SETLIGHTBLEND:
        encsetsave[enchantindex][valueindex] = chrlight[character];
        chrlight[character] = evesetvalue[enchanttype][valueindex];
        break;

      case SETALPHABLEND:
        encsetsave[enchantindex][valueindex] = chralpha[character];
        chralpha[character] = evesetvalue[enchanttype][valueindex];
        break;

      case SETSHEEN:
        encsetsave[enchantindex][valueindex] = chrsheen[character];
        chrsheen[character] = evesetvalue[enchanttype][valueindex];
        break;

      case SETFLYTOHEIGHT:
        encsetsave[enchantindex][valueindex] = chrflyheight[character];
        if (chrflyheight[character] == 0 && chrzpos[character] > -2)
        {
          chrflyheight[character] = evesetvalue[enchanttype][valueindex];
        }
        break;

      case SETWALKONWATER:
        encsetsave[enchantindex][valueindex] = chrwaterwalk[character];
        if (!chrwaterwalk[character])
        {
          chrwaterwalk[character] = evesetvalue[enchanttype][valueindex];
        }
        break;

      case SETCANSEEINVISIBLE:
        encsetsave[enchantindex][valueindex] = chrcanseeinvisible[character];
        chrcanseeinvisible[character] = evesetvalue[enchanttype][valueindex];
        break;

      case SETMISSILETREATMENT:
        encsetsave[enchantindex][valueindex] = chrmissiletreatment[character];
        chrmissiletreatment[character] = evesetvalue[enchanttype][valueindex];
        break;

      case SETCOSTFOREACHMISSILE:
        encsetsave[enchantindex][valueindex] = chrmissilecost[character];
        chrmissilecost[character] = evesetvalue[enchanttype][valueindex];
        chrmissilehandler[character] = encowner[enchantindex];
        break;

      case SETMORPH:
        encsetsave[enchantindex][valueindex] = (chrtexture[character] - madskinstart[chrmodel[character]]) % MAXSKIN;
        // Special handler for morph
        change_character(character, enchanttype, 0, LEAVEALL); // LEAVEFIRST);
        chralert[character] |= ALERTIFCHANGED;
        break;

      case SETCHANNEL:
        encsetsave[enchantindex][valueindex] = chrcanchannel[character];
        chrcanchannel[character] = evesetvalue[enchanttype][valueindex];
        break;

      }
    }
  }
}

//--------------------------------------------------------------------------------------------
void getadd(int MIN, int value, int MAX, int* valuetoadd)
{
  // ZZ> This function figures out what value to add should be in order
  //     to not overflow the MIN and MAX bounds
  int newvalue;

  newvalue = value + (*valuetoadd);
  if (newvalue < MIN)
  {
    // Increase valuetoadd to fit
    *valuetoadd = MIN - value;
    if (*valuetoadd > 0)  *valuetoadd = 0;
    return;
  }


  if (newvalue > MAX)
  {
    // Decrease valuetoadd to fit
    *valuetoadd = MAX - value;
    if (*valuetoadd < 0)  *valuetoadd = 0;
  }
}

//--------------------------------------------------------------------------------------------
void fgetadd(float MIN, float value, float MAX, float* valuetoadd)
{
  // ZZ> This function figures out what value to add should be in order
  //     to not overflow the MIN and MAX bounds
  float newvalue;


  newvalue = value + (*valuetoadd);
  if (newvalue < MIN)
  {
    // Increase valuetoadd to fit
    *valuetoadd = MIN - value;
    if (*valuetoadd > 0)  *valuetoadd = 0;
    return;
  }


  if (newvalue > MAX)
  {
    // Decrease valuetoadd to fit
    *valuetoadd = MAX - value;
    if (*valuetoadd < 0)  *valuetoadd = 0;
  }
}

//--------------------------------------------------------------------------------------------
void add_enchant_value(Uint16 enchantindex, Uint8 valueindex,
                       Uint16 enchanttype)
{
  // ZZ> This function does cumulative modification to character stats
  int valuetoadd, newvalue;
  float fvaluetoadd, fnewvalue;
  Uint16 character;


  character = enctarget[enchantindex];
  valuetoadd = 0;
  switch (valueindex)
  {
  case ADDJUMPPOWER:
    fnewvalue = chrjump[character];
    fvaluetoadd = eveaddvalue[enchanttype][valueindex] / 16.0;
    fgetadd(0, fnewvalue, 30.0, &fvaluetoadd);
    valuetoadd = fvaluetoadd * 16.0; // Get save value
    fvaluetoadd = valuetoadd / 16.0;
    chrjump[character] += fvaluetoadd;

    break;

  case ADDBUMPDAMPEN:
    fnewvalue = chrbumpdampen[character];
    fvaluetoadd = eveaddvalue[enchanttype][valueindex] / 128.0;
    fgetadd(0, fnewvalue, 1.0, &fvaluetoadd);
    valuetoadd = fvaluetoadd * 128.0; // Get save value
    fvaluetoadd = valuetoadd / 128.0;
    chrbumpdampen[character] += fvaluetoadd;
    break;

  case ADDBOUNCINESS:
    fnewvalue = chrdampen[character];
    fvaluetoadd = eveaddvalue[enchanttype][valueindex] / 128.0;
    fgetadd(0, fnewvalue, 0.95, &fvaluetoadd);
    valuetoadd = fvaluetoadd * 128.0; // Get save value
    fvaluetoadd = valuetoadd / 128.0;
    chrdampen[character] += fvaluetoadd;
    break;

  case ADDDAMAGE:
    newvalue = chrdamageboost[character];
    valuetoadd = eveaddvalue[enchanttype][valueindex] << 6;
    getadd(0, newvalue, 4096, &valuetoadd);
    chrdamageboost[character] += valuetoadd;
    break;

  case ADDSIZE:
    fnewvalue = chrsizegoto[character];
    fvaluetoadd = eveaddvalue[enchanttype][valueindex] / 128.0;
    fgetadd(0.5, fnewvalue, 2.0, &fvaluetoadd);
    valuetoadd = fvaluetoadd * 128.0; // Get save value
    fvaluetoadd = valuetoadd / 128.0;
    chrsizegoto[character] += fvaluetoadd;
    chrsizegototime[character] = SIZETIME;
    break;

  case ADDACCEL:
    fnewvalue = chrmaxaccel[character];
    fvaluetoadd = eveaddvalue[enchanttype][valueindex] / 25.0;
    fgetadd(0, fnewvalue, 1.5, &fvaluetoadd);
    valuetoadd = fvaluetoadd * 1000.0; // Get save value
    fvaluetoadd = valuetoadd / 1000.0;
    chrmaxaccel[character] += fvaluetoadd;
    break;

  case ADDRED:
    newvalue = chrredshift[character];
    valuetoadd = eveaddvalue[enchanttype][valueindex];
    getadd(0, newvalue, 6, &valuetoadd);
    chrredshift[character] += valuetoadd;
    break;

  case ADDGRN:
    newvalue = chrgrnshift[character];
    valuetoadd = eveaddvalue[enchanttype][valueindex];
    getadd(0, newvalue, 6, &valuetoadd);
    chrgrnshift[character] += valuetoadd;
    break;

  case ADDBLU:
    newvalue = chrblushift[character];
    valuetoadd = eveaddvalue[enchanttype][valueindex];
    getadd(0, newvalue, 6, &valuetoadd);
    chrblushift[character] += valuetoadd;
    break;

  case ADDDEFENSE:
    newvalue = chrdefense[character];
    valuetoadd = eveaddvalue[enchanttype][valueindex];
    getadd(55, newvalue, 255, &valuetoadd);  // Don't fix again!
    chrdefense[character] += valuetoadd;
    break;

  case ADDMANA:
    newvalue = chrmanamax[character];
    valuetoadd = eveaddvalue[enchanttype][valueindex] << 6;
    getadd(0, newvalue, HIGHSTAT, &valuetoadd);
    chrmanamax[character] += valuetoadd;
    chrmana[character] += valuetoadd;
    if (chrmana[character] < 0)  chrmana[character] = 0;
    break;

  case ADDLIFE:
    newvalue = chrlifemax[character];
    valuetoadd = eveaddvalue[enchanttype][valueindex] << 6;
    getadd(LOWSTAT, newvalue, HIGHSTAT, &valuetoadd);
    chrlifemax[character] += valuetoadd;
    chrlife[character] += valuetoadd;
    if (chrlife[character] < 1)  chrlife[character] = 1;
    break;

  case ADDSTRENGTH:
    newvalue = chrstrength[character];
    valuetoadd = eveaddvalue[enchanttype][valueindex] << 6;
    getadd(0, newvalue, PERFECTSTAT, &valuetoadd);
    chrstrength[character] += valuetoadd;
    break;

  case ADDWISDOM:
    newvalue = chrwisdom[character];
    valuetoadd = eveaddvalue[enchanttype][valueindex] << 6;
    getadd(0, newvalue, PERFECTSTAT, &valuetoadd);
    chrwisdom[character] += valuetoadd;
    break;

  case ADDINTELLIGENCE:
    newvalue = chrintelligence[character];
    valuetoadd = eveaddvalue[enchanttype][valueindex] << 6;
    getadd(0, newvalue, PERFECTSTAT, &valuetoadd);
    chrintelligence[character] += valuetoadd;
    break;

  case ADDDEXTERITY:
    newvalue = chrdexterity[character];
    valuetoadd = eveaddvalue[enchanttype][valueindex] << 6;
    getadd(0, newvalue, PERFECTSTAT, &valuetoadd);
    chrdexterity[character] += valuetoadd;
    break;
  }

  encaddsave[enchantindex][valueindex] = valuetoadd;  // Save the value for undo
}


//--------------------------------------------------------------------------------------------
Uint16 spawn_enchant(Uint16 owner, Uint16 target,
                             Uint16 spawner, Uint16 enchantindex, Uint16 modeloptional)
{
  // ZZ> This function enchants a target, returning the enchantment index or MAXENCHANT
  //     if failed
  Uint16 enchanttype, overlay;
  int add, cnt;


  if (modeloptional < MAXMODEL)
  {
    // The enchantment type is given explicitly
    enchanttype = modeloptional;
  }
  else
  {
    // The enchantment type is given by the spawner
    enchanttype = chrmodel[spawner];
  }


  // Target and owner must both be alive and on and valid
  if (target < MAXCHR)
  {
    if (!chron[target] || !chralive[target])
      return MAXENCHANT;
  }
  else
  {
    // Invalid target
    return MAXENCHANT;
  }
  if (owner < MAXCHR)
  {
    if (!chron[owner] || !chralive[owner])
      return MAXENCHANT;
  }
  else
  {
    // Invalid target
    return MAXENCHANT;
  }


  if (evevalid[enchanttype])
  {
    if (enchantindex == MAXENCHANT)
    {
      // Should it choose an inhand item?
      if (everetarget[enchanttype])
      {
        // Is at least one valid?
        if (chrholdingwhich[target][0] == MAXCHR && chrholdingwhich[target][1] == MAXCHR)
        {
          // No weapons to pick
          return MAXENCHANT;
        }
        // Left, right, or both are valid
        if (chrholdingwhich[target][0] == MAXCHR)
        {
          // Only right hand is valid
          target = chrholdingwhich[target][1];
        }
        else
        {
          // Pick left hand
          target = chrholdingwhich[target][0];
        }
      }


      // Make sure it's valid
      if (evedontdamagetype[enchanttype] != DAMAGENULL)
      {
        if ((chrdamagemodifier[target][evedontdamagetype[enchanttype]]&7) >= 3)  // Invert | Shift = 7
        {
          return MAXENCHANT;
        }
      }
      if (eveonlydamagetype[enchanttype] != DAMAGENULL)
      {
        if (chrdamagetargettype[target] != eveonlydamagetype[enchanttype])
        {
          return MAXENCHANT;
        }
      }


      // Find one to use
      enchantindex = get_free_enchant();
    }
    else
    {
      numfreeenchant--;  // To keep it in order
    }
    if (enchantindex < MAXENCHANT)
    {
      // Make a new one
      encon[enchantindex] = btrue;
      enctarget[enchantindex] = target;
      encowner[enchantindex] = owner;
      encspawner[enchantindex] = spawner;
      if (spawner < MAXCHR)
      {
        chrundoenchant[spawner] = enchantindex;
      }
      enceve[enchantindex] = enchanttype;
      enctime[enchantindex] = evetime[enchanttype];
      encspawntime[enchantindex] = 1;
      encownermana[enchantindex] = eveownermana[enchanttype];
      encownerlife[enchantindex] = eveownerlife[enchanttype];
      enctargetmana[enchantindex] = evetargetmana[enchanttype];
      enctargetlife[enchantindex] = evetargetlife[enchanttype];



      // Add it as first in the list
      encnextenchant[enchantindex] = chrfirstenchant[target];
      chrfirstenchant[target] = enchantindex;


      // Now set all of the specific values, morph first
      set_enchant_value(enchantindex, SETMORPH, enchanttype);
      for(cnt=SETDAMAGETYPE; cnt<=SETCOSTFOREACHMISSILE; cnt++)
      {
        set_enchant_value(enchantindex, cnt, enchanttype);
      }
      set_enchant_value(enchantindex, SETCHANNEL, enchanttype);


      // Now do all of the stat adds
      add = 0;
      while (add < EVE_ADD_COUNT)
      {
        add_enchant_value(enchantindex, add, enchanttype);
        add++;
      }


      // Create an overlay character?
      encoverlay[enchantindex] = MAXCHR;
      if (eveoverlay[enchanttype])
      {
        overlay = spawn_one_character(chrxpos[target], chrypos[target], chrzpos[target],
                                      enchanttype, chrteam[target], 0, chrturnleftright[target],
                                      NULL, MAXCHR);
        if (overlay < MAXCHR)
        {
          encoverlay[enchantindex] = overlay;  // Kill this character on end...
          chraitarget[overlay] = target;
          chraistate[overlay] = eveoverlay[enchanttype];
          chroverlay[overlay] = btrue;


          // Start out with ActionMJ...  Object activated
          if (madactionvalid[chrmodel[overlay]][ACTIONMJ])
          {
            chraction[overlay] = ACTIONMJ;
            chrlip[overlay] = 0;
            chrflip[overlay] = 0.0f;
            chrframe[overlay] = madactionstart[chrmodel[overlay]][ACTIONMJ];
            chrlastframe[overlay] = chrframe[overlay];
            chractionready[overlay] = bfalse;
          }
          chrlight[overlay] = 254;  // Assume it's transparent...
        }
      }
    }
    return enchantindex;
  }
  return MAXENCHANT;
}

//--------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------
void load_action_names(char* loadname)
{
  // ZZ> This function loads all of the 2 letter action names
  FILE* fileread;
  int cnt;
  char first, second;

  fileread = fs_fileOpen(PRI_NONE, NULL, loadname, "r");
  if (NULL != fileread)
  {
    cnt = 0;
    while (cnt < MAXACTION)
    {
      goto_colon(fileread);
      fscanf(fileread, "%c%c", &first, &second);
      cActionName[cnt][0] = first;
      cActionName[cnt][1] = second;
      cnt++;
    }
    fs_fileClose(fileread);
  }
  else log_warning("Problems loading model action codes (%s)\n", loadname);
}

//--------------------------------------------------------------------------------------------
void get_name(FILE* fileread, char *szName)
{
  // ZZ> This function loads a string of up to MAXCAPNAMESIZE characters, parsing
  //     it for underscores.  The szName argument is rewritten with the null terminated
  //     string
  int cnt;
  char cTmp;
  STRING szTmp;


  fscanf(fileread, "%s", szTmp);
  cnt = 0;
  while (cnt < MAXCAPNAMESIZE - 1)
  {
    cTmp = szTmp[cnt];
    if (cTmp == '_')  cTmp = ' ';
    szName[cnt] = cTmp;
    cnt++;
  }
  szName[cnt] = 0;
}

//--------------------------------------------------------------------------------------------
void read_setup(char* filename)
{
  // ZZ> This function loads the setup file

  ConfigFilePtr lConfigSetup;
  char lCurSectionName[64];
  bool_t lTempBool;
  Sint32 lTmpInt;
  char lTmpStr[24];


  lConfigSetup = OpenConfigFile(filename);
  if (lConfigSetup == NULL)
  {
    //Major Error
    log_error("Could not find setup file (%s).\n", filename);
  }
  else
  {
    globalname = filename; // heu!?

    /*********************************************

    GRAPHIC Section

    *********************************************/

    strcpy(lCurSectionName, "GRAPHIC");

    //Draw z reflection?
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "Z_REFLECTION", &CData.zreflect) == 0)
    {
      CData.zreflect = bfalse; // default
    }

    //Max number of vertrices (Should always be 100!)
    if (GetConfigIntValue(lConfigSetup, lCurSectionName, "MAX_NUMBER_VERTICES", &lTmpInt) == 0)
    {
      lTmpInt = 100; // default
    }
    CData.maxtotalmeshvertices = lTmpInt * 1024;

    //Do CData.fullscreen?
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "FULLSCREEN", &CData.fullscreen) == 0)
    {
      CData.fullscreen = bfalse; // default
    }

    //Screen Size
    if (GetConfigIntValue(lConfigSetup, lCurSectionName, "SCREENSIZE_X", &lTmpInt) == 0)
    {
      lTmpInt = 640; // default
    }
    CData.scrx = lTmpInt;
    if (GetConfigIntValue(lConfigSetup, lCurSectionName, "SCREENSIZE_Y", &lTmpInt) == 0)
    {
      lTmpInt = 480; // default
    }
    CData.scry = lTmpInt;

    //Color depth
    if (GetConfigIntValue(lConfigSetup, lCurSectionName, "COLOR_DEPTH", &lTmpInt) == 0)
    {
      lTmpInt = 16; // default
    }
    CData.scrd = lTmpInt;

    //The z depth
    if (GetConfigIntValue(lConfigSetup, lCurSectionName, "Z_DEPTH", &lTmpInt) == 0)
    {
      lTmpInt = 16; // default
    }
    CData.scrz = lTmpInt;

    //Max number of messages displayed
    if (GetConfigIntValue(lConfigSetup, lCurSectionName, "MAX_TEXT_MESSAGE", &lTmpInt) == 0)
    {
      lTmpInt = 1; // default
    }
    CData.messageon = btrue;
    CData.maxmessage = lTmpInt;
    if (CData.maxmessage < 1)  { CData.maxmessage = 1;  CData.messageon = bfalse; }
    if (CData.maxmessage > MAXMESSAGE)  { CData.maxmessage = MAXMESSAGE; }

    //Show status bars? (Life, mana, character icons, etc.)
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "STATUS_BAR", &CData.staton) == 0)
    {
      CData.staton = btrue; // default
    }
    CData.wraptolerance = 32;
    if (CData.staton)
    {
      CData.wraptolerance = 90;
    }

    //Perspective correction
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "PERSPECTIVE_CORRECT", &lTempBool) == 0)
    {
      CData.perspective = GL_FASTEST; // default
    }
    CData.perspective = lTempBool ? GL_NICEST : GL_FASTEST;

    //Enable dithering?
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "DITHERING", &CData.dither) == 0)
    {
      CData.dither = bfalse; // default
    }

    //Reflection fadeout
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "FLOOR_REFLECTION_FADEOUT", &lTempBool) == 0)
    {
      lTempBool = bfalse; // default
    }

    if (lTempBool)
    {
      CData.reffadeor = 0;
    }
    else
    {
      CData.reffadeor = 255;
    }

    //Draw Reflection?
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "REFLECTION", &CData.refon) == 0)
    {
      CData.refon = bfalse; // default
    }

    //Draw shadows?
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "SHADOWS", &CData.shaon) == 0)
    {
      CData.shaon = bfalse; // default
    }

    //Draw good shadows
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "SHADOW_AS_SPRITE", &CData.shasprite) == 0)
    {
      CData.shasprite = btrue; // default
    }

    //Draw phong mapping?
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "PHONG", &CData.phongon) == 0)
    {
      CData.phongon = btrue; // default
    }

    //Draw water with more layers?
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "MULTI_LAYER_WATER", &CData.twolayerwateron) == 0)
    {
      CData.twolayerwateron = bfalse; // default
    }

    //TODO: This is not implemented
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "OVERLAY", &CData.overlayvalid) == 0)
    {
      CData.overlayvalid = bfalse; // default
    }

    //Allow backgrounds?
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "BACKGROUND", &CData.backgroundvalid) == 0)
    {
      CData.backgroundvalid = bfalse; // default
    }

    //Enable fog?
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "FOG", &CData.fogallowed) == 0)
    {
      CData.fogallowed = bfalse; // default
    }

    //Do gourad CData.shading?
    CData.shading = GL_FLAT; // default
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "GOURAUD_SHADING", &lTempBool) != 0)
    {
      CData.shading = lTempBool ? GL_SMOOTH : GL_FLAT;
    }

    //Enable CData.antialiasing?
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "ANTIALIASING", &CData.antialiasing) == 0)
    {
      CData.antialiasing = bfalse; // default
    }

    //Do we do texture filtering?

    if (GetConfigValue(lConfigSetup, lCurSectionName, "TEXTURE_FILTERING", lTmpStr, 24) == 0)
    {
      CData.texturefilter = 1;
    }
    else if (isdigit(lTmpStr[0]))
    {
      sscanf(lTmpStr, "%d", &CData.texturefilter);
      if (CData.texturefilter >= TX_ANISOTROPIC)
      {
        int tmplevel = CData.texturefilter - TX_ANISOTROPIC + 1;
        userAnisotropy = 1 << tmplevel;
      }
    }
    else if (lTmpStr[0] == 'L' || lTmpStr[0] == 'l')  CData.texturefilter = TX_LINEAR;
    else if (lTmpStr[0] == 'B' || lTmpStr[0] == 'b')  CData.texturefilter = TX_BILINEAR;
    else if (lTmpStr[0] == 'T' || lTmpStr[0] == 't')  CData.texturefilter = TX_TRILINEAR_2;
    else if (lTmpStr[0] == 'A' || lTmpStr[0] == 'a')  CData.texturefilter = TX_ANISOTROPIC + log2Anisotropy;

    if (GetConfigValue(lConfigSetup, lCurSectionName, "PARTICLE_EFFECTS", lTmpStr, 24) == 0)
    {
      CData.particletype = PART_NORMAL;	//Default
    }
	else if (lTmpStr[0] == 'N' || lTmpStr[0] == 'n')  CData.particletype = PART_NORMAL;
    else if (lTmpStr[0] == 'S' || lTmpStr[0] == 's')  CData.particletype = PART_SMOOTH;
    else if (lTmpStr[0] == 'F' || lTmpStr[0] == 'f')  CData.particletype = PART_FAST;

    /*********************************************

    SOUND Section

    *********************************************/

    strcpy(lCurSectionName, "SOUND");

    //Enable sound
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "SOUND", &CData.soundvalid) == 0)
    {
      CData.soundvalid = bfalse; // default
    }

    //Enable music
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "MUSIC", &CData.musicvalid) == 0)
    {
      CData.musicvalid = bfalse; // default
    }

    //Music volume
    if (GetConfigIntValue(lConfigSetup, lCurSectionName, "MUSIC_VOLUME", &CData.musicvolume) == 0)
    {
      CData.musicvolume = 50; // default
    }

    //Sound volume
    if (GetConfigIntValue(lConfigSetup, lCurSectionName, "SOUND_VOLUME", &CData.soundvolume) == 0)
    {
      CData.soundvolume = 75; // default
    }

    //Max number of sound channels playing at the same time
    if (GetConfigIntValue(lConfigSetup, lCurSectionName, "MAX_SOUND_CHANNEL", &CData.maxsoundchannel) == 0)
    {
      CData.maxsoundchannel = 16; // default
    }
    if (CData.maxsoundchannel < 8) CData.maxsoundchannel = 8;
    if (CData.maxsoundchannel > 32) CData.maxsoundchannel = 32;

    //The output buffer size
    if (GetConfigIntValue(lConfigSetup, lCurSectionName, "OUPUT_BUFFER_SIZE", &CData.buffersize) == 0)
    {
      CData.buffersize = 2048; // default
    }
    if (CData.buffersize < 512) CData.buffersize = 512;
    if (CData.buffersize > 8196) CData.buffersize = 8196;


    /*********************************************

    CONTROL Section

    *********************************************/

    strcpy(lCurSectionName, "CONTROL");

    //Camera control mode
    if (GetConfigValue(lConfigSetup, lCurSectionName, "AUTOTURN_CAMERA", lTmpStr, 24) == 0)
    {
      strcpy(lTmpStr, "GOOD");   // default
    }

    if (lTmpStr[0] == 'G' || lTmpStr[0] == 'g')  CData.autoturncamera = 255;
    if (lTmpStr[0] == 'T' || lTmpStr[0] == 't')  CData.autoturncamera = btrue;
    if (lTmpStr[0] == 'F' || lTmpStr[0] == 'f')  CData.autoturncamera = bfalse;

    //[claforte] Force CData.autoturncamera to bfalse, or else it doesn't move right.
    //CData.autoturncamera = bfalse;



    /*********************************************

    NETWORK Section

    *********************************************/

    strcpy(lCurSectionName, "NETWORK");

    //Enable networking systems?
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "NETWORK_ON", &CData.networkon) == 0)
    {
      CData.networkon = bfalse; // default
    }

    //Max CData.lag
    if (GetConfigIntValue(lConfigSetup, lCurSectionName, "LAG_TOLERANCE", &lTmpInt) == 0)
    {
      lTmpInt = 2; // default
    }
    CData.lag = lTmpInt;

    //Name or IP of the host or the target to join
    if (GetConfigValue(lConfigSetup, lCurSectionName, "HOST_NAME", CData.nethostname, 64) == 0)
    {
      strcpy(CData.nethostname, "no host");   // default
    }

    //Multiplayer name
    if (GetConfigValue(lConfigSetup, lCurSectionName, "MULTIPLAYER_NAME", CData.netmessagename, 64) == 0)
    {
      strcpy(CData.netmessagename, "little Raoul");   // default
    }


    /*********************************************

    DEBUG Section

    *********************************************/

    strcpy(lCurSectionName, "DEBUG");

    //Show the FPS counter?
    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "DISPLAY_FPS", &lTempBool) == 0)
    {
      lTempBool = btrue; // default
    }
    CData.fpson = lTempBool;

    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "HIDE_MOUSE", &CData.HideMouse) == 0)
    {
      CData.HideMouse = btrue; // default
    }

    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "GRAB_MOUSE", &lTempBool) == 0)
    {
      CData.GrabMouse = SDL_GRAB_ON; // default
    }
    CData.GrabMouse = lTempBool ? SDL_GRAB_ON : SDL_GRAB_OFF;

    if (GetConfigBooleanValue(lConfigSetup, lCurSectionName, "DEVELOPER_MODE", &CData.DevMode) == 0)
    {
      CData.DevMode = btrue; // default
    }
    CloseConfigFile(lConfigSetup);

  }
}
//--------------------------------------------------------------------------------------------
void log_madused(char *savename)
{
  // ZZ> This is a debug function for checking model loads
  FILE* hFileWrite;
  int cnt;

  hFileWrite = fs_fileOpen(PRI_NONE, NULL, savename, "w");
  if (hFileWrite)
  {
    fprintf(hFileWrite, "Slot usage for objects in last module loaded...\n");
    fprintf(hFileWrite, "%d of %d frames used...\n", madloadframe, MAXFRAME);
    cnt = 0;
    while (cnt < MAXMODEL)
    {
      fprintf(hFileWrite, "%3d %32s %s\n", cnt, capclassname[cnt], madname[cnt]);
      cnt++;
    }
    fs_fileClose(hFileWrite);
  }
}

//---------------------------------------------------------------------------------------------
void make_lightdirectionlookup()
{
  // ZZ> This function builds the lighting direction table
  //     The table is used to find which direction the light is coming
  //     from, based on the four corner vertices of a mesh tile.
  Uint32 cnt;
  Uint16 tl, tr, br, bl;
  int x, y;

  for (cnt = 0; cnt < 65536; cnt++)
  {
    tl = (cnt & 0xf000) >> 12;
    tr = (cnt & 0x0f00) >> 8;
    br = (cnt & 0x00f0) >> 4;
    bl = (cnt & 0x000f);
    x = br + tr - bl - tl;
    y = br + bl - tl - tr;
    lightdirectionlookup[cnt] = (atan2(-y, x) + PI) * RAD_TO_BYTE;
  }
}

float sinlut[MAXLIGHTROTATION];
float coslut[MAXLIGHTROTATION];

//---------------------------------------------------------------------------------------------
float spek_global_lighting(int rotation, int normal, float lx, float ly, float lz)
{
  // ZZ> This function helps make_spektable
  float fTmp;
  float nx, ny, nz;
  float sinrot, cosrot;

  nx = md2normals[normal][0];
  ny = md2normals[normal][1];
  nz = md2normals[normal][2];
  sinrot = sinlut[rotation];
  cosrot = coslut[rotation];
  fTmp = cosrot * nx + sinrot * ny;
  ny = cosrot * ny - sinrot * nx;
  nx = fTmp;

  fTmp = 0;
  if (nx*lx > 0) fTmp += nx * lx;
  if (ny*ly > 0) fTmp += ny * ly;
  if (nz*lz > 0) fTmp += nz * lz;

  return fTmp;
}

//---------------------------------------------------------------------------------------------
float spek_local_lighting(int rotation, int normal)
{
  // ZZ> This function helps make_spektable
  float fTmp;
  float nx, ny, nz;
  float sinrot, cosrot;

  nx = md2normals[normal][0];
  ny = md2normals[normal][1];
  nz = md2normals[normal][2];
  sinrot = sinlut[rotation];
  cosrot = coslut[rotation];
  fTmp = cosrot * nx + sinrot * ny;

  fTmp = 0;
  if (cosrot*nx + sinrot*ny > 0) fTmp += cosrot * nx + sinrot * ny;

  return fTmp;
}


//---------------------------------------------------------------------------------------------
void make_spektable(float lx, float ly, float lz)
{
  // ZZ> This function makes a light table to fake directional lighting
  int cnt, tnc;
  float flight;

  // Build a lookup table for sin/cos
  for (cnt = 0; cnt < MAXLIGHTROTATION; cnt++)
  {
    sinlut[cnt] = sin(TWO_PI * cnt / MAXLIGHTROTATION);
    coslut[cnt] = cos(TWO_PI * cnt / MAXLIGHTROTATION);
  }

  flight = sqrt(lx * lx + ly * ly + lz * lz);
  for (cnt = 0; cnt < MD2LIGHTINDICES - 1; cnt++)  // Spikey mace
  {
    for (tnc = 0; tnc < MAXLIGHTROTATION; tnc++)
    {
      spek_global[tnc][cnt] = lightambi * flight * spek_global_lighting(tnc, cnt, lx / flight, ly / flight, lz / flight);
    }
  }

  for (cnt = 0; cnt < MD2LIGHTINDICES - 1; cnt++)  // Spikey mace
  {
    for (tnc = 0; tnc < MAXLIGHTROTATION; tnc++)
    {
      spek_local[tnc][cnt] = spek_local_lighting(tnc, cnt);
    }
  }

  // Fill in index number 162 for the spike mace
  for (tnc = 0; tnc < MAXLIGHTROTATION; tnc++)
  {
    spek_global[tnc][MD2LIGHTINDICES-1] = 0;
    spek_local[tnc][cnt]                = 0;
  }
}

//---------------------------------------------------------------------------------------------
void make_lighttospek(void)
{
  // ZZ> This function makes a light table to fake directional lighting
  int cnt, tnc;
//  Uint8 spek;
//  float fTmp, fPow;


  // New routine
  for (cnt = 0; cnt < MAXSPEKLEVEL; cnt++)
  {
    for (tnc = 0; tnc < 256; tnc++)
    {
      lighttospek[cnt][tnc] = 255 * pow(tnc / 255.0f, 1.0 + cnt / 2.0f);
      lighttospek[cnt][tnc] = MIN(255, lighttospek[cnt][tnc]);

      //fTmp = tnc/256.0;
      //fPow = (fTmp*4.0)+1;
      //fTmp = pow(fTmp, fPow);
      //fTmp = fTmp*cnt*255.0/MAXSPEKLEVEL;
      //if(fTmp<0) fTmp=0;
      //if(fTmp>255) fTmp=255;
      //spek = fTmp;
      //spek >>=1;
      //lighttospek[cnt][tnc] = spek;
    }
  }
}

//---------------------------------------------------------------------------------------------
int vertexconnected(int modelindex, int vertex)
{
  // ZZ> This function returns 1 if the model vertex is connected, 0 otherwise
  int cnt, tnc, entry;

  entry = 0;
  for (cnt = 0; cnt < madcommands[modelindex]; cnt++)
  {
    for (tnc = 0; tnc < madcommandsize[modelindex][cnt]; tnc++)
    {
      if (madcommandvrt[modelindex][entry] == vertex)
      {
        // The vertex is used
        return 1;
      }
      entry++;
    }
  }

  // The vertex is not used
  return 0;
}

//---------------------------------------------------------------------------------------------
void get_madtransvertices(int modelindex)
{
  // ZZ> This function gets the number of vertices to transform for a model...
  //     That means every one except the grip ( unconnected ) vertices
  int cnt, trans = 0;

  for (cnt = 0; cnt < madvertices[modelindex]; cnt++)
    trans += vertexconnected(modelindex, cnt);

  madtransvertices[modelindex] = trans;
}

//---------------------------------------------------------------------------------------------
int rip_md2_header(void)
{
  // ZZ> This function makes sure an md2 is really an md2
  int iTmp;
  int* ipIntPointer;

  // Check the file type
  ipIntPointer = (int*) cLoadBuffer;
  iTmp = ipIntPointer[0];

#if SDL_BYTEORDER != SDL_LIL_ENDIAN
  iTmp = SDL_Swap32(iTmp);
#endif

  if (iTmp != MD2START) return bfalse;

  return btrue;
}

//---------------------------------------------------------------------------------------------
void fix_md2_normals(Uint16 modelindex)
{
  // ZZ> This function helps light not flicker so much
  int cnt, tnc;
  Uint8 indexofcurrent, indexofnext, indexofnextnext, indexofnextnextnext;
  Uint8 indexofnextnextnextnext;
  Uint32 frame;

  frame = madframestart[modelindex];
  cnt = 0;
  while (cnt < madvertices[modelindex])
  {
    tnc = 0;
    while (tnc < madframes[modelindex])
    {
      indexofcurrent = madvrta[frame][cnt];
      indexofnext = madvrta[frame+1][cnt];
      indexofnextnext = madvrta[frame+2][cnt];
      indexofnextnextnext = madvrta[frame+3][cnt];
      indexofnextnextnextnext = madvrta[frame+4][cnt];
      if (indexofcurrent == indexofnextnext && indexofnext != indexofcurrent)
      {
        madvrta[frame+1][cnt] = indexofcurrent;
      }
      if (indexofcurrent == indexofnextnextnext)
      {
        if (indexofnext != indexofcurrent)
        {
          madvrta[frame+1][cnt] = indexofcurrent;
        }
        if (indexofnextnext != indexofcurrent)
        {
          madvrta[frame+2][cnt] = indexofcurrent;
        }
      }
      if (indexofcurrent == indexofnextnextnextnext)
      {
        if (indexofnext != indexofcurrent)
        {
          madvrta[frame+1][cnt] = indexofcurrent;
        }
        if (indexofnextnext != indexofcurrent)
        {
          madvrta[frame+2][cnt] = indexofcurrent;
        }
        if (indexofnextnextnext != indexofcurrent)
        {
          madvrta[frame+3][cnt] = indexofcurrent;
        }
      }
      tnc++;
    }
    cnt++;
  }
}

#if SDL_BYTEORDER != SDL_LIL_ENDIAN
float LoadFloatByteswapped( float *ptr )
{
  union {
    float f;
    int i;
  } u;

  u.f = *ptr;
  u.i = SDL_Swap32(u.i);

  return u.f;
}
#endif

//---------------------------------------------------------------------------------------------
void rip_md2_commands(Uint16 modelindex)
{
  // ZZ> This function converts an md2's GL commands into our little command list thing
  int iTmp;
  float fTmpu, fTmpv;
  int iNumVertices;
  int tnc;

  char* cpCharPointer = (char*) cLoadBuffer;
  int* ipIntPointer = (int*) cLoadBuffer;
  float* fpFloatPointer = (float*) cLoadBuffer;

  // Number of GL commands in the MD2
  int iNumCommands = ipIntPointer[9];

#if SDL_BYTEORDER != SDL_LIL_ENDIAN
  iNumCommands = SDL_Swap32(iNumCommands);
#endif

  // Offset (in DWORDS) from the start of the file to the gl command list.
  int iCommandOffset = ipIntPointer[15] >> 2;

#if SDL_BYTEORDER != SDL_LIL_ENDIAN
  iCommandOffset = SDL_Swap32(iCommandOffset);
#endif

  // Read in each command
  // iNumCommands isn't the number of commands, rather the number of dwords in
  // the command list...  Use iCommandCount to figure out how many we use
  int iCommandCount = 0;
  int entry = 0;

  int cnt = 0;
  while (cnt < iNumCommands)
  {
    iNumVertices = ipIntPointer[iCommandOffset];

#if SDL_BYTEORDER != SDL_LIL_ENDIAN
    iNumVertices = SDL_Swap32(iNumVertices);
#endif

    iCommandOffset++;
    cnt++;

    if (iNumVertices != 0)
    {
      if (iNumVertices < 0)
      {
        // Fans start with a negative
        iNumVertices = -iNumVertices;
        // PORT: madcommandtype[modelindex][iCommandCount] = (Uint8) D3DPT_TRIANGLEFAN;
        madcommandtype[modelindex][iCommandCount] = GL_TRIANGLE_FAN;
        madcommandsize[modelindex][iCommandCount] = (Uint8) iNumVertices;
      }
      else
      {
        // Strips start with a positive
        // PORT: madcommandtype[modelindex][iCommandCount] = (Uint8) D3DPT_TRIANGLESTRIP;
        madcommandtype[modelindex][iCommandCount] = GL_TRIANGLE_STRIP;
        madcommandsize[modelindex][iCommandCount] = (Uint8) iNumVertices;
      }

      // Read in vertices for each command
      tnc = 0;
      while (tnc < iNumVertices)
      {
        fTmpu = fpFloatPointer[iCommandOffset];  iCommandOffset++;  cnt++;
        fTmpv = fpFloatPointer[iCommandOffset];  iCommandOffset++;  cnt++;
        iTmp = ipIntPointer[iCommandOffset];  iCommandOffset++;  cnt++;

#if SDL_BYTEORDER != SDL_LIL_ENDIAN
        fTmpu = LoadFloatByteswapped(&fTmpu);
        fTmpv = LoadFloatByteswapped(&fTmpv);
        iTmp = SDL_Swap32(iTmp);
#endif
        madcommandu[modelindex][entry] = fTmpu - (.5 / 64); // GL doesn't align correctly
        madcommandv[modelindex][entry] = fTmpv - (.5 / 64); // with D3D
        madcommandvrt[modelindex][entry] = (Uint16) iTmp;
        entry++;
        tnc++;
      }
      iCommandCount++;
    }
  }
  madcommands[modelindex] = iCommandCount;
}

//---------------------------------------------------------------------------------------------
int rip_md2_frame_name(int frame)
{
  // ZZ> This function gets frame names from the load buffer, it returns
  //     btrue if the name in cFrameName[] is valid
  int iFrameOffset;
  int iNumVertices;
  int iNumFrames;
  int cnt;
  int* ipNamePointer;
  int* ipIntPointer;
  bool_t foundname;

  // Jump to the Frames section of the md2 data
  ipNamePointer = (int*) cFrameName;
  ipIntPointer = (int*) cLoadBuffer;


  iNumVertices = ipIntPointer[6];
  iNumFrames = ipIntPointer[10];
  iFrameOffset = ipIntPointer[14] >> 2;

#if SDL_BYTEORDER != SDL_LIL_ENDIAN
  iNumVertices = SDL_Swap32(iNumVertices);
  iNumFrames = SDL_Swap32(iNumFrames);
  iFrameOffset = SDL_Swap32(iFrameOffset);
#endif


  // Chug through each frame
  foundname = bfalse;
  cnt = 0;
  while (cnt < iNumFrames && !foundname)
  {
    iFrameOffset += 6;
    if (cnt == frame)
    {
      ipNamePointer[0] = ipIntPointer[iFrameOffset]; iFrameOffset++;
      ipNamePointer[1] = ipIntPointer[iFrameOffset]; iFrameOffset++;
      ipNamePointer[2] = ipIntPointer[iFrameOffset]; iFrameOffset++;
      ipNamePointer[3] = ipIntPointer[iFrameOffset]; iFrameOffset++;
      foundname = btrue;
    }
    else
    {
      iFrameOffset += 4;
    }
    iFrameOffset += iNumVertices;
    cnt++;
  }
  cFrameName[15] = 0;  // Make sure it's null terminated
  return foundname;
}

//---------------------------------------------------------------------------------------------
void rip_md2_frames(Uint16 modelindex)
{
  // ZZ> This function gets frames from the load buffer and adds them to
  //     the indexed model
  Uint8 cTmpx, cTmpy, cTmpz;
  Uint8 cTmpNormalIndex;
  float fRealx, fRealy, fRealz;
  float fScalex, fScaley, fScalez;
  float fTranslatex, fTranslatey, fTranslatez;
  int iFrameOffset;
  int iNumVertices;
  int iNumFrames;
  int cnt, tnc;
  char* cpCharPointer;
  int* ipIntPointer;
  float* fpFloatPointer;


  // Jump to the Frames section of the md2 data
  cpCharPointer = (char*) cLoadBuffer;
  ipIntPointer = (int*) cLoadBuffer;
  fpFloatPointer = (float*) cLoadBuffer;


  iNumVertices = ipIntPointer[6];
  iNumFrames = ipIntPointer[10];
  iFrameOffset = ipIntPointer[14] >> 2;

#if SDL_BYTEORDER != SDL_LIL_ENDIAN
  iNumVertices = SDL_Swap32(iNumVertices);
  iNumFrames = SDL_Swap32(iNumFrames);
  iFrameOffset = SDL_Swap32(iFrameOffset);
#endif


  // Read in each frame
  madframestart[modelindex] = madloadframe;
  madframes[modelindex] = iNumFrames;
  madvertices[modelindex] = iNumVertices;
  madscale[modelindex] = (float)(1.0 / 320.0); // Scale each vertex float to fit it in a short
  cnt = 0;
  while (cnt < iNumFrames && madloadframe < MAXFRAME)
  {
    fScalex = fpFloatPointer[iFrameOffset]; iFrameOffset++;
    fScaley = fpFloatPointer[iFrameOffset]; iFrameOffset++;
    fScalez = fpFloatPointer[iFrameOffset]; iFrameOffset++;
    fTranslatex = fpFloatPointer[iFrameOffset]; iFrameOffset++;
    fTranslatey = fpFloatPointer[iFrameOffset]; iFrameOffset++;
    fTranslatez = fpFloatPointer[iFrameOffset]; iFrameOffset++;

#if SDL_BYTEORDER != SDL_LIL_ENDIAN
    fScalex = LoadFloatByteswapped(&fScalex);
    fScaley = LoadFloatByteswapped(&fScaley);
    fScalez = LoadFloatByteswapped(&fScalez);

    fTranslatex = LoadFloatByteswapped(&fTranslatex);
    fTranslatey = LoadFloatByteswapped(&fTranslatey);
    fTranslatez = LoadFloatByteswapped(&fTranslatez);
#endif

    iFrameOffset += 4;
    tnc = 0;
    while (tnc < iNumVertices)
    {
      // This should work because it's reading a single character
      cTmpx = cpCharPointer[(iFrameOffset<<2)];
      cTmpy = cpCharPointer[(iFrameOffset<<2)+1];
      cTmpz = cpCharPointer[(iFrameOffset<<2)+2];
      cTmpNormalIndex = cpCharPointer[(iFrameOffset<<2)+3];
      fRealx = (cTmpx * fScalex) + fTranslatex;
      fRealy = (cTmpy * fScaley) + fTranslatey;
      fRealz = (cTmpz * fScalez) + fTranslatez;
      //            fRealx = (cTmpx*fScalex);
      //            fRealy = (cTmpy*fScaley);
      //            fRealz = (cTmpz*fScalez);
      //            madvrtx[madloadframe][tnc] = (Sint16) (fRealx*256); // HUK
      madvrtx[madloadframe][tnc] = (Sint16)(-fRealx * 256);
      madvrty[madloadframe][tnc] = (Sint16)(fRealy * 256);
      madvrtz[madloadframe][tnc] = (Sint16)(fRealz * 256);
      madvrta[madloadframe][tnc] = cTmpNormalIndex;
      iFrameOffset++;
      tnc++;
    }
    madloadframe++;
    cnt++;
  }
}

//---------------------------------------------------------------------------------------------
int load_one_md2(char* szLoadname, Uint16 modelindex)
{
  // ZZ> This function loads an id md2 file, storing the converted data in the indexed model
  //    int iFileHandleRead;
  size_t iBytesRead = 0;
  int iReturnValue;

  // Read the input file
  FILE *file = fs_fileOpen(PRI_NONE, NULL, szLoadname, "rb");
  if (NULL==file) return bfalse;

  // Read up to MD2MAXLOADSIZE bytes from the file into the cLoadBuffer array.
  iBytesRead = fread(cLoadBuffer, 1, MD2MAXLOADSIZE, file);
  if (iBytesRead == 0)
    return bfalse;

  // Check the header
  // TODO: Verify that the header's filesize correspond to iBytesRead.
  iReturnValue = rip_md2_header();
  if (0==iReturnValue)
    return bfalse;

  // Get the frame vertices
  rip_md2_frames(modelindex);
  // Get the commands
  rip_md2_commands(modelindex);
  // Fix them normals
  //fix_md2_normals(modelindex);
  // Figure out how many vertices to transform
  get_madtransvertices(modelindex);

  fs_fileClose(file);

  return btrue;
}

//--------------------------------------------------------------------------------------------
void make_enviro(void)
{
  // ZZ> This function sets up the environment mapping table
  int cnt;
  float z;
  float x, y;

  // Find the environment map positions
  for (cnt = 0; cnt < MD2LIGHTINDICES; cnt++)
  {
    x = md2normals[cnt][0];
    y = md2normals[cnt][1];
    x = 1.0f + atan2(y, x)/PI;
    x--;

    if (x < 0)
      x--;

    indextoenvirox[cnt] = x;
  }

  for (cnt = 0; cnt < 256; cnt++)
  {
    z = cnt / 256.0;  // Z is between 0 and 1
    lighttoenviroy[cnt] = z;
  }
}

//--------------------------------------------------------------------------------------------
void show_stat(Uint16 statindex)
{
  // ZZ> This function shows the more specific stats for a character
  Uint16 character, level;
  char gender[8];

  if (statdelay == 0)
  {
    if (statindex < numstat)
    {
      character = statlist[statindex];


      // Name
      debug_message(1, "=%s=", chrname[character]);

      // Level and gender and class
      gender[0] = 0;
      if (chralive[character])
      {
        if (chrgender[character] == GENMALE)
        {
          snprintf(gender, sizeof(gender), "male ");
        }
        if (chrgender[character] == GENFEMALE)
        {
          snprintf(gender, sizeof(gender), "female ");
        }

        level = chrexperiencelevel[character];
        if (level == 0)
          debug_message(1, " 1st level %s%s", gender, capclassname[chrmodel[character]]);
        if (level == 1)
          debug_message(1, " 2nd level %s%s", gender, capclassname[chrmodel[character]]);
        if (level == 2)
          debug_message(1, " 3rd level %s%s", gender, capclassname[chrmodel[character]]);
        if (level >  2)
          debug_message(1, " %dth level %s%s", level + 1, gender, capclassname[chrmodel[character]]);
      }
      else
      {
        debug_message(1, " Dead %s", capclassname[chrmodel[character]]);
      }

      // Stats
      debug_message(1, " STR:%2d ~WIS:%2d ~DEF:%d", chrstrength[character] >> 8, chrwisdom[character] >> 8, 255 - chrdefense[character]);
      debug_message(1, " INT:%2d ~DEX:%2d ~EXP:%d", chrintelligence[character] >> 8, chrdexterity[character] >> 8, chrexperience[character]);

      statdelay = 10;
    }
  }
}


//--------------------------------------------------------------------------------------------
void check_stats()
{
  // ZZ> This function lets the players check character stats
  if (!netmessagemode)
  {
    if (SDLKEYDOWN(SDLK_1))  show_stat(0);
    if (SDLKEYDOWN(SDLK_2))  show_stat(1);
    if (SDLKEYDOWN(SDLK_3))  show_stat(2);
    if (SDLKEYDOWN(SDLK_4))  show_stat(3);
    if (SDLKEYDOWN(SDLK_5))  show_stat(4);
    if (SDLKEYDOWN(SDLK_6))  show_stat(5);
    if (SDLKEYDOWN(SDLK_7))  show_stat(6);
    if (SDLKEYDOWN(SDLK_8))  show_stat(7);

    // !!!BAD!!!  CHEAT
    if (CData.DevMode && SDLKEYDOWN(SDLK_x))
    {
      if (SDLKEYDOWN(SDLK_1) && plaindex[0] < MAXCHR)  give_experience(plaindex[0], 25, XPDIRECT);
      if (SDLKEYDOWN(SDLK_2) && plaindex[1] < MAXCHR)  give_experience(plaindex[1], 25, XPDIRECT);
      if (SDLKEYDOWN(SDLK_3) && plaindex[2] < MAXCHR)  give_experience(plaindex[2], 25, XPDIRECT);
      if (SDLKEYDOWN(SDLK_4) && plaindex[3] < MAXCHR)  give_experience(plaindex[3], 25, XPDIRECT);
      if (SDLKEYDOWN(SDLK_5) && plaindex[4] < MAXCHR)  give_experience(plaindex[4], 25, XPDIRECT);
      if (SDLKEYDOWN(SDLK_6) && plaindex[5] < MAXCHR)  give_experience(plaindex[5], 25, XPDIRECT);
      if (SDLKEYDOWN(SDLK_7) && plaindex[6] < MAXCHR)  give_experience(plaindex[6], 25, XPDIRECT);
      if (SDLKEYDOWN(SDLK_8) && plaindex[7] < MAXCHR)  give_experience(plaindex[7], 25, XPDIRECT);
      statdelay = 0;
    }
  }
}

void check_screenshot()
{
  //This function checks if we want to take a screenshot
  if (SDLKEYDOWN(SDLK_F11))
  {
    if (!dump_screenshot())                 //Take the shot, returns bfalse if failed
    {
      debug_message(1, "Error writing screenshot");
      log_warning("Error writing screenshot\n");      //Log the error in log.txt
    }
  }
}

bool_t dump_screenshot()
{
  // dumps the current screen (GL context) to a new bitmap file
  // right now it dumps it to whatever the current directory is

  // returns btrue if successful, bfalse otherwise

  SDL_Surface *screen, *temp;
  Uint8 *pixels;
  STRING buff;
  int i;
  FILE *test;

  screen = SDL_GetVideoSurface();
  temp = SDL_CreateRGBSurface(SDL_SWSURFACE, screen->w, screen->h, 24,
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
                              0x000000FF, 0x0000FF00, 0x00FF0000, 0
#else
                              0x00FF0000, 0x0000FF00, 0x000000FF, 0
#endif
                             );


  if (temp == NULL)
    return bfalse;

  pixels = malloc(3 * screen->w * screen->h);
  if (pixels == NULL)
  {
    SDL_FreeSurface(temp);
    return bfalse;
  }

  glReadPixels(0, 0, screen->w, screen->h, GL_RGB, GL_UNSIGNED_BYTE, pixels);

  for (i = 0; i < screen->h; i++)
    memcpy(((char *) temp->pixels) + temp->pitch * i, pixels + 3 * screen->w * (screen->h - i - 1), screen->w * 3);
  free(pixels);

  // find the next EGO??.BMP file for writing
  i = 0;
  test = NULL;

  do
  {
    if (test != NULL)
      fs_fileClose(test);

    snprintf(buff, sizeof(buff), "ego%02d.bmp", i);

    // lame way of checking if the file already exists...
    test = fs_fileOpen(PRI_NONE, NULL, buff, "rb");
    i++;

  }
  while ((test != NULL) && (i < 100));

  SDL_SaveBMP(temp, buff);
  SDL_FreeSurface(temp);

  debug_message(1, "Saved to %s", buff);

  return btrue;
}

//--------------------------------------------------------------------------------------------
void add_stat(Uint16 character)
{
  // ZZ> This function adds a status display to the do list
  if (numstat < MAXSTAT)
  {
    statlist[numstat] = character;
    chrstaton[character] = btrue;
    numstat++;
  }
}

//--------------------------------------------------------------------------------------------
void move_to_top(Uint16 character)
{
  // ZZ> This function puts the character on top of the statlist
  int cnt, oldloc;


  // Find where it is
  oldloc = numstat;

  for (cnt = 0; cnt < numstat; cnt++)
    if (statlist[cnt] == character)
    {
      oldloc = cnt;
      cnt = numstat;
    }

  // Change position
  if (oldloc < numstat)
  {
    // Move all the lower ones up
    while (oldloc > 0)
    {
      oldloc--;
      statlist[oldloc+1] = statlist[oldloc];
    }
    // Put the character in the top slot
    statlist[0] = character;
  }
}

//--------------------------------------------------------------------------------------------
void sort_stat()
{
  // ZZ> This function puts all of the local players on top of the statlist
  int cnt;

  for (cnt = 0; cnt < numpla; cnt++)
    if (plavalid[cnt] && pladevice[cnt] != INPUTNONE)
    {
      move_to_top(plaindex[cnt]);
    }
}

//--------------------------------------------------------------------------------------------
void move_water(float dUpdate)
{
  // ZZ> This function animates the water overlays
  int layer;

  for (layer = 0; layer < MAXWATERLAYER; layer++)
  {
    waterlayeru[layer] += waterlayeruadd[layer] * dUpdate;
    waterlayerv[layer] += waterlayervadd[layer] * dUpdate;
    if (waterlayeru[layer] > 1.0)  waterlayeru[layer] -= 1.0;
    if (waterlayerv[layer] > 1.0)  waterlayerv[layer] -= 1.0;
    if (waterlayeru[layer] < -1.0)  waterlayeru[layer] += 1.0;
    if (waterlayerv[layer] < -1.0)  waterlayerv[layer] += 1.0;
    waterlayerframe[layer] = ((int)(waterlayerframe[layer] + waterlayerframeadd[layer] * dUpdate)) & WATERFRAMEAND;
  }
}

//--------------------------------------------------------------------------------------------
void play_action(Uint16 character, Uint16 action, Uint8 actionready)
{
  // ZZ> This function starts a generic action for a character
  if (madactionvalid[chrmodel[character]][action])
  {
    chrnextaction[character] = ACTIONDA;
    chraction[character] = action;
    chrlip[character] = 0;
    chrflip[character] = 0.0f;
    chrlastframe[character] = chrframe[character];
    chrframe[character] = madactionstart[chrmodel[character]][chraction[character]];
    chractionready[character] = actionready;
  }
}

//--------------------------------------------------------------------------------------------
void set_frame(Uint16 character, Uint16 frame, Uint8 lip)
{
  // ZZ> This function sets the frame for a character explicitly...  This is used to
  //     rotate Tank turrets
  chrnextaction[character] = ACTIONDA;
  chraction[character] = ACTIONDA;
  chrlip[character] = (lip << 6);
  chrlastframe[character] = madactionstart[chrmodel[character]][ACTIONDA] + frame;
  chrframe[character] = madactionstart[chrmodel[character]][ACTIONDA] + frame + 1;
  chractionready[character] = btrue;
}

//--------------------------------------------------------------------------------------------
void reset_character_alpha(Uint16 character)
{
  // ZZ> This function fixes an item's transparency
  Uint16 enchant, mount;

  if (character != MAXCHR)
  {
    mount = chrattachedto[character];
    if (chron[character] && mount != MAXCHR && chrisitem[character] && chrtransferblend[mount])
    {
      // Okay, reset transparency
      enchant = chrfirstenchant[character];
      while (enchant < MAXENCHANT)
      {
        unset_enchant_value(enchant, SETALPHABLEND);
        unset_enchant_value(enchant, SETLIGHTBLEND);
        enchant = encnextenchant[enchant];
      }
      chralpha[character] = capalpha[chrmodel[character]];
      chrlight[character] = caplight[chrmodel[character]];
      enchant = chrfirstenchant[character];
      while (enchant < MAXENCHANT)
      {
        set_enchant_value(enchant, SETALPHABLEND, enceve[enchant]);
        set_enchant_value(enchant, SETLIGHTBLEND, enceve[enchant]);
        enchant = encnextenchant[enchant];
      }
    }
  }
}

//--------------------------------------------------------------------------------------------
int generate_number(int numbase, int numrand)
{
  int tmp;

  // ZZ> This function generates a random number
  if (numrand <= 0)
  {
    log_error("One of the data pairs in randomization is wrong! (%i, %i)\n", numbase, numrand);
  }
  tmp = (rand() % numrand) + numbase;
  return (tmp);
}

//--------------------------------------------------------------------------------------------
void drop_money(Uint16 character, Uint16 money)
{
  // ZZ> This function drops some of a character's money
  Uint16 huns, tfives, fives, ones, cnt;

  if (money > chrmoney[character])  money = chrmoney[character];
  if (money > 0 && chrzpos[character] > -2)
  {
    chrmoney[character] -= money;
    huns   = money / 100;  money -= huns   * 100;
    tfives = money /  25;  money -= tfives *  25;
    fives  = money /   5;  money -= fives  *   5;
    ones   = money;

    for (cnt = 0; cnt < ones; cnt++)
      spawn_one_particle(chrxpos[character], chrypos[character],  chrzpos[character], 0, MAXMODEL, PRTPIP_COIN_001, MAXCHR, SPAWNLAST, NULLTEAM, MAXCHR, cnt, MAXCHR);

    for (cnt = 0; cnt < fives; cnt++)
      spawn_one_particle(chrxpos[character], chrypos[character],  chrzpos[character], 0, MAXMODEL, PRTPIP_COIN_005, MAXCHR, SPAWNLAST, NULLTEAM, MAXCHR, cnt, MAXCHR);

    for (cnt = 0; cnt < tfives; cnt++)
      spawn_one_particle(chrxpos[character], chrypos[character],  chrzpos[character], 0, MAXMODEL, PRTPIP_COIN_025, MAXCHR, SPAWNLAST, NULLTEAM, MAXCHR, cnt, MAXCHR);

    for (cnt = 0; cnt < huns; cnt++)
      spawn_one_particle(chrxpos[character], chrypos[character],  chrzpos[character], 0, MAXMODEL, PRTPIP_COIN_100, MAXCHR, SPAWNLAST, NULLTEAM, MAXCHR, cnt, MAXCHR);

    chrdamagetime[character] = DAMAGETIME;  // So it doesn't grab it again
  }
}

//--------------------------------------------------------------------------------------------
void call_for_help(Uint16 character)
{
  // ZZ> This function issues a call for help to all allies
  Uint8 team;
  Uint16 cnt;

  team = chrteam[character];
  teamsissy[team] = character;

  for (cnt = 0; cnt < MAXCHR; cnt++)
    if (chron[cnt] && cnt != character && !teamhatesteam[chrteam[cnt]][team])
      chralert[cnt] |= ALERTIFCALLEDFORHELP;
}

//--------------------------------------------------------------------------------------------
void give_experience(Uint16 character, int amount, Uint8 xptype)
{
  // ZZ> This function gives a character experience, and pawns off level gains to
  //     another function
  int newamount;
  int curlevel, nextexperience;
  int number;
  int profile;


  if (chrinvictus[character]) return;


  // Figure out how much experience to give
  profile = chrmodel[character];
  newamount = amount;
  if (xptype < MAXEXPERIENCETYPE)
  {
    newamount = amount * capexperiencerate[profile][xptype];
  }
  newamount += chrexperience[character];
  if (newamount > MAXXP)  newamount = MAXXP;
  chrexperience[character] = newamount;


  // Do level ups and stat changes
  curlevel       = chrexperiencelevel[character];
  nextexperience = (curlevel<MAXLEVEL-1) ? capexperienceforlevel[profile][curlevel+1] : capexperiencecoeff[profile]*pow(curlevel+1,capexperiencepower[profile]);
  if (chrexperience[character] >= nextexperience)
  {
    // The character is ready to advance...
    if (chrisplayer[character])
    {
      debug_message(1, "%s gained a level!!!", chrname[character]);

      snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s", CData.basicdat_dir, CData.lvlup_sound);
      sound = Mix_LoadWAV(CStringTmp1);

      Mix_PlayChannel(-1, sound, 0);
    }
    chrexperiencelevel[character]++;

    // Size
    if((chrsizegoto[character] + capsizeperlevel[profile]) < 1+(capsizeperlevel[profile]*10)) chrsizegoto[character] += capsizeperlevel[profile];
	chrsizegototime[character] = SIZETIME*100;

    // Strength
    number = generate_number(capstrengthperlevelbase[profile], capstrengthperlevelrand[profile]);
    number += chrstrength[character];
    if (number > PERFECTSTAT) number = PERFECTSTAT;
    chrstrength[character] = number;

    // Wisdom
    number = generate_number(capwisdomperlevelbase[profile], capwisdomperlevelrand[profile]);
    number += chrwisdom[character];
    if (number > PERFECTSTAT) number = PERFECTSTAT;
    chrwisdom[character] = number;

    // Intelligence
    number = generate_number(capintelligenceperlevelbase[profile], capintelligenceperlevelrand[profile]);
    number += chrintelligence[character];
    if (number > PERFECTSTAT) number = PERFECTSTAT;
    chrintelligence[character] = number;

    // Dexterity
    number = generate_number(capdexterityperlevelbase[profile], capdexterityperlevelrand[profile]);
    number += chrdexterity[character];
    if (number > PERFECTSTAT) number = PERFECTSTAT;
    chrdexterity[character] = number;

    // Life
    number = generate_number(caplifeperlevelbase[profile], caplifeperlevelrand[profile]);
    number += chrlifemax[character];
    if (number > PERFECTBIG) number = PERFECTBIG;
    chrlife[character] += (number - chrlifemax[character]);
    chrlifemax[character] = number;

    // Mana
    number = generate_number(capmanaperlevelbase[profile], capmanaperlevelrand[profile]);
    number += chrmanamax[character];
    if (number > PERFECTBIG) number = PERFECTBIG;
    chrmana[character] += (number - chrmanamax[character]);
    chrmanamax[character] = number;

    // Mana Return
    number = generate_number(capmanareturnperlevelbase[profile], capmanareturnperlevelrand[profile]);
    number += chrmanareturn[character];
    if (number > PERFECTSTAT) number = PERFECTSTAT;
    chrmanareturn[character] = number;

    // Mana Flow
    number = generate_number(capmanaflowperlevelbase[profile], capmanaflowperlevelrand[profile]);
    number += chrmanaflow[character];
    if (number > PERFECTSTAT) number = PERFECTSTAT;
    chrmanaflow[character] = number;
  }
}


//--------------------------------------------------------------------------------------------
void give_team_experience(Uint8 team, int amount, Uint8 xptype)
{
  // ZZ> This function gives a character experience, and pawns off level gains to
  //     another function
  int cnt;

  for (cnt = 0; cnt < MAXCHR; cnt++)
    if (chrteam[cnt] == team && chron[cnt])
      give_experience(cnt, amount, xptype);
}


//--------------------------------------------------------------------------------------------
void setup_alliances(char *modname)
{
  // ZZ> This function reads the alliance file
  STRING newloadname, szTemp;
  Uint8 teama, teamb;
  FILE *fileread;


  // Load the file
  snprintf(newloadname, sizeof(CStringTmp1), "%s%s/%s", modname, CData.gamedat_dir, CData.alliance_file);
  fileread = fs_fileOpen(PRI_NONE, NULL, newloadname, "r");
  if (NULL != fileread)
  {
    while (goto_colon_yesno(fileread))
    {
      fscanf(fileread, "%s", szTemp);
      teama = (szTemp[0] - 'A') % MAXTEAM;
      fscanf(fileread, "%s", szTemp);
      teamb = (szTemp[0] - 'A') % MAXTEAM;
      teamhatesteam[teama][teamb] = bfalse;
    }
    fs_fileClose(fileread);
  }
}

//grfx.c
//--------------------------------------------------------------------------------------------
void load_mesh_fans()
{
  // ZZ> This function loads fan types for the terrain
  int cnt, entry;
  int numfantype, fantype, bigfantype, vertices;
  int numcommand, command, commandsize;
  int itmp;
  float ftmp;
  FILE* fileread;
  float offx, offy;


  // Initialize all mesh types to 0
  entry = 0;
  while (entry < MAXMESHTYPE)
  {
    meshcommandnumvertices[entry] = 0;
    meshcommands[entry] = 0;
    entry++;
  }


  // Open the file and go to it
  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s", CData.basicdat_dir, CData.fans_file);
  fileread = fs_fileOpen(PRI_NONE, NULL, CStringTmp1, "r");
  if (NULL != fileread)
  {
    goto_colon(fileread);
    fscanf(fileread, "%d", &numfantype);
    fantype = 0;
    bigfantype = MAXMESHTYPE / 2; // Duplicate for 64x64 tiles
    while (fantype < numfantype)
    {
      goto_colon(fileread);
      fscanf(fileread, "%d", &vertices);
      meshcommandnumvertices[fantype] = vertices;
      meshcommandnumvertices[bigfantype] = vertices;  // Dupe
      cnt = 0;
      while (cnt < vertices)
      {
        goto_colon(fileread);
        fscanf(fileread, "%d", &itmp);
        goto_colon(fileread);
        fscanf(fileread, "%f", &ftmp);
        meshcommandu[fantype][cnt] = ftmp;
        meshcommandu[bigfantype][cnt] = ftmp;  // Dupe
        goto_colon(fileread);
        fscanf(fileread, "%f", &ftmp);
        meshcommandv[fantype][cnt] = ftmp;
        meshcommandv[bigfantype][cnt] = ftmp;  // Dupe
        cnt++;
      }


      goto_colon(fileread);
      fscanf(fileread, "%d", &numcommand);
      meshcommands[fantype] = numcommand;
      meshcommands[bigfantype] = numcommand;  // Dupe
      entry = 0;
      command = 0;
      while (command < numcommand)
      {
        goto_colon(fileread);
        fscanf(fileread, "%d", &commandsize);
        meshcommandsize[fantype][command] = commandsize;
        meshcommandsize[bigfantype][command] = commandsize;  // Dupe
        cnt = 0;
        while (cnt < commandsize)
        {
          goto_colon(fileread);
          fscanf(fileread, "%d", &itmp);
          meshcommandvrt[fantype][entry] = itmp;
          meshcommandvrt[bigfantype][entry] = itmp;  // Dupe
          entry++;
          cnt++;
        }
        command++;
      }
      fantype++;
      bigfantype++;  // Dupe
    }
    fs_fileClose(fileread);
  }


  // Correct all of them silly texture positions for seamless tiling
  entry = 0;
  while (entry < MAXMESHTYPE / 2)
  {
    cnt = 0;
    while (cnt < meshcommandnumvertices[entry])
    {
      //            meshcommandu[entry][cnt] = ((.5/32)+(meshcommandu[entry][cnt]*31/32))/8;
      //            meshcommandv[entry][cnt] = ((.5/32)+(meshcommandv[entry][cnt]*31/32))/8;
      meshcommandu[entry][cnt] = ((.6 / 32) + (meshcommandu[entry][cnt] * 30.8 / 32)) / 8;
      meshcommandv[entry][cnt] = ((.6 / 32) + (meshcommandv[entry][cnt] * 30.8 / 32)) / 8;
      cnt++;
    }
    entry++;
  }
  // Do for big tiles too
  while (entry < MAXMESHTYPE)
  {
    cnt = 0;
    while (cnt < meshcommandnumvertices[entry])
    {
      //            meshcommandu[entry][cnt] = ((.5/64)+(meshcommandu[entry][cnt]*63/64))/4;
      //            meshcommandv[entry][cnt] = ((.5/64)+(meshcommandv[entry][cnt]*63/64))/4;
      meshcommandu[entry][cnt] = ((.6 / 64) + (meshcommandu[entry][cnt] * 62.8 / 64)) / 4;
      meshcommandv[entry][cnt] = ((.6 / 64) + (meshcommandv[entry][cnt] * 62.8 / 64)) / 4;
      cnt++;
    }
    entry++;
  }


  // Make tile texture offsets
  entry = 0;
  while (entry < MAXTILETYPE)
  {
    offx = (entry & 7) / 8.0;
    offy = (entry >> 3) / 8.0;
    meshtileoffu[entry] = offx;
    meshtileoffv[entry] = offy;
    entry++;
  }
}

//--------------------------------------------------------------------------------------------
void make_fanstart()
{
  // ZZ> This function builds a look up table to ease calculating the
  //     fan number given an x,y pair
  int cnt;


  cnt = 0;
  while (cnt < meshsizey)
  {
    meshfanstart[cnt] = meshsizex * cnt;
    cnt++;
  }

  cnt = 0;
  while (cnt < (meshsizey >> 2))
  {
    meshblockstart[cnt] = (meshsizex >> 2) * cnt;
    cnt++;
  }
}

//--------------------------------------------------------------------------------------------
void make_twist()
{
  // ZZ> This function precomputes surface normals and steep hill acceleration for
  //     the mesh
  int cnt;
  int x, y;
  float fx,fy,fz,ftmp;
  float xslide, yslide;

  
  for (cnt = 0; cnt < 256; cnt++)
  {
    y = cnt >> 4;
    x = cnt & 15;

    fy = y - 7;  // -7 to 8
    fx = x - 7;  // -7 to 8
    mapudtwist[cnt] = 32768 + fy * SLOPE;
    maplrtwist[cnt] = 32768 + fx * SLOPE;

    ftmp = fx*fx + fy*fy;
    if(ftmp > 121.0f)
    {
      fz = 0.0f;
      ftmp = sqrt(ftmp);
      fx /= ftmp;
      fy /= ftmp;
    }
    else
    {
      fz = sqrt(1.0f - ftmp/121.0f);
      fx /= 11.0f;
      fy /= 11.0f;
    }
    
    xslide = fx;
    yslide = fy;

    normtwistx[cnt] = fx;
    normtwisty[cnt] =-fy;
    normtwistz[cnt] = fz;

    flattwist[cnt] = fz > (1.0-SLIDEFIX);
  }
}

//--------------------------------------------------------------------------------------------
bool_t load_mesh(char *modname)
{
  // ZZ> This function loads the level.mpd file
  FILE* fileread;
  STRING newloadname;
  int itmp, cnt;
  float ftmp;
  int fan;
  int numvert, numfan;
  int x, y, vert;

  snprintf(newloadname, sizeof(newloadname), "%s%s/%s", modname, CData.gamedat_dir, CData.mesh_file);
  fileread = fs_fileOpen(PRI_NONE, NULL, newloadname, "rb");
  if (NULL == fileread) return bfalse;


#if SDL_BYTEORDER == SDL_LIL_ENDIAN
  fread(&itmp, 4, 1, fileread);  if (itmp != MAPID) return bfalse;
  fread(&itmp, 4, 1, fileread);  numvert = itmp;
  fread(&itmp, 4, 1, fileread);  meshsizex = itmp;
  fread(&itmp, 4, 1, fileread);  meshsizey = itmp;
#else
  fread(&itmp, 4, 1, fileread);  if ((int)SDL_Swap32(itmp) != MAPID) return bfalse;
  fread(&itmp, 4, 1, fileread);  numvert = (int)SDL_Swap32(itmp);
  fread(&itmp, 4, 1, fileread);  meshsizex = (int)SDL_Swap32(itmp);
  fread(&itmp, 4, 1, fileread);  meshsizey = (int)SDL_Swap32(itmp);
#endif

  numfan = meshsizex * meshsizey;
  meshedgex = meshsizex * 128;
  meshedgey = meshsizey * 128;
  numfanblock = ((meshsizex >> 2)) * ((meshsizey >> 2));  // MESHSIZEX MUST BE MULTIPLE OF 4
  watershift = 3;
  if (meshsizex > 16)  watershift++;
  if (meshsizex > 32)  watershift++;
  if (meshsizex > 64)  watershift++;
  if (meshsizex > 128)  watershift++;
  if (meshsizex > 256)  watershift++;


  // Load fan data
  fan = 0;
  while (fan < numfan)
  {
    fread(&itmp, 4, 1, fileread);

#if SDL_BYTEORDER != SDL_LIL_ENDIAN
    itmp = SDL_Swap32(itmp);
#endif

    meshtype[fan] = itmp >> 24;
    meshfx[fan] = itmp >> 16;
    meshtile[fan] = itmp;

    fan++;
  }
  // Load fan data
  fan = 0;
  while (fan < numfan)
  {
    fread(&itmp, 1, 1, fileread);
    meshtwist[fan] = itmp;
    fan++;
  }


  // Load vertex x data
  cnt = 0;
  while (cnt < numvert)
  {
    fread(&ftmp, 4, 1, fileread);

    meshvrtx[cnt] = ftmp;

#if SDL_BYTEORDER != SDL_LIL_ENDIAN
    meshvrtx[cnt] = LoadFloatByteswapped(&meshvrtx[cnt]);
#endif
    cnt++;
  }
  // Load vertex y data
  cnt = 0;
  while (cnt < numvert)
  {
    fread(&ftmp, 4, 1, fileread);

    meshvrty[cnt] = ftmp;

#if SDL_BYTEORDER != SDL_LIL_ENDIAN
    meshvrty[cnt] = LoadFloatByteswapped(&meshvrty[cnt]);
#endif

    cnt++;
  }
  // Load vertex z data
  cnt = 0;
  while (cnt < numvert)
  {
    fread(&ftmp, 4, 1, fileread);

#if SDL_BYTEORDER == SDL_LIL_ENDIAN
    meshvrtz[cnt] = ftmp / 16.0;  // Cartman uses 4 bit fixed point for Z
#else
    meshvrtz[cnt] = (LoadFloatByteswapped(&ftmp)) / 16.0;  // Cartman uses 4 bit fixed point for Z
#endif

    cnt++;
  }

  //if(CData.shading==GL_FLAT && !rtscontrol)
  //{
  //  // Assume fullbright
  //  cnt = 0;
  //  while(cnt < numvert)
  //  {
  //    meshvrta[cnt] = 255;
  //    meshvrtl[cnt] = 255;
  //    cnt++;
  //  }
  //}
  //else
  {
    // Load vertex a data
    cnt = 0;
    while (cnt < numvert)
    {
      fread(&itmp, 1, 1, fileread);

      meshvrta[cnt] = itmp;
      meshvrtl[cnt] = 0;

      cnt++;
    }
  }
  fs_fileClose(fileread);

  make_fanstart();


  vert = 0;
  y = 0;
  while (y < meshsizey)
  {
    x = 0;
    while (x < meshsizex)
    {
      fan = meshfanstart[y] + x;
      meshvrtstart[fan] = vert;
      vert += meshcommandnumvertices[meshtype[fan]];
      x++;
    }
    y++;
  }

  return btrue;
}

//--------------------------------------------------------------------------------------------
void check_respawn()
{
  int cnt;

  for(cnt=0; cnt<MAXCHR; cnt++)
  {
    // Let players respawn
    if (respawnvalid && !chralive[cnt] && HAS_BITS(chrlatchbutton[cnt],LATCHBUTTONRESPAWN))
    {
      respawn_character(cnt);
      teamleader[chrteam[cnt]] = cnt;
      chralert[cnt] |= ALERTIFCLEANEDUP;

      // Cost some experience for doing this...
      chrexperience[cnt] *= EXPKEEP;
    }
    chrlatchbutton[cnt] &= ~LATCHBUTTONRESPAWN;
  }

};




//--------------------------------------------------------------------------------------------
void update_game(float dUpdate)
{
  // ZZ> This function does several iterations of character movements and such
  //     to keep the game in sync.
  int    cnt, numdead;

  // Check for all local players being dead
  alllocalpladead = bfalse;
  somelocalpladead = bfalse;
  localseeinvisible = bfalse;
  localseekurse = bfalse;
  numdead = 0;
  for (cnt = 0; cnt < MAXPLAYER; cnt++)
  {
    if(!plavalid[cnt] || INPUTNONE==pladevice[cnt]) continue;

    if (!chralive[plaindex[cnt]] && chrislocalplayer[plaindex[cnt]])
    {
      numdead++;
    };

    if (chrcanseeinvisible[plaindex[cnt]])
    {
      localseeinvisible = btrue;
    }

    if (chrcanseekurse[plaindex[cnt]])
    {
      localseekurse = btrue;
    }

  }

  somelocalpladead = (numdead > 0);
  alllocalpladead  = (numdead >= numlocalpla);

  // This is the main game loop
  msgtimechange = 0;

  // [claforte Jan 6th 2001]
  // TODO: Put that back in place once networking is functional.
  // Important stuff to keep in sync
  while ( (wldclock<allclock) && (AClientState.numplatimes > 0))
  {
    srand(randsave);

    PROFILE_BEGIN(resize_characters);
    resize_characters(dUpdate);
    PROFILE_END(resize_characters);

    PROFILE_BEGIN(keep_weapons_with_holders);
    keep_weapons_with_holders();
    PROFILE_END(keep_weapons_with_holders);

    PROFILE_BEGIN(let_ai_think);
    let_ai_think();
    PROFILE_END(let_ai_think);

    PROFILE_BEGIN(do_weather_spawn);
    do_weather_spawn(dUpdate);
    PROFILE_END(do_weather_spawn);

    PROFILE_BEGIN(do_enchant_spawn);
    do_enchant_spawn(dUpdate);
    PROFILE_END(do_enchant_spawn);


    // unbuffer the updated latches after let_ai_think() and before move_characters()
    PROFILE_BEGIN(cl_unbufferLatches);
    cl_unbufferLatches(&AClientState);
    PROFILE_END(cl_unbufferLatches);

    PROFILE_BEGIN(sv_unbufferLatches);
    sv_unbufferLatches(&AServerState);
    PROFILE_END(sv_unbufferLatches);


    PROFILE_BEGIN(check_respawn);
    check_respawn();
    PROFILE_END(check_respawn);


    PROFILE_BEGIN(move_characters);
    move_characters(dUpdate);
    PROFILE_END(move_characters);

    PROFILE_BEGIN(move_particles);
    move_particles(dUpdate);
    PROFILE_END(move_particles);

    PROFILE_BEGIN(make_character_matrices);
    make_character_matrices();
    PROFILE_END(make_character_matrices);

    PROFILE_BEGIN(attach_particles);
    attach_particles();
    PROFILE_END(attach_particles);

    PROFILE_BEGIN(make_onwhichfan);
    make_onwhichfan();
    PROFILE_END(make_onwhichfan);

    PROFILE_BEGIN(bump_characters);
    bump_characters(dUpdate);
    PROFILE_END(bump_characters);

    PROFILE_BEGIN(stat_return);
    stat_return(dUpdate);
    PROFILE_END(stat_return);

    PROFILE_BEGIN(pit_kill);
    pit_kill(dUpdate);
    PROFILE_END(pit_kill);


    // Generate the new seed
    randsave += *((Uint32*) & md2normals[wldframe&127][0]);
    randsave += *((Uint32*) & md2normals[randsave&127][1]);

    // Stuff for which sync doesn't matter
    PROFILE_BEGIN(flash_select);
    flash_select(dUpdate);
    PROFILE_END(flash_select);

    PROFILE_BEGIN(animate_tiles);
    animate_tiles(dUpdate);
    PROFILE_END(animate_tiles);

    PROFILE_BEGIN(move_water);
    move_water(dUpdate);
    PROFILE_END(move_water);

    // Timers
    wldclock += UPDATESKIP;
    wldframe++;
    msgtimechange++;
    if (statdelay > 0)  statdelay--;
    statclock++;
  }

  {
    if (AClientState.numplatimes == 0)
    {
      // The remote ran out of messages, and is now twiddling its thumbs...
      // Make it go slower so it doesn't happen again
      wldclock += UPDATESKIP/4.0f;
    }
    else if (!hostactive && AClientState.numplatimes > 3)
    {
      // The host has too many messages, and is probably experiencing control
      // CData.lag...  Speed it up so it gets closer to sync
      wldclock -= UPDATESKIP/4.0f;
    }
  }
}

//--------------------------------------------------------------------------------------------
void update_timers()
{
  // ZZ> This function updates the game timers
  lstclock = allclock;
  allclock = SDL_GetTicks() - sttclock;
  ups_clock += allclock - lstclock;
  fps_clock += allclock - lstclock;

  if(ups_loops > 0)
  {
    stabilized_ups        = stabilized_ups*0.99 + 0.01* (float)ups_loops / ((float)ups_clock / TICKS_PER_SEC );
    stabilized_ups_weight = stabilized_ups_weight*0.99 + 0.01;
  };

  if(fps_loops > 0)
  {
    stabilized_fps        = stabilized_fps*0.99 + 0.01 * (float)fps_loops / ((float)fps_clock / TICKS_PER_SEC);
    stabilized_fps_weight = stabilized_fps_weight*0.99 + 0.01;
  };

  if (ups_clock >= TICKS_PER_SEC)
  {
    ups_clock = 0;
    ups_loops = 0;
    create_szfpstext(stabilized_fps/stabilized_fps_weight, stabilized_ups/stabilized_ups_weight);
  }

  if (fps_clock >= TICKS_PER_SEC)
  {
    fps_clock = 0;
    fps_loops = 0;
    create_szfpstext(stabilized_fps/stabilized_fps_weight, stabilized_ups/stabilized_ups_weight);
  }

}

//--------------------------------------------------------------------------------------------
void read_pair(FILE* fileread)
{
  // ZZ> This function reads a damage/stat pair ( eg. 5-9 )
  char cTmp;
  float  fBase, fRand;

  fscanf(fileread, "%f", &fBase);  // The first number
  pairbase = fBase * 256;
  cTmp = get_first_letter(fileread);  // The hyphen
  if (cTmp != '-')
  {
    // Not in correct format, so fail
    pairrand = 1;
    return;
  }
  fscanf(fileread, "%f", &fRand);  // The second number
  pairrand = fRand * 256;
  pairrand -= pairbase;
  if (pairrand < 1)
    pairrand = 1;
}

//--------------------------------------------------------------------------------------------
void undo_pair(int base, int rand)
{
  // ZZ> This function generates a damage/stat pair ( eg. 3-6.5 )
  //     from the base and random values.  It set pairfrom and
  //     pairto
  pairfrom = base / 256.0;
  pairto = rand / 256.0;
  if (pairfrom < 0.0)  pairfrom = 0.0;
  if (pairto < 0.0)  pairto = 0.0;
  pairto += pairfrom;
}

//--------------------------------------------------------------------------------------------
void ftruthf(FILE* filewrite, char* text, Uint8 truth)
{
  // ZZ> This function kinda mimics fprintf for the output of
  //     btrue bfalse statements

  fprintf(filewrite, text);
  if (truth)
  {
    fprintf(filewrite, "TRUE\n");
  }
  else
  {
    fprintf(filewrite, "FALSE\n");
  }
}

//--------------------------------------------------------------------------------------------
void fdamagf(FILE* filewrite, char* text, Uint8 damagetype)
{
  // ZZ> This function kinda mimics fprintf for the output of
  //     SLASH CRUSH POKE HOLY EVIL FIRE ICE ZAP statements
  fprintf(filewrite, text);
  if (damagetype == DAMAGESLASH)
    fprintf(filewrite, "SLASH\n");
  if (damagetype == DAMAGECRUSH)
    fprintf(filewrite, "CRUSH\n");
  if (damagetype == DAMAGEPOKE)
    fprintf(filewrite, "POKE\n");
  if (damagetype == DAMAGEHOLY)
    fprintf(filewrite, "HOLY\n");
  if (damagetype == DAMAGEEVIL)
    fprintf(filewrite, "EVIL\n");
  if (damagetype == DAMAGEFIRE)
    fprintf(filewrite, "FIRE\n");
  if (damagetype == DAMAGEICE)
    fprintf(filewrite, "ICE\n");
  if (damagetype == DAMAGEZAP)
    fprintf(filewrite, "ZAP\n");
  if (damagetype == DAMAGENULL)
    fprintf(filewrite, "NONE\n");
}

//--------------------------------------------------------------------------------------------
void factiof(FILE* filewrite, char* text, Uint8 action)
{
  // ZZ> This function kinda mimics fprintf for the output of
  //     SLASH CRUSH POKE HOLY EVIL FIRE ICE ZAP statements
  fprintf(filewrite, text);
  if (action == ACTIONDA)
    fprintf(filewrite, "WALK\n");
  if (action == ACTIONUA)
    fprintf(filewrite, "UNARMED\n");
  if (action == ACTIONTA)
    fprintf(filewrite, "THRUST\n");
  if (action == ACTIONSA)
    fprintf(filewrite, "SLASH\n");
  if (action == ACTIONCA)
    fprintf(filewrite, "CHOP\n");
  if (action == ACTIONBA)
    fprintf(filewrite, "BASH\n");
  if (action == ACTIONLA)
    fprintf(filewrite, "LONGBOW\n");
  if (action == ACTIONXA)
    fprintf(filewrite, "XBOW\n");
  if (action == ACTIONFA)
    fprintf(filewrite, "FLING\n");
  if (action == ACTIONPA)
    fprintf(filewrite, "PARRY\n");
  if (action == ACTIONZA)
    fprintf(filewrite, "ZAP\n");
}

//--------------------------------------------------------------------------------------------
void fgendef(FILE* filewrite, char* text, Uint8 gender)
{
  // ZZ> This function kinda mimics fprintf for the output of
  //     MALE FEMALE OTHER statements

  fprintf(filewrite, text);
  if (gender == GENMALE)
    fprintf(filewrite, "MALE\n");
  if (gender == GENFEMALE)
    fprintf(filewrite, "FEMALE\n");
  if (gender == GENOTHER)
    fprintf(filewrite, "OTHER\n");
}

//--------------------------------------------------------------------------------------------
void fpairof(FILE* filewrite, char* text, int base, int rand)
{
  // ZZ> This function mimics fprintf in spitting out
  //     damage/stat pairs
  undo_pair(base, rand);
  fprintf(filewrite, text);
  fprintf(filewrite, "%4.2f-%4.2f\n", pairfrom, pairto);
}

//--------------------------------------------------------------------------------------------
void funderf(FILE* filewrite, char* text, char* usename)
{
  // ZZ> This function mimics fprintf in spitting out
  //     a name with underscore spaces
  char cTmp;
  int cnt;


  fprintf(filewrite, text);
  cnt = 0;
  cTmp = usename[0];
  cnt++;
  while (cTmp != 0)
  {
    if (cTmp == ' ')
    {
      fprintf(filewrite, "_");
    }
    else
    {
      fprintf(filewrite, "%c", cTmp);
    }
    cTmp = usename[cnt];
    cnt++;
  }
  fprintf(filewrite, "\n");
}

//--------------------------------------------------------------------------------------------
void get_message(FILE* fileread)
{
  // ZZ> This function loads a string into the message buffer, making sure it
  //     is null terminated.
  int cnt;
  char cTmp;
  STRING szTmp;


  if (msgtotal < MAXTOTALMESSAGE)
  {
    if (msgtotalindex >= MESSAGEBUFFERSIZE)
    {
      msgtotalindex = MESSAGEBUFFERSIZE - 1;
    }
    msgindex[msgtotal] = msgtotalindex;
    fscanf(fileread, "%s", szTmp);
    szTmp[255] = 0;
    cTmp = szTmp[0];
    cnt = 1;
    while (cTmp != 0 && msgtotalindex < MESSAGEBUFFERSIZE - 1)
    {
      if (cTmp == '_')  cTmp = ' ';
      msgtext[msgtotalindex] = cTmp;
      msgtotalindex++;
      cTmp = szTmp[cnt];
      cnt++;
    }
    msgtext[msgtotalindex] = 0;  msgtotalindex++;
    msgtotal++;
  }
}

//--------------------------------------------------------------------------------------------
void load_all_messages(char *loadname, int object)
{
  // ZZ> This function loads all of an objects messages
  FILE *fileread;


  madmsgstart[object] = 0;
  fileread = fs_fileOpen(PRI_NONE, NULL, loadname, "r");
  if (NULL != fileread)
  {
    madmsgstart[object] = msgtotal;
    while (goto_colon_yesno(fileread))
    {
      get_message(fileread);
    }
    fs_fileClose(fileread);
  }
}


//--------------------------------------------------------------------------------------------
void reset_teams()
{
  // ZZ> This function makes everyone hate everyone else
  int teama, teamb;


  teama = 0;
  while (teama < MAXTEAM)
  {
    // Make the team hate everyone
    teamb = 0;
    while (teamb < MAXTEAM)
    {
      teamhatesteam[teama][teamb] = btrue;
      teamb++;
    }
    // Make the team like itself
    teamhatesteam[teama][teama] = bfalse;
    // Set defaults
    teamleader[teama] = NOLEADER;
    teamsissy[teama] = 0;
    teammorale[teama] = 0;
    teama++;
  }


  // Keep the null team neutral
  teama = 0;
  while (teama < MAXTEAM)
  {
    teamhatesteam[teama][NULLTEAM] = bfalse;
    teamhatesteam[NULLTEAM][teama] = bfalse;
    teama++;
  }
}

//--------------------------------------------------------------------------------------------
void reset_messages()
{
  // ZZ> This makes messages safe to use
  int cnt;

  msgtotal = 0;
  msgtotalindex = 0;
  msgtimechange = 0;
  msgstart = 0;
  cnt = 0;
  while (cnt < MAXMESSAGE)
  {
    msgtime[cnt] = 0;
    cnt++;
  }
  cnt = 0;
  while (cnt < MAXTOTALMESSAGE)
  {
    msgindex[cnt] = 0;
    cnt++;
  }
  msgtext[0] = 0;
}

//--------------------------------------------------------------------------------------------
void make_randie()
{
  // ZZ> This function makes the random number table
  int tnc, cnt;


  // Fill in the basic values
  cnt = 0;
  while (cnt < MAXRAND)
  {
    randie[cnt] = rand() << 1;
    cnt++;
  }


  // Keep adjusting those values
  tnc = 0;
  while (tnc < 20)
  {
    cnt = 0;
    while (cnt < MAXRAND)
    {
      randie[cnt] += (Uint16)rand();
      cnt++;
    }
    tnc++;
  }

  // All done
  randindex = 0;
}

//--------------------------------------------------------------------------------------------
void reset_timers()
{
  // ZZ> This function resets the timers...
  sttclock = SDL_GetTicks();
  allclock = 0;
  lstclock = 0;
  wldclock = 0;
  statclock = 0;
  pitclock = 0;  pitskill = bfalse;
  wldframe = 0;
  allframe = 0;
  outofsync = bfalse;

  ups_loops = 0;
  ups_clock = 0;
  fps_loops = 0;
  fps_clock = 0;
}

extern int mnu_Run(float deltaTime);
extern int initMenus();

//--------------------------------------------------------------------------------------------
void set_default_config_data()
{
  strncpy(CData.basicdat_dir, "basicdat", sizeof(STRING));
  strncpy(CData.gamedat_dir, "gamedat" , sizeof(STRING));
  strncpy(CData.mnu_dir, "menu" , sizeof(STRING));
  strncpy(CData.globalparticles_dir, "globalparticles" , sizeof(STRING));
  strncpy(CData.modules_dir, "modules" , sizeof(STRING));
  strncpy(CData.music_dir, "music" , sizeof(STRING));
  strncpy(CData.objects_dir, "objects" , sizeof(STRING));
  strncpy(CData.import_dir, "import" , sizeof(STRING));
  strncpy(CData.players_dir, "players", sizeof(STRING));

  strncpy(CData.nullicon_bitmap, "nullicon.bmp" , sizeof(STRING));
  strncpy(CData.keybicon_bitmap, "keybicon.bmp" , sizeof(STRING));
  strncpy(CData.mousicon_bitmap, "mousicon.bmp" , sizeof(STRING));
  strncpy(CData.joyaicon_bitmap, "joyaicon.bmp" , sizeof(STRING));
  strncpy(CData.joybicon_bitmap, "joybicon.bmp" , sizeof(STRING));

  strncpy(CData.tile0_bitmap, "tile0.bmp" , sizeof(STRING));
  strncpy(CData.tile1_bitmap, "tile1.bmp" , sizeof(STRING));
  strncpy(CData.tile2_bitmap, "tile2.bmp" , sizeof(STRING));
  strncpy(CData.tile3_bitmap, "tile3.bmp" , sizeof(STRING));
  strncpy(CData.watertop_bitmap, "watertop.bmp" , sizeof(STRING));
  strncpy(CData.waterlow_bitmap, "waterlow.bmp" , sizeof(STRING));
  strncpy(CData.phong_bitmap, "phong.bmp" , sizeof(STRING));
  strncpy(CData.plan_bitmap, "plan.bmp" , sizeof(STRING));
  strncpy(CData.blip_bitmap, "blip.bmp" , sizeof(STRING));
  strncpy(CData.font_bitmap, "font.bmp" , sizeof(STRING));
  strncpy(CData.icon_bitmap, "icon.bmp" , sizeof(STRING));
  strncpy(CData.bars_bitmap, "bars.bmp" , sizeof(STRING));
  strncpy(CData.particle_bitmap, "particle_normal.bmp" , sizeof(STRING));
  strncpy(CData.title_bitmap, "title.bmp" , sizeof(STRING));

  strncpy(CData.menu_main_bitmap, "menu_main.png" , sizeof(STRING));
  strncpy(CData.menu_advent_bitmap, "menu_advent.png" , sizeof(STRING));
  strncpy(CData.menu_sleepy_bitmap, "menu_sleepy.png" , sizeof(STRING));
  strncpy(CData.menu_gnome_bitmap, "menu_gnome.png" , sizeof(STRING));


  strncpy(CData.slotused_file, "slotused.txt" , sizeof(STRING));
  strncpy(CData.passage_file, "passage.txt" , sizeof(STRING));
  strncpy(CData.aicodes_file, "aicodes.txt" , sizeof(STRING));
  strncpy(CData.actions_file, "actions.txt" , sizeof(STRING));
  strncpy(CData.alliance_file, "alliance.txt" , sizeof(STRING));
  strncpy(CData.fans_file, "fans.txt" , sizeof(STRING));
  strncpy(CData.fontdef_file, "font.txt" , sizeof(STRING));
  strncpy(CData.mnu_file, "menu.txt" , sizeof(STRING));
  strncpy(CData.money1_file, "1money.txt" , sizeof(STRING));
  strncpy(CData.money5_file, "5money.txt" , sizeof(STRING));
  strncpy(CData.money25_file, "25money.txt" , sizeof(STRING));
  strncpy(CData.money100_file, "100money.txt" , sizeof(STRING));
  strncpy(CData.weather4_file, "weather4.txt" , sizeof(STRING));
  strncpy(CData.weather5_file, "weather5.txt" , sizeof(STRING));
  strncpy(CData.script_file, "script.txt" , sizeof(STRING));
  strncpy(CData.ripple_file, "ripple.txt" , sizeof(STRING));
  strncpy(CData.scancode_file, "scancode.txt" , sizeof(STRING));
  strncpy(CData.playlist_file, "playlist.txt" , sizeof(STRING));
  strncpy(CData.spawn_file, "spawn.txt" , sizeof(STRING));
  strncpy(CData.wawalite_file, "wawalite.txt" , sizeof(STRING));
  strncpy(CData.defend_file, "defend.txt" , sizeof(STRING));
  strncpy(CData.splash_file, "splash.txt" , sizeof(STRING));
  strncpy(CData.mesh_file, "level.mpd" , sizeof(STRING));
  strncpy(CData.setup_file, "setup.txt" , sizeof(STRING));
  strncpy(CData.log_file, "log.txt", sizeof(STRING));
  strncpy(CData.controls_file, "controls.txt", sizeof(STRING));
  strncpy(CData.data_file, "data.txt", sizeof(STRING));
  strncpy(CData.copy_file, "copy.txt", sizeof(STRING));
  strncpy(CData.enchant_file, "enchant.txt", sizeof(STRING));
  strncpy(CData.message_file, "message.txt", sizeof(STRING));
  strncpy(CData.naming_file, "naming.txt", sizeof(STRING));
  strncpy(CData.modules_file, "modules.txt", sizeof(STRING));
  strncpy(CData.setup_file, "setup.txt", sizeof(STRING));
  strncpy(CData.skin_file, "skin.txt", sizeof(STRING));
  strncpy(CData.credits_file, "credits.txt", sizeof(STRING));
  strncpy(CData.quest_file, "quest.txt", sizeof(STRING));


  CData.uifont_points  = 20;
  CData.uifont_points2 = 18;
  strncpy(CData.uifont_ttf, "Negatori.ttf" , sizeof(STRING));

  strncpy(CData.coinget_sound, "coinget.wav" , sizeof(STRING));
  strncpy(CData.defend_sound, "defend.wav" , sizeof(STRING));
  strncpy(CData.coinfall_sound, "coinfall.wav" , sizeof(STRING));
  strncpy(CData.lvlup_sound, "lvlup.wav" , sizeof(STRING));

  CData.fullscreen = bfalse;
  CData.zreflect = bfalse;
  CData.maxtotalmeshvertices = 256*256*6;
  CData.scrd = 8;                   // Screen bit depth
  CData.scrx = 320;                 // Screen X size
  CData.scry = 200;                 // Screen Y size
  CData.scrz = 16;                  // Screen z-buffer depth ( 8 unsupported )
  CData.maxmessage = MAXMESSAGE;    //
  CData.messageon  = btrue;           // Messages?
  CData.wraptolerance = 80;     // Status bar
  CData.staton = btrue;                 // Draw the status bars?
  CData.overlayon = bfalse;      
  CData.perspective = GL_FASTEST;    
  CData.dither = bfalse;         
  CData.shading = GL_FLAT;       
  CData.antialiasing = bfalse;   
  CData.refon = bfalse;          
  CData.shaon = bfalse;          
  CData.texturefilter = TX_LINEAR;       
  CData.wateron = btrue;         
  CData.shasprite = bfalse;  
  CData.phongon = btrue;                // Do phong overlay? OUTDATED?
  CData.zreflect = bfalse; 
  CData.reffadeor = 0;              // 255 = Don't fade reflections
  CData.twolayerwateron = bfalse;        // Two layer water?
  CData.overlayvalid = bfalse;               // Allow large overlay?
  CData.backgroundvalid = bfalse;            // Allow large background?
  CData.fogallowed = btrue;          //
  CData.particletype = PART_NORMAL;
  CData.soundvalid = bfalse;     //Allow playing of sound?
  CData.musicvalid = bfalse;     // Allow music and loops?
  CData.networkon  = btrue;              // Try to connect?
  CData.lag        = 3;                                // Lag tolerance
  strncpy(CData.nethostname, "no host", sizeof(CData.nethostname));                            // Name for hosting session
  strncpy(CData.netmessagename, "little Raoul", sizeof(CData.netmessagename));                 // Name for messages
  CData.fpson = btrue;               // FPS displayed?
  // Debug option
  CData.GrabMouse = SDL_GRAB_ON;
  CData.HideMouse = bfalse;
  CData.DevMode  = btrue;
  // Debug option
};

//--------------------------------------------------------------------------------------------
#undef DEBUG_UPDATE_CLAMP

int SDL_main(int argc, char **argv)
{
  // ZZ> This is where the program starts and all the high level stuff happens
  double frameDuration;
  int    menuResult;
  bool_t menuActive = btrue;
  Uint32 time_current, update_last, frame_last;
  float dFrame, dUpdate;
  Uint32 dTimeUpdate, dTimeFrame;

  PROFILE_INIT(resize_characters);
  PROFILE_INIT(keep_weapons_with_holders);
  PROFILE_INIT(let_ai_think);
  PROFILE_INIT(do_weather_spawn);
  PROFILE_INIT(do_enchant_spawn);
  PROFILE_INIT(cl_unbufferLatches);
  PROFILE_INIT(sv_unbufferLatches);
  PROFILE_INIT(check_respawn);
  PROFILE_INIT(move_characters);
  PROFILE_INIT(move_particles);
  PROFILE_INIT(make_character_matrices);
  PROFILE_INIT(attach_particles);
  PROFILE_INIT(make_onwhichfan);
  PROFILE_INIT(bump_characters);
  PROFILE_INIT(stat_return);
  PROFILE_INIT(pit_kill);
  PROFILE_INIT(flash_select);
  PROFILE_INIT(animate_tiles);
  PROFILE_INIT(move_water);
  PROFILE_INIT(figure_out_what_to_draw);
  PROFILE_INIT(draw_main);

  
  set_default_config_data();

  // Initialize logging first, so that we can use it everywhere.
  log_init();
  log_setLoggingLevel(2);

  // start initializing the various subsystems
  log_message("Starting Egoboo %s...\n", VERSION);

  sys_initialize();
  g_clk_state = clock_create_state();
  clock_init(g_clk_state);
  fs_init();

  read_setup(CData.setup_file);

  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s", CData.basicdat_dir, CData.scancode_file);
  read_all_tags(CStringTmp1);

  read_controls(CData.controls_file);
  reset_ai_script();

  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s", CData.basicdat_dir, CData.aicodes_file);
  load_ai_codes(CStringTmp1);

  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s", CData.basicdat_dir, CData.actions_file);
  load_action_names(CStringTmp1);

  sdlinit(argc, argv);
  glinit(argc, argv);
  net_initialize();

  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s", CData.basicdat_dir, CData.uifont_ttf);
  ui_initialize(CStringTmp1, CData.uifont_points);

  sdlmixer_initialize();

  if (!get_mesh_memory())
  {
    log_error("Unable to initialize Mesh Memory - Reduce the maximum number of vertices (See SETUP.TXT)\n");
    return bfalse;
  }

  //Load stuff into memory
  prime_icons();
  prime_titleimage();
  make_textureoffset();  // THIS SHOULD WORK
  make_lightdirectionlookup(); // THIS SHOULD WORK
  make_turntosin();  // THIS SHOULD WORK
  make_enviro(); // THIS SHOULD WORK
  load_mesh_fans(); // THIS SHOULD WORK
  load_all_menu_images();
  load_all_music_sounds();
  initMenus();        //Start the game menu

  // initialize the bitmap font so we can use the cursor
  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s", CData.basicdat_dir, CData.font_bitmap);
  snprintf(CStringTmp2, sizeof(CStringTmp2), "%s/%s", CData.basicdat_dir, CData.fontdef_file);
  if (!load_font(CStringTmp1, CStringTmp2))
  {
    log_warning("UI unable to use load bitmap font for cursor. Files missing from %s directory\n", CData.basicdat_dir);
  };


  // Let the normal OS mouse cursor work
  SDL_WM_GrabInput(SDL_GRAB_OFF);
  SDL_ShowCursor(SDL_DISABLE);
  mouseon = bfalse;

  // Network's temporarily disabled

  clock_frameStep(g_clk_state);
  frameDuration = clock_getFrameDuration(g_clk_state);
  gameActive = bfalse;
  menuActive = btrue;

  //DEBUG
  CData.scrd = 32;				//We need to force 32 bit, since SDL_image crashes without it
  //DEBUG END

  while (gameActive || menuActive)
  {
    // Clock updates each frame
    clock_frameStep(g_clk_state);
    frameDuration = clock_getFrameDuration(g_clk_state);

    // Either run through the menus, or jump into the game
    if (menuActive)
    {
      //Play the menu music
      play_music(0, 500, -1);

      // do menus
      do_clear();
      read_input();

      //Pressed panic button
      if (SDLKEYDOWN(SDLK_q) && SDLKEYDOWN(SDLK_LCTRL))
      {
        log_info("User pressed escape button (LCTRL+Q)... Quitting game.\n");
        menuActive = bfalse;
        gameActive = bfalse;
      }

     
      ui_beginFrame(frameDuration);

      menuResult = mnu_Run((float)frameDuration);
      switch (menuResult)
      {
        case 1:
          // Go ahead and start the game
          menuActive = bfalse;
          gameActive = btrue;
          hostactive = CData.networkon;
          break;

        case - 1:
          // The user selected "Quit"
          menuActive = bfalse;
          gameActive = bfalse;
          break;
      }
      ui_endFrame();
     
      request_pageflip();

      do_pageflip();
    }

    // Do the game
    // Did we get through all the menus?
    if (gameActive)
    {
      SDL_WM_GrabInput(CData.GrabMouse);
      mouseon = btrue;

      // Start a new module
      srand((Uint32)-1);
	 
      load_module(pickedmodule);  // :TODO: Seems to be the next part to fix

      pressed = bfalse;
      make_onwhichfan();
      reset_camera();
      reset_timers();
      figure_out_what_to_draw();
      make_character_matrices();
      attach_particles();

      if (CData.networkon)
      {
        log_info("SDL_main: Loading module %s...\n", pickedmodule);
        netmessagemode = bfalse;
        netmessagedelay = 20;
        net_sayHello();
      }

      // Let the game go
      moduleActive = btrue;
      randsave = 0;
      srand(0);
      time_current = SDL_GetTicks();
      update_last = time_current - UPDATESKIP;
      frame_last  = time_current - FRAMESKIP;
      dFrame = dUpdate = 0.0f;
      while (moduleActive)
      {
        time_current = SDL_GetTicks();

        dTimeFrame  = time_current - frame_last;
        dTimeUpdate = time_current - update_last;

        if(dTimeFrame > FRAMESKIP) 
        {
          dFrame  += (float)(dTimeFrame )/(float)FRAMESKIP;
          frame_last = time_current;
        };

        if(dTimeUpdate > UPDATESKIP) 
        {
          dUpdate += (float)(dTimeUpdate)/(float)UPDATESKIP;
          update_last = time_current;
        };

        update_timers();

#ifdef DEBUG_UPDATE_CLAMP
        log_info("wldframe == %d\t dframe == %2.4f\n", wldframe, dUpdate);
#endif

        // This is the control loop
        read_input();
        check_screenshot();

        //Do game pause if needed
        //Check for pause key   //TODO: What to do in network games?
        if (!SDLKEYDOWN(SDLK_F8)) startpause = btrue;
        if (SDLKEYDOWN(SDLK_F8) && keyon && startpause)
        {
          if (gamepaused) gamepaused = bfalse;
          else gamepaused = btrue;
          startpause = bfalse;
        }

        while(dUpdate>=1.0)
        {
          // Do important things
          if (CData.networkon && waitingforplayers)
          {
            wldclock = allclock;
          }
          else if (gamepaused)
          {
            wldclock = allclock;
          }
          else
          {
            check_stats();
            set_local_latches();
            update_timers();
            check_passage_music();

            // NETWORK PORT
            {
              // buffer the existing latches
              input_net_message();
              cl_bufferLatches(&AClientState);
              sv_bufferLatches(&AServerState);

              // upload the information
              cl_talkToHost(&AClientState);
              sv_talkToRemotes(&AServerState);

              // download/handle any queued packets
              listen_for_packets();
            }

            update_game(UPDATESCALE);
          }

          dUpdate -= 1.0f;
          ups_loops++;

#ifdef DEBUG_UPDATE_CLAMP
          log_info("\t\twldframe == %d\t dframe == %2.4f\n", wldframe, dUpdate);
#endif

          move_camera(UPDATESCALE);

          // let the OS breathe
          SDL_Delay(1);
        };
          
        if(dFrame >= 1.0)
        {
          // Do the display stuff
          PROFILE_BEGIN(figure_out_what_to_draw);
          figure_out_what_to_draw();
          PROFILE_END(figure_out_what_to_draw);

          PROFILE_BEGIN(draw_main);
          draw_main((float)frameDuration);
          PROFILE_END(draw_main);

          dFrame = 0.0f;
        }

        do_pageflip();


        // Check for quitters
        // :TODO: nolocalplayers is not set correctly
        if (SDLKEYDOWN(SDLK_ESCAPE) /*|| nolocalplayers*/)
        {
          quit_module();
          gameActive = bfalse;
          menuActive = btrue;
        }

        // let the OS breathe
        SDL_Delay(1);
      }
      release_module();
      close_session();

      // Let the normal OS mouse cursor work
      SDL_WM_GrabInput(CData.GrabMouse);
      //SDL_ShowCursor(SDL_ENABLE);
    }

  }

  log_info("============ PROFILE =================\n");
  log_info("\tresize_characters - %lf\n", 1e6*PROFILE_QUERY(resize_characters));
  log_info("\tkeep_weapons_with_holders - %lf\n", 1e6*PROFILE_QUERY(keep_weapons_with_holders));
  log_info("\tlet_ai_think - %lf\n", 1e6*PROFILE_QUERY(let_ai_think));
  log_info("\tdo_weather_spawn - %lf\n", 1e6*PROFILE_QUERY(do_weather_spawn));
  log_info("\tdo_enchant_spawn - %lf\n", 1e6*PROFILE_QUERY(do_enchant_spawn));
  log_info("\tcl_unbufferLatches - %lf\n", 1e6*PROFILE_QUERY(cl_unbufferLatches));
  log_info("\tsv_unbufferLatches - %lf\n", 1e6*PROFILE_QUERY(sv_unbufferLatches));
  log_info("\tcheck_respawn - %lf\n", 1e6*PROFILE_QUERY(check_respawn));
  log_info("\tmove_characters - %lf\n", 1e6*PROFILE_QUERY(move_characters));
  log_info("\tmove_particles - %lf\n", 1e6*PROFILE_QUERY(move_particles));
  log_info("\tmake_character_matrices - %lf\n", 1e6*PROFILE_QUERY(make_character_matrices));
  log_info("\tattach_particles - %lf\n", 1e6*PROFILE_QUERY(attach_particles));
  log_info("\tmake_onwhichfan - %lf\n", 1e6*PROFILE_QUERY(make_onwhichfan));
  log_info("\tbump_characters - %lf\n", 1e6*PROFILE_QUERY(bump_characters));
  log_info("\tstat_return - %lf\n", 1e6*PROFILE_QUERY(stat_return));
  log_info("\tpit_kill - %lf\n", 1e6*PROFILE_QUERY(pit_kill));
  log_info("\tflash_select - %lf\n", 1e6*PROFILE_QUERY(flash_select));
  log_info("\tanimate_tiles - %lf\n", 1e6*PROFILE_QUERY(animate_tiles));
  log_info("\tmove_water - %lf\n", 1e6*PROFILE_QUERY(move_water));
  log_info("======================================\n");
  log_info("\tfigure_out_what_to_draw - %lf\n", 1e6*PROFILE_QUERY(figure_out_what_to_draw));
  log_info("\tdraw_main - %lf\n", 1e6*PROFILE_QUERY(draw_main));
  log_info("======================================\n\n");

  PROFILE_FREE(resize_characters);
  PROFILE_FREE(keep_weapons_with_holders);
  PROFILE_FREE(let_ai_think);
  PROFILE_FREE(do_weather_spawn);
  PROFILE_FREE(do_enchant_spawn);
  PROFILE_FREE(cl_unbufferLatches);
  PROFILE_FREE(sv_unbufferLatches);
  PROFILE_FREE(check_respawn);
  PROFILE_FREE(move_characters);
  PROFILE_FREE(move_particles);
  PROFILE_FREE(make_character_matrices);
  PROFILE_FREE(attach_particles);
  PROFILE_FREE(make_onwhichfan);
  PROFILE_FREE(bump_characters);
  PROFILE_FREE(stat_return);
  PROFILE_FREE(pit_kill);
  PROFILE_FREE(flash_select);
  PROFILE_FREE(animate_tiles);
  PROFILE_FREE(move_water);

  quit_game();
  if (!memory_cleanUp()) log_message("WARNING! COULD NOT CLEAN MEMORY AND SHUTDOWN SUPPORT SYSTEMS!!\n");

  return btrue;
}
