/* Egoboo - particle.c
 * Manages particle systems.
 */

/*
    This file is part of Egoboo.

    Egoboo is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Egoboo is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Egoboo.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "egoboo.h"
#include "mathstuff.h"
#include "Log.h"


//--------------------------------------------------------------------------------------------
void make_prtlist(void)
{
  // ZZ> This function figures out which particles are visible, and it sets up dynamic
  //     lighting
  int cnt, tnc, disx, disy, distance, slot;


  // Don't really make a list, just set to visible or not
  numdynalight = 0;
  dynadistancetobeat = MAXDYNADIST;
  cnt = 0;
  while (cnt < MAXPRT)
  {
    prtinview[cnt] = bfalse;
    if (prton[cnt])
    {
      prtinview[cnt] = meshinrenderlist[prtonwhichfan[cnt]];
      // Set up the lights we need
      if (prtdynalighton[cnt])
      {
        disx = prtxpos[cnt] - camtrackx;
        disx = ABS(disx);
        disy = prtypos[cnt] - camtracky;
        disy = ABS(disy);
        distance = disx + disy;
        if (distance < dynadistancetobeat)
        {
          if (numdynalight < MAXDYNA)
          {
            // Just add the light
            slot = numdynalight;
            dynadistance[slot] = distance;
            numdynalight++;
          }
          else
          {
            // Overwrite the worst one
            slot = 0;
            tnc = 1;
            dynadistancetobeat = dynadistance[0];
            while (tnc < MAXDYNA)
            {
              if (dynadistance[tnc] > dynadistancetobeat)
              {
                slot = tnc;
              }
              tnc++;
            }
            dynadistance[slot] = distance;


            // Find the new distance to beat
            tnc = 1;
            dynadistancetobeat = dynadistance[0];
            while (tnc < MAXDYNA)
            {
              if (dynadistance[tnc] > dynadistancetobeat)
              {
                dynadistancetobeat = dynadistance[tnc];
              }
              tnc++;
            }
          }
          dynalightlistx[slot] = prtxpos[cnt];
          dynalightlisty[slot] = prtypos[cnt];
          dynalightlistz[slot] = prtzpos[cnt];
          dynalightlevel[slot] = prtdynalightlevel[cnt];
          dynalightfalloff[slot] = prtdynalightfalloff[cnt];
        }
      }
    }
    cnt++;
  }
}

//--------------------------------------------------------------------------------------------
void free_one_particle_no_sound(Uint16 particle)
{
  // ZZ> This function sticks a particle back on the free particle stack
  freeprtlist[numfreeprt] = particle;
  numfreeprt++;
  prton[particle] = bfalse;
}

//--------------------------------------------------------------------------------------------
void play_particle_sound(Uint16 particle, Sint8 sound)
{
  //This function plays a sound effect for a particle
  if (sound != -1)
  {
    //Play local sound or else global (coins for example)
    if (prtmodel[particle] != MAXMODEL) play_sound(prtxpos[particle], prtypos[particle], capwaveindex[prtmodel[particle]][sound]);
    else  play_sound(prtxpos[particle], prtypos[particle], globalwave[sound]);
  }
}

//--------------------------------------------------------------------------------------------
void free_one_particle(Uint16 particle)
{
  // ZZ> This function sticks a particle back on the free particle stack and
  //     plays the sound associated with the particle
  int child;
  if (prtspawncharacterstate[particle] != SPAWNNOCHARACTER)
  {
    child = spawn_one_character(prtxpos[particle], prtypos[particle], prtzpos[particle],
                                prtmodel[particle], prtteam[particle], 0, prtfacing[particle],
                                NULL, MAXCHR);
    if (child != MAXCHR)
    {
      chraistate[child] = prtspawncharacterstate[particle];
      chraiowner[child] = prtchr[particle];
    }
  }
  play_particle_sound(particle, pipsoundend[prtpip[particle]]);
  freeprtlist[numfreeprt] = particle;
  numfreeprt++;
  prton[particle] = bfalse;
}

//--------------------------------------------------------------------------------------------
int get_free_particle(int force)
{
  // ZZ> This function gets an unused particle.  If all particles are in use
  //     and force is set, it grabs the first unimportant one.  The particle
  //     index is the return value
  Uint16 particle;


  // Return MAXPRT if we can't find one
  particle = MAXPRT;
  if (numfreeprt == 0)
  {
    if (force)
    {
      // Gotta find one, so go through the list
      particle = 0;
      while (particle < MAXPRT)
      {
        if (prtbumpsize[particle] == 0)
        {
          // Found one
          return particle;
        }
        particle++;
      }
    }
  }
  else
  {
    if (force || numfreeprt > (MAXPRT / 4))
    {
      // Just grab the next one
      numfreeprt--;
      particle = freeprtlist[numfreeprt];
    }
  }
  return particle;
}

//--------------------------------------------------------------------------------------------
Uint16 spawn_one_particle(float x, float y, float z,
                                  Uint16 facing, Uint16 model, Uint16 pip,
                                  Uint16 characterattach, Uint16 grip, Uint8 team,
                                  Uint16 characterorigin, Uint16 multispawn, Uint16 oldtarget)
{
  // ZZ> This function spawns a new particle, and returns the number of that particle
  int cnt, velocity;
  float xvel, yvel, zvel, tvel;
  int offsetfacing, newrand;
  Uint16 local_pip = MAXPRTPIP;

  if(pip >= MAXPRTPIP) 
    return MAXPRT;

  // Convert from local pip to global pip
  if (model < MAXMODEL && pip < PRTPIP_PEROBJECT_COUNT)
    local_pip = madprtpip[model][pip];
  
  // assume we were given a global pip
  if(MAXPRTPIP==local_pip)
    local_pip = pip;


  cnt = get_free_particle(pipforce[local_pip]);
  if (cnt == MAXPRT) return MAXPRT;

  // Necessary data for any part
  prton[cnt] = btrue;
  prtgopoof[cnt] = bfalse;
  prtpip[cnt] = local_pip;
  prtmodel[cnt] = model;
  prtinview[cnt] = bfalse;
  prtlevel[cnt] = 0;
  prtteam[cnt] = team;
  prtchr[cnt] = characterorigin;
  prtdamagetype[cnt] = pipdamagetype[local_pip];
  prtspawncharacterstate[cnt] = SPAWNNOCHARACTER;


  // Lighting and sound
  prtdynalighton[cnt] = bfalse;
  if (multispawn == 0)
  {
    prtdynalighton[cnt] = pipdynalightmode[local_pip];
    if (pipdynalightmode[local_pip] == DYNALOCAL)
    {
      prtdynalighton[cnt] = bfalse;
    }
  }
  prtdynalightlevel[cnt] = pipdynalevel[local_pip];
  prtdynalightfalloff[cnt] = pipdynafalloff[local_pip];



  // Set character attachments ( characterattach==MAXCHR means none )
  prtattachedtocharacter[cnt] = characterattach;
  prtgrip[cnt] = grip;



  // Correct facing
  facing += pipfacingbase[local_pip];


  // Targeting...
  offsetfacing = 0;
  zvel = 0;
  newrand = RANDIE;
  z += pipzspacingbase[local_pip] + (newrand & pipzspacingrand[local_pip]) - (pipzspacingrand[local_pip] >> 1);
  newrand = RANDIE;
  velocity = (pipxyvelbase[local_pip] + (newrand & pipxyvelrand[local_pip]));
  prttarget[cnt] = oldtarget;
  if (pipnewtargetonspawn[local_pip])
  {
    if (piptargetcaster[local_pip])
    {
      // Set the target to the caster
      prttarget[cnt] = characterorigin;
    }
    else
    {
      // Find a target
      prttarget[cnt] = prt_find_target(x, y, facing, piptargetangle[local_pip], piponlydamagefriendly[local_pip], bfalse, team, characterorigin, oldtarget);
      if (prttarget[cnt] != MAXCHR && !piphoming[local_pip])
      {
        facing -= glouseangle;
      }
      // Correct facing for dexterity...
      offsetfacing = 0;
      if (chrdexterity[characterorigin] < PERFECTSTAT)
      {
        // Correct facing for randomness
        offsetfacing = RANDIE;
        offsetfacing &= pipfacingrand[local_pip];
        offsetfacing -= (pipfacingrand[local_pip] >> 1);
        offsetfacing = (offsetfacing * (PERFECTSTAT - chrdexterity[characterorigin])) >> 13;  // Divided by PERFECTSTAT
      }
      if (prttarget[cnt] != MAXCHR && pipzaimspd[local_pip] != 0)
      {
        // These aren't velocities...  This is to do aiming on the Z axis
        if (velocity > 0)
        {
          xvel = chrxpos[prttarget[cnt]] - x;
          yvel = chrypos[prttarget[cnt]] - y;
          tvel = sqrt(xvel * xvel + yvel * yvel) / velocity;  // This is the number of steps...
          if (tvel > 0)
          {
            zvel = (chrzpos[prttarget[cnt]] + (chrbumpsize[prttarget[cnt]] >> 1) - z) / tvel;  // This is the zvel alteration
            if (zvel < -(pipzaimspd[local_pip] >> 1)) zvel = -(pipzaimspd[local_pip] >> 1);
            if (zvel > pipzaimspd[local_pip]) zvel = pipzaimspd[local_pip];
          }
        }
      }
    }
    // Does it go away?
    if (prttarget[cnt] == MAXCHR && pipneedtarget[local_pip])
    {
      free_one_particle(cnt);
      return MAXPRT;
    }
    // Start on top of target
    if (prttarget[cnt] != MAXCHR && pipstartontarget[local_pip])
    {
      x = chrxpos[prttarget[cnt]];
      y = chrypos[prttarget[cnt]];
    }
  }
  else
  {
    // Correct facing for randomness
    offsetfacing = RANDIE;
    offsetfacing &= pipfacingrand[local_pip];
    offsetfacing -= (pipfacingrand[local_pip] >> 1);
  }
  facing += offsetfacing;
  prtfacing[cnt] = facing;
  facing >>= 2;


  // Location data from arguments
  newrand = RANDIE;
  x += turntosin[(facing+8192+TRIGTABLE_SHIFT) & TRIGTABLE_MASK] * (pipxyspacingbase[local_pip] + (newrand & pipxyspacingrand[local_pip]));
  y += turntosin[(facing+8192) & TRIGTABLE_MASK] * (pipxyspacingbase[local_pip] + (newrand & pipxyspacingrand[local_pip]));
  if (x < 0)  x = 0;
  if (x > meshedgex - 2)  x = meshedgex - 2;
  if (y < 0)  y = 0;
  if (y > meshedgey - 2)  y = meshedgey - 2;
  prtxpos[cnt] = x;
  prtypos[cnt] = y;
  prtzpos[cnt] = z;


  // Velocity data
  xvel = turntosin[(facing+8192+TRIGTABLE_SHIFT) & TRIGTABLE_MASK] * velocity;
  yvel = turntosin[(facing+8192) & TRIGTABLE_MASK] * velocity;
  newrand = RANDIE;
  zvel += pipzvelbase[local_pip] + (newrand & pipzvelrand[local_pip]) - (pipzvelrand[local_pip] >> 1);
  prtxvel[cnt] = xvel;
  prtyvel[cnt] = yvel;
  prtzvel[cnt] = zvel;


  // Template values
  prtbumpsize[cnt] = pipbumpsize[local_pip];
  prtbumpsizebig[cnt] = prtbumpsize[cnt] + (prtbumpsize[cnt] >> 1);
  prtbumpheight[cnt] = pipbumpheight[local_pip];
  prttype[cnt] = piptype[local_pip];


  // Image data
  newrand = RANDIE;
  prtrotate[cnt] = (newrand & piprotaterand[local_pip]) + piprotatebase[local_pip];
  prtrotateadd[cnt] = piprotateadd[local_pip];
  prtsize[cnt] = pipsizebase[local_pip];
  prtsizeadd[cnt] = pipsizeadd[local_pip];
  prtimage[cnt] = 0;
  newrand = RANDIE;
  prtimageadd[cnt] = pipimageadd[local_pip] + (newrand & pipimageaddrand[local_pip]);
  prtimagestt[cnt] = pipimagebase[local_pip] << 8;
  prtimagemax[cnt] = pipnumframes[local_pip] << 8;
  prttime[cnt] = piptime[local_pip];
  if (pipendlastframe[local_pip])
  {
    if(prtimageadd[cnt] != 0) prttime[cnt] = 0.0;
  }


  // Set onwhichfan...
  prtonwhichfan[cnt] = mesh_get_fan(prtxpos[cnt], prtypos[cnt]);


  // Damage stuff
  prtdamagebase[cnt] = pipdamagebase[local_pip];
  prtdamagerand[cnt] = pipdamagerand[local_pip];


  // Spawning data
  prtspawntime[cnt] = pipcontspawntime[local_pip];
  if (prtspawntime[cnt] != 0)
  {
    prtspawntime[cnt] = 1;
    if (prtattachedtocharacter[cnt] != MAXCHR)
    {
      prtspawntime[cnt]++; // Because attachment takes an update before it happens
    }
  }


  // Sound effect
  play_particle_sound(cnt, pipsoundspawn[local_pip]);

  return cnt;
}

//--------------------------------------------------------------------------------------------
Uint8 __prthitawall(Uint16 particle)
{
  // ZZ> This function returns nonzero if the particle hit a wall
  Uint32 fan;


  fan = mesh_get_fan(prtxpos[particle], prtypos[particle]);
  if(OFFEDGE==fan) return 0;

  // !!!BAD!!! Should really do bound checking...
  if (pipbumpmoney[prtpip[particle]])
  {
    return meshfx[fan] & (MESHFXIMPASS|MESHFXWALL);
  }
  else
  {
    return meshfx[fan]&(MESHFXIMPASS);
  }
}

//--------------------------------------------------------------------------------------------
void disaffirm_attached_particles(Uint16 character)
{
  // ZZ> This function makes sure a character has no attached particles
  Uint16 particle;

  particle = 0;
  while (particle < MAXPRT)
  {
    if (prton[particle] && prtattachedtocharacter[particle] == character)
    {
      free_one_particle(particle);
    }
    particle++;
  }

  // Set the alert for disaffirmation ( wet torch )
  chralert[character] |= ALERTIFDISAFFIRMED;
}

//--------------------------------------------------------------------------------------------
Uint16 number_of_attached_particles(Uint16 character)
{
  // ZZ> This function returns the number of particles attached to the given character
  Uint16 cnt, particle;

  cnt = 0;
  particle = 0;
  while (particle < MAXPRT)
  {
    if (prton[particle] && prtattachedtocharacter[particle] == character)
    {
      cnt++;
    }
    particle++;
  }

  return cnt;
}

//--------------------------------------------------------------------------------------------
void reaffirm_attached_particles(Uint16 character)
{
  // ZZ> This function makes sure a character has all of it's particles
  Uint16 numberattached;
  Uint16 particle;

  numberattached = number_of_attached_particles(character);
  while (numberattached < capattachedprtamount[chrmodel[character]])
  {
    particle = spawn_one_particle(chrxpos[character], chrypos[character], chrzpos[character], 0, chrmodel[character], capattachedprttype[chrmodel[character]], character, SPAWNLAST + numberattached, chrteam[character], character, numberattached, MAXCHR);
    if (particle != MAXPRT)
    {
      attach_particle_to_character(particle, character, prtgrip[particle]);
    }
    numberattached++;
  }

  // Set the alert for reaffirmation ( for exploding barrels with fire )
  chralert[character] |= ALERTIFREAFFIRMED;
}


//--------------------------------------------------------------------------------------------
void move_particles(float dUpdate)
{
  // ZZ> This is the particle physics function
  int cnt, tnc;
  Uint32 fan;
  Uint16 facing, pip, particle;
  float level;

  float loc_noslipfriction, loc_homingfriction, loc_homingaccel;

  loc_noslipfriction = pow(noslipfriction, dUpdate);

  // actually destroy all particles that requested destruction last time through the loop
  for(cnt = 0; cnt < MAXPRT; cnt++)
  {
    if(!prton[cnt] || !prtgopoof[cnt]) continue;

    // To make it easier
    pip = prtpip[cnt];
    facing = prtfacing[cnt];
    tnc = 0;
    while (tnc < pipendspawnamount[pip])
    {
      spawn_one_particle(prtxpos[cnt], prtypos[cnt], prtzpos[cnt],
        facing, prtmodel[cnt], pipendspawnpip[pip],
        MAXCHR, SPAWNLAST, prtteam[cnt], prtchr[cnt], tnc, prttarget[cnt]);
      facing += pipendspawnfacingadd[pip];
      tnc++;
    }
    free_one_particle(cnt);
  }

  for(cnt = 0; cnt < MAXPRT; cnt++)
  {
    if (!prton[cnt]) continue;

    prtonwhichfan[cnt] = OFFEDGE;
    prtlevel[cnt] = 0;
    fan = mesh_get_fan(prtxpos[cnt], prtypos[cnt]);
    prtonwhichfan[cnt] = fan;
    prtlevel[cnt] = (OFFEDGE==fan) ? 0 : get_level(prtxpos[cnt], prtypos[cnt], fan, bfalse);

    // To make it easier
    pip = prtpip[cnt];

    loc_homingfriction = pow(piphomingfriction[pip], dUpdate);
    loc_homingaccel    = piphomingaccel[pip];

    // Animate particle
    prtimage[cnt] += prtimageadd[cnt] * dUpdate;
    if (prtimage[cnt] >= prtimagemax[cnt])
    {
      if(pipendlastframe[pip])
      {
        prtimage[cnt]  = prtimagemax[cnt] - (1<<8);
        prtgopoof[cnt] = btrue;
      }
      else
      {
        prtimage[cnt] = 0;
      }
    };

    prtrotate[cnt] += prtrotateadd[cnt] * dUpdate;
    prtsize[cnt] = (prtsize[cnt] + prtsizeadd[cnt] < 0) ? 0 : prtsize[cnt] + prtsizeadd[cnt];

    // Change dyna light values
    prtdynalightlevel[cnt]   += pipdynalightleveladd[pip] * dUpdate;
    prtdynalightfalloff[cnt] += pipdynalightfalloffadd[pip] * dUpdate;


    // Make it sit on the floor...  Shift is there to correct for sprite size
    level = prtlevel[cnt] + ((float)prtsize[cnt] / (float)(1 << 9));


    if (prtattachedtocharacter[cnt] == MAXCHR)
    {
      prtzpos[cnt] += prtzvel[cnt] * dUpdate;

      // Check floor collision and do iterative physics
      if ( prtzpos[cnt] < level )
      {
        prtzpos[cnt]  = level + 0.001f;
        prtxvel[cnt] *= loc_noslipfriction;
        prtyvel[cnt] *= loc_noslipfriction;

        if (pipendground[pip]) prtgopoof[cnt] = btrue;

        if ( prtzvel[cnt] < -STOPBOUNCINGPART)
        {
          // Make it bounce
          prtzvel[cnt] *= -pipdampen[pip]*pipdampen[pip];

          // Play the sound for hitting the floor [FSND]
          play_particle_sound(cnt, pipsoundfloor[pip]);
        }
        else if(prtzvel[cnt] < STOPBOUNCINGPART)
        {
          prtzvel[cnt] = 0.0f;
        };

      }
      else
      {
        float lerp = (prtzpos[cnt] - (prtlevel[cnt] + PLATTOLERANCE)) / (float)PLATTOLERANCE;
        if(lerp < 0.2f) lerp = 0.2f;
        if(lerp > 1.0f) lerp = 1.0f;

        prtzvel[cnt] += gravity * lerp * dUpdate;

        prtxvel[cnt] -= (1.0f-noslipfriction)*lerp*dUpdate * prtxvel[cnt];
        prtyvel[cnt] -= (1.0f-noslipfriction)*lerp*dUpdate * prtyvel[cnt];
      }

      prtxpos[cnt] += prtxvel[cnt] * dUpdate;
      if(0 != __prthitawall(cnt))
      {
        // Play the sound for hitting a wall [WSND]
        play_particle_sound(cnt, pipsoundwall[pip]);
        prtxpos[cnt] -= prtxvel[cnt] * dUpdate;
        prtxvel[cnt] *= -pipdampen[pip];
        if (pipendwall[pip])
        {
          prtgopoof[cnt] = btrue;
        }
        else
        {
          // Change facing
          facing = prtfacing[cnt];
          if (facing < 32768)
          {
            facing -= NORTH;
            facing = ~facing;
            facing += NORTH;
          }
          else
          {
            facing -= SOUTH;
            facing = ~facing;
            facing += SOUTH;
          }
          prtfacing[cnt] = facing;
        }
      }

      prtypos[cnt] += prtyvel[cnt] * dUpdate;
      if(0 != __prthitawall(cnt))
      {
        prtypos[cnt] -= prtyvel[cnt] * dUpdate;
        prtyvel[cnt] *= -pipdampen[pip];
        if (pipendwall[pip])
        {
          prtgopoof[cnt] = btrue;
        }
        else
        {
          // Change facing
          facing = prtfacing[cnt];
          if (facing < 16384 || facing > 49152)
          {
            facing = ~facing;
          }
          else
          {
            facing -= EAST;
            facing = ~facing;
            facing += EAST;
          }
          prtfacing[cnt] = facing;
        }
      }
    }

    // Do speed limit on Z
    if (prtzvel[cnt] < -pipspdlimit[pip])  prtzvel[cnt] = -pipspdlimit[pip];

    // Do homing
    if (piphoming[pip] && prttarget[cnt] != MAXCHR)
    {
      if (!chralive[prttarget[cnt]])
      {
        prtgopoof[cnt] = btrue;
      }
      else
      {
        if (prtattachedtocharacter[cnt] == MAXCHR)
        {
          prtxvel[cnt] *= loc_homingfriction;
          prtyvel[cnt] *= loc_homingfriction;
          prtzvel[cnt] *= loc_homingfriction;

          prtxvel[cnt] += (chrxpos[prttarget[cnt]] - prtxpos[cnt]) * loc_homingaccel * dUpdate * 4.0f;
          prtyvel[cnt] += (chrypos[prttarget[cnt]] - prtypos[cnt]) * loc_homingaccel * dUpdate * 4.0f;
          prtzvel[cnt] += (chrzpos[prttarget[cnt]] + (chrbumpheight[prttarget[cnt]] >> 1) - prtzpos[cnt]) * loc_homingaccel * dUpdate * 4.0f;
        }
      }
    }


    // Spawn new particles if continually spawning
    if (pipcontspawnamount[pip] > 0.0f)
    {
      prtspawntime[cnt] -= dUpdate;
      if (prtspawntime[cnt] <= 0.0f)
      {
        prtspawntime[cnt] = pipcontspawntime[pip];
        facing = prtfacing[cnt];
        tnc = 0;
        while (tnc < pipcontspawnamount[pip])
        {
          particle = spawn_one_particle(prtxpos[cnt], prtypos[cnt], prtzpos[cnt],
            facing, prtmodel[cnt], pipcontspawnpip[pip],
            MAXCHR, SPAWNLAST, prtteam[cnt], prtchr[cnt], tnc, prttarget[cnt]);
          if (pipfacingadd[prtpip[cnt]] != 0 && particle < MAXPRT)
          {
            // Hack to fix velocity
            prtxvel[particle] += prtxvel[cnt];
            prtyvel[particle] += prtyvel[cnt];
          }
          facing += pipcontspawnfacingadd[pip];
          tnc++;
        }
      }
    }


    // Check underwater
    if (prtzpos[cnt] < waterdouselevel && HAS_BITS(meshfx[prtonwhichfan[cnt]],MESHFXWATER) && pipendwater[pip])
    {
      // Splash for particles is just a ripple
      spawn_one_particle(prtxpos[cnt], prtypos[cnt], watersurfacelevel,
        0, MAXMODEL, PRTPIP_RIPPLE, MAXCHR, SPAWNLAST, NULLTEAM, MAXCHR, 0, MAXCHR);


      // Check for disaffirming character
      if (prtattachedtocharacter[cnt] != MAXCHR && prtchr[cnt] == prtattachedtocharacter[cnt])
      {
        // Disaffirm the whole character
        disaffirm_attached_particles(prtattachedtocharacter[cnt]);
      }
      else
      {
        prtgopoof[cnt] = btrue;
      }
    }

    // Down the particle timer
    if (prttime[cnt] > 0.0f)
    {
      prttime[cnt] -= dUpdate;
      if(prttime[cnt]<=0.0f) prtgopoof[cnt] = btrue;
    };

    prtfacing[cnt] += pipfacingadd[pip] * dUpdate;
  }

}

//--------------------------------------------------------------------------------------------
void attach_particles()
{
  // ZZ> This function attaches particles to their characters so everything gets
  //     drawn right
  int cnt;

  cnt = 0;
  while (cnt < MAXPRT)
  {
    if (prton[cnt] && prtattachedtocharacter[cnt] != MAXCHR)
    {
      attach_particle_to_character(cnt, prtattachedtocharacter[cnt], prtgrip[cnt]);
      // Correct facing so swords knock characters in the right direction...
      if (pipdamfx[prtpip[cnt]]&DAMFXTURN)
        prtfacing[cnt] = chrturnleftright[prtattachedtocharacter[cnt]];
    }
    cnt++;
  }
}

//--------------------------------------------------------------------------------------------
void free_all_particles()
{
  // ZZ> This function resets the particle allocation lists
  numfreeprt = 0;
  while (numfreeprt < MAXPRT)
  {
    prton[numfreeprt] = bfalse;
    freeprtlist[numfreeprt] = numfreeprt;
    numfreeprt++;
  }
}

//--------------------------------------------------------------------------------------------
void setup_particles()
{
  // ZZ> This function sets up particle data

  particletexture = 0;

  // Reset the allocation table
  free_all_particles();
}

//--------------------------------------------------------------------------------------------
Uint16 terp_dir(Uint16 majordir, float dx, float dy, float dUpdate)
{
  // ZZ> This function returns a direction between the major and minor ones, closer
  //     to the major.

  Uint16 rotate_sin, minordir;
  Sint16 diff_dir;
  const float turnspeed = 2000.0f;

  if(ABS(dx) + ABS(dy) > TURNSPD*TURNSPD)
  {
    minordir = vec_to_turn(dx,dy);

    diff_dir = ((Sint16)minordir) - ((Sint16)majordir);
    
    if( (diff_dir > -turnspeed * dUpdate) && (diff_dir < turnspeed * dUpdate))
    {
      rotate_sin = (Uint16)diff_dir;
    }
    else
    {
      rotate_sin = (Uint16)(turnspeed * dUpdate * SGN(diff_dir));
    };

    return majordir + rotate_sin;
  }
  else
    return majordir;
}

//--------------------------------------------------------------------------------------------
//Uint16 terp_dir_fast(Uint16 majordir, float dx, float dy, float dUpdate)
//{
//  // ZZ> This function returns a direction between the major and minor ones, closer
//  //     to the major, but not by much.  Makes turning faster.
//  //float fkeep = pow(0.9f, dUpdate);
//  Uint16 rotate_sin, minordir;
//  Sint16 diff_dir;
//
//  if(ABS(dx) + ABS(dy) > TURNSPD*TURNSPD)
//  {
//    minordir = vec_to_turn(dx,dy);
//
//    diff_dir = ((Sint16)minordir) - ((Sint16)majordir);
//    
//    if( (diff_dir > -1000 * dUpdate) && (diff_dir < 1000 * dUpdate))
//    {
//      rotate_sin = (Uint16)diff_dir;
//    }
//    else
//    {
//      rotate_sin = (Uint16)(1000 * dUpdate * SGN(diff_dir));
//    };
//
//    return majordir + rotate_sin;
//  }
//  else
//    return majordir;
//}

////--------------------------------------------------------------------------------------------
//Uint16 terp_dir(Uint16 majordir, float dx, float dy, float dUpdate)
//{
//  // ZZ> This function returns a direction between the major and minor ones, closer
//  //     to the major.
//  float maj_dx, maj_dy, tmp_dx, tmp_dy;
//  //float fkeep = pow(0.9f, dUpdate);
//  float len, dotprod, crossprod, ang_diff;
//  Uint16 rotate_sin, rotate_cos;
//
//  len = sqrt(dx*dx + dy*dy);
//  if(len > TURNSPD)
//  {
//    turn_to_vec(majordir, &maj_dx, &maj_dy);
//    dotprod   = (maj_dx*dx + maj_dy*dy); //  / len;
//    crossprod = (maj_dx*dy - maj_dy*dx); //  / len;
//
//    ang_diff = atan2(crossprod, dotprod) * RAD_TO_SHORT;
//    
//    if( (ang_diff > -1000 * dUpdate) && (ang_diff < 1000 * dUpdate))
//    {
//      rotate_sin = ang_diff;
//    }
//    else
//    {
//      rotate_sin = 1000 * dUpdate * SGN(ang_diff);
//    };
//
//    rotate_cos = rotate_sin + 8192;
//
//    tmp_dx = turntosin[rotate_cos>>2] * maj_dx + turntosin[rotate_sin>>2] * maj_dy;
//    tmp_dy =-turntosin[rotate_sin>>2] * maj_dx + turntosin[rotate_cos>>2] * maj_dy;
//
//    return vec_to_turn(tmp_dx, tmp_dy);
//  }
//  else
//    return majordir;
//}
//
////--------------------------------------------------------------------------------------------
//Uint16 terp_dir_fast(Uint16 majordir, float dx, float dy, float dUpdate)
//{
//  // ZZ> This function returns a direction between the major and minor ones, closer
//  //     to the major, but not by much.  Makes turning faster.
//  float maj_dx, maj_dy, tmp_dx, tmp_dy;
//  //float fkeep = pow(0.9f, dUpdate);
//  float len, dotprod, crossprod, ang_diff;
//  Uint16 rotate_sin, rotate_cos;
//
//  len = sqrt(dx*dx + dy*dy);
//  if(len > TURNSPD)
//  {
//    turn_to_vec(majordir, &maj_dx, &maj_dy);
//    dotprod   = (maj_dx*dx + maj_dy*dy); //  / len;
//    crossprod = (maj_dx*dy - maj_dy*dx); //  / len;
//
//    ang_diff = atan2(crossprod, dotprod) * RAD_TO_SHORT;
//    
//    if( (ang_diff > -1000 * dUpdate) && (ang_diff < 1000 * dUpdate))
//    {
//      rotate_sin = ang_diff;
//    }
//    else
//    {
//      rotate_sin = 1000 * dUpdate * SGN(ang_diff);
//    };
//
//    rotate_cos = rotate_sin + 8192;
//
//    tmp_dx = turntosin[rotate_cos>>2] * maj_dx + turntosin[rotate_sin>>2] * maj_dy;
//    tmp_dy =-turntosin[rotate_sin>>2] * maj_dx + turntosin[rotate_cos>>2] * maj_dy;
//
//    return vec_to_turn(tmp_dx, tmp_dy);
//  }
//  else
//    return majordir;
//}

//--------------------------------------------------------------------------------------------
void spawn_bump_particles(Uint16 character, Uint16 particle)
{
  // ZZ> This function is for catching characters on fire and such
  int cnt;
  Sint16 x, y, z;
  int distance, bestdistance;
  Uint16 frame;
  Uint16 facing, bestvertex;
  Uint16 amount;
  Uint16 pip;
  Uint16 vertices;
  Uint16 direction, left, right, model;
  float fsin, fcos;


  pip = prtpip[particle];
  amount = pipbumpspawnamount[pip];


  if (amount != 0 || pipspawnenchant[pip])
  {
    // Only damage if hitting from proper direction
    model = chrmodel[character];
    vertices = madvertices[model];
    direction = RAD_TO_TURN(atan2(prtyvel[particle], prtxvel[particle]));
    direction = chrturnleftright[character] - direction + 32768;
    if (madframefx[chrframe[character]]&MADFXINVICTUS)
    {
      // I Frame
      if (pipdamfx[pip]&DAMFXBLOC)
      {
        left = 65535;
        right = 0;
      }
      else
      {
        direction -= capiframefacing[model];
        left = (~capiframeangle[model]);
        right = capiframeangle[model];
      }
    }
    else
    {
      // N Frame
      direction -= capnframefacing[model];
      left = (~capnframeangle[model]);
      right = capnframeangle[model];
    }
    // Check that direction
    if (direction <= left && direction >= right)
    {
      // Spawn new enchantments
      if (pipspawnenchant[pip])
      {
        spawn_enchant(prtchr[particle], character, MAXCHR, MAXENCHANT, prtmodel[particle]);
      }
      // Spawn particles
      if (amount != 0 && !capresistbumpspawn[chrmodel[character]] && !chrinvictus[character] && vertices != 0 && (chrdamagemodifier[character][prtdamagetype[particle]]&DAMAGESHIFT) < 3)
      {
        if (amount == 1)
        {
          // A single particle ( arrow? ) has been stuck in the character...
          // Find best vertex to attach to
          bestvertex = 0;
          bestdistance = 9999999;
          z = -chrzpos[character] + prtzpos[particle] + RAISE;
          facing = prtfacing[particle] - chrturnleftright[character] - 16384;
          facing >>= 2;
          fsin = turntosin[facing & TRIGTABLE_MASK];
          fcos = turntosin[(facing+TRIGTABLE_SHIFT) & TRIGTABLE_MASK];
          y = 8192;
          x = -y * fsin;
          y = y * fcos;
          z <<= 10;///chrscale[character];
          frame = madframestart[chrmodel[character]];
          cnt = 0;
          while (cnt < vertices)
          {
            distance = ABS(x - madvrtx[frame][vertices-cnt-1]) + ABS(y - madvrty[frame][vertices-cnt-1]) + (ABS(z - madvrtz[frame][vertices-cnt-1]));
            if (distance < bestdistance)
            {
              bestdistance = distance;
              bestvertex = cnt;
            }
            cnt++;
          }
          spawn_one_particle(chrxpos[character], chrypos[character], chrzpos[character], 0, prtmodel[particle], pipbumpspawnpip[pip],
                             character, bestvertex + 1, prtteam[particle], prtchr[particle], cnt, character);
        }
        else
        {
          amount = (amount * vertices) >> 5;  // Correct amount for size of character
          cnt = 0;
          while (cnt < amount)
          {
            spawn_one_particle(chrxpos[character], chrypos[character], chrzpos[character], 0, prtmodel[particle], pipbumpspawnpip[pip],
                               character, rand() % vertices, prtteam[particle], prtchr[particle], cnt, character);
            cnt++;
          }
        }
      }
    }
  }
}

//--------------------------------------------------------------------------------------------
bool_t prt_is_over_water(int cnt)
{
  // This function returns btrue if the particle is over a water tile

  Uint32 fan;

  if (cnt < MAXPRT)
  {
    fan = mesh_get_fan(prtxpos[cnt], prtypos[cnt]);
    if (HAS_BITS(meshfx[fan],MESHFXWATER))  return (OFFEDGE != fan);
  }

  return bfalse;
}

//--------------------------------------------------------------------------------------------
void do_weather_spawn(float dUpdate)
{
  // ZZ> This function drops snowflakes or rain or whatever, also swings the camera
  Uint16 particle;
  int cnt;
  float x, y, z;
  bool_t foundone = bfalse;

  if(weathertime > 0)
  {
  weathertime -= dUpdate;
  if (weathertime < 0) weathertime = 0;

  if (weathertime == 0)
  {
    weathertime = weathertimereset;

    // Find a valid player
    foundone = bfalse;
    cnt = 0;
    while (cnt < MAXPLAYER)
    {
      weatherplayer = (weatherplayer + 1) & (MAXPLAYER - 1);
      if (plavalid[weatherplayer])
      {
        foundone = btrue;
        cnt = MAXPLAYER;
      }
      cnt++;
    }
   }


    // Did we find one?
    if (foundone)
    {
      // Yes, but is the character valid?
      cnt = plaindex[weatherplayer];
      if (chron[cnt] && !chrinpack[cnt])
      {
        // Yes, so spawn over that character
        x = chrxpos[cnt];
        y = chrypos[cnt];
        z = chrzpos[cnt];
        particle = spawn_one_particle(x, y, z, 0, MAXMODEL, PRTPIP_WEATHER_1, MAXCHR, SPAWNLAST, NULLTEAM, MAXCHR, 0, MAXCHR);
        if (weatheroverwater && particle != MAXPRT)
        {
          if (!prt_is_over_water(particle))
          {
            free_one_particle_no_sound(particle);
          }
        }

      }
    }

  }
  camswing = (camswing + camswingrate) & TRIGTABLE_MASK;
}

//--------------------------------------------------------------------------------------------
Uint32 load_one_particle(char *szLoadName, int object, int pip)
{
  // ZZ> This function loads a particle template, returning MAXPRTPIP if the file wasn't
  //     found
  FILE* fileread;
  IDSZ idsz;
  int iTmp;
  float fTmp;
  char cTmp;
  Uint32 retval = MAXPRTPIP;



  fileread = fs_fileOpen(PRI_NONE, NULL, szLoadName, "r");
  if (NULL == fileread) return retval;

  retval = numpip;
  numpip++;


  // General data
  globalname = szLoadName;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  pipforce[retval] = bfalse;
  if (cTmp == 'T' || cTmp == 't')  pipforce[retval] = btrue;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  if (cTmp == 'L' || cTmp == 'l')  piptype[retval] = PRTLIGHTSPRITE;
  if (cTmp == 'S' || cTmp == 's')  piptype[retval] = PRTSOLIDSPRITE;
  if (cTmp == 'T' || cTmp == 't')  piptype[retval] = PRTALPHASPRITE;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipimagebase[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipnumframes[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipimageadd[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipimageaddrand[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); piprotatebase[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); piprotaterand[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); piprotateadd[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipsizebase[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipsizeadd[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%f", &fTmp); pipspdlimit[retval] = fTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipfacingadd[retval] = iTmp;


  // Ending conditions
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  pipendwater[retval] = btrue;
  if (cTmp == 'F' || cTmp == 'f')  pipendwater[retval] = bfalse;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  pipendbump[retval] = btrue;
  if (cTmp == 'F' || cTmp == 'f')  pipendbump[retval] = bfalse;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  pipendground[retval] = btrue;
  if (cTmp == 'F' || cTmp == 'f')  pipendground[retval] = bfalse;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  pipendlastframe[retval] = btrue;
  if (cTmp == 'F' || cTmp == 'f')  pipendlastframe[retval] = bfalse;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); piptime[retval] = iTmp;


  // Collision data
  goto_colon(fileread);  fscanf(fileread, "%f", &fTmp); pipdampen[retval] = fTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipbumpmoney[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipbumpsize[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipbumpheight[retval] = iTmp;
  goto_colon(fileread);  read_pair(fileread);
  pipdamagebase[retval] = pairbase;
  pipdamagerand[retval] = pairrand;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  if (cTmp == 'S' || cTmp == 's') pipdamagetype[retval] = DAMAGESLASH;
  if (cTmp == 'C' || cTmp == 'c') pipdamagetype[retval] = DAMAGECRUSH;
  if (cTmp == 'P' || cTmp == 'p') pipdamagetype[retval] = DAMAGEPOKE;
  if (cTmp == 'H' || cTmp == 'h') pipdamagetype[retval] = DAMAGEHOLY;
  if (cTmp == 'E' || cTmp == 'e') pipdamagetype[retval] = DAMAGEEVIL;
  if (cTmp == 'F' || cTmp == 'f') pipdamagetype[retval] = DAMAGEFIRE;
  if (cTmp == 'I' || cTmp == 'i') pipdamagetype[retval] = DAMAGEICE;
  if (cTmp == 'Z' || cTmp == 'z') pipdamagetype[retval] = DAMAGEZAP;


  // Lighting data
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  pipdynalightmode[retval] = DYNAOFF;
  if (cTmp == 'T' || cTmp == 't') pipdynalightmode[retval] = DYNAON;
  if (cTmp == 'L' || cTmp == 'l') pipdynalightmode[retval] = DYNALOCAL;
  goto_colon(fileread);  fscanf(fileread, "%f", &fTmp); pipdynalevel[retval] = fTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipdynafalloff[retval] = iTmp;
  if (pipdynafalloff[retval] > MAXFALLOFF)  pipdynafalloff[retval] = MAXFALLOFF;



  // Initial spawning of this particle
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipfacingbase[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipfacingrand[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipxyspacingbase[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipxyspacingrand[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipzspacingbase[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipzspacingrand[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipxyvelbase[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipxyvelrand[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipzvelbase[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipzvelrand[retval] = iTmp;


  // Continuous spawning of other particles
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipcontspawntime[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipcontspawnamount[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipcontspawnfacingadd[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipcontspawnpip[retval] = iTmp;


  // End spawning of other particles
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipendspawnamount[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipendspawnfacingadd[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipendspawnpip[retval] = iTmp;

  // Bump spawning of attached particles
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipbumpspawnamount[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipbumpspawnpip[retval] = iTmp;


  // Random stuff  !!!BAD!!! Not complete
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipdazetime[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); pipgrogtime[retval] = iTmp;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  pipspawnenchant[retval] = bfalse;
  if (cTmp == 'T' || cTmp == 't') pipspawnenchant[retval] = btrue;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  pipcauseknockback[retval] = bfalse;
  if (cTmp == 'T' || cTmp == 't') pipcauseknockback[retval] = btrue;
  goto_colon(fileread);
  pipcausepancake[retval] = bfalse;
  if (cTmp == 'T' || cTmp == 't') pipcausepancake[retval] = btrue;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  pipneedtarget[retval] = bfalse;
  if (cTmp == 'T' || cTmp == 't') pipneedtarget[retval] = btrue;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  piptargetcaster[retval] = bfalse;
  if (cTmp == 'T' || cTmp == 't') piptargetcaster[retval] = btrue;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  pipstartontarget[retval] = bfalse;
  if (cTmp == 'T' || cTmp == 't') pipstartontarget[retval] = btrue;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  piponlydamagefriendly[retval] = bfalse;
  if (cTmp == 'T' || cTmp == 't') piponlydamagefriendly[retval] = btrue;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp);
  if (iTmp < -1) iTmp = -1;
  if (iTmp > MAXWAVE - 1) iTmp = MAXWAVE - 1;
  pipsoundspawn[retval] = iTmp;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp);
  if (iTmp < -1) iTmp = -1;
  if (iTmp > MAXWAVE - 1) iTmp = MAXWAVE - 1;
  pipsoundend[retval] = iTmp;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  pipfriendlyfire[retval] = bfalse;
  if (cTmp == 'T' || cTmp == 't') pipfriendlyfire[retval] = btrue;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  piphateonly[retval] = bfalse;
  if (cTmp == 'T' || cTmp == 't') piphateonly[retval] = btrue;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  pipnewtargetonspawn[retval] = bfalse;
  if (cTmp == 'T' || cTmp == 't') pipnewtargetonspawn[retval] = btrue;
  goto_colon(fileread);  fscanf(fileread, "%d", &iTmp); piptargetangle[retval] = iTmp >> 1;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  piphoming[retval] = bfalse;
  if (cTmp == 'T' || cTmp == 't') piphoming[retval] = btrue;
  goto_colon(fileread);  fscanf(fileread, "%f", &fTmp); piphomingfriction[retval] = fTmp;
  goto_colon(fileread);  fscanf(fileread, "%f", &fTmp); piphomingaccel[retval] = fTmp;
  goto_colon(fileread);  cTmp = get_first_letter(fileread);
  piprotatetoface[retval] = bfalse;
  if (cTmp == 'T' || cTmp == 't') piprotatetoface[retval] = btrue;
  goto_colon(fileread);		//BAD! Not used
  goto_colon(fileread);  fscanf(fileread, "%f", &fTmp); pipmanadrain[numpip] = fTmp*256;		//Mana drain (Mana damage)



  // Clear expansions...
  pipzaimspd[retval] = 0;
  pipsoundfloor[retval] = -1;
  pipsoundwall[retval] = -1;
  pipendwall[retval] = pipendground[retval];
  pipdamfx[retval] = DAMFXTURN;
  if (piphoming[retval])  pipdamfx[retval] = DAMFXNONE;
  pipallowpush[retval] = btrue;
  pipdynalightfalloffadd[retval] = 0;
  pipdynalightleveladd[retval] = 0;
  pipintdamagebonus[retval] = bfalse;
  pipwisdamagebonus[retval] = bfalse;
  // Read expansions
  while (goto_colon_yesno(fileread))
  {
    idsz = get_idsz(fileread);
    fscanf(fileread, "%c%d", &cTmp, &iTmp);

    if (MAKE_IDSZ("TURN") == idsz)  pipdamfx[retval] = DAMFXNONE;
    if (MAKE_IDSZ("ZSPD") == idsz)  pipzaimspd[retval] = iTmp;
    if (MAKE_IDSZ("FSND") == idsz)  pipsoundfloor[retval] = iTmp;
    if (MAKE_IDSZ("WSND") == idsz)  pipsoundwall[retval] = iTmp;
    if (MAKE_IDSZ("WEND") == idsz)  pipendwall[retval] = iTmp;
    if (MAKE_IDSZ("ARMO") == idsz)  pipdamfx[retval] |= DAMFXARMO;
    if (MAKE_IDSZ("BLOC") == idsz)  pipdamfx[retval] |= DAMFXBLOC;
    if (MAKE_IDSZ("ARRO") == idsz)  pipdamfx[retval] |= DAMFXARRO;
    if (MAKE_IDSZ("TIME") == idsz)  pipdamfx[retval] |= DAMFXTIME;
    if (MAKE_IDSZ("PUSH") == idsz)  pipallowpush[retval] = iTmp;
    if (MAKE_IDSZ("DLEV") == idsz)  pipdynalightleveladd[retval] = iTmp / 1000.0;
    if (MAKE_IDSZ("DRAD") == idsz)  pipdynalightfalloffadd[retval] = iTmp / 1000.0;
    if (MAKE_IDSZ("IDAM") == idsz)  pipintdamagebonus[retval] = iTmp;
    if (MAKE_IDSZ("WDAM") == idsz)  pipwisdamagebonus[retval] = iTmp;
  }


  // Make sure it's referenced properly
  madprtpip[object][pip] = retval;

  fs_fileClose(fileread);

  return retval;
}

//--------------------------------------------------------------------------------------------
void reset_particles(char* modname)
{
  // ZZ> This resets all particle data and reads in the coin and water particles
  int cnt, object;

  // Load in the standard global particles ( the coins for example )
  //BAD! This should only be needed once at the start of the game
  numpip = 0;

  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s/%s", CData.basicdat_dir, CData.globalparticles_dir, CData.money1_file);
  if (MAXPRTPIP == load_one_particle(CStringTmp1, 0, 0))
  {
    log_error("Data file was not found! (%s)\n", CStringTmp1);
  }

  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s/%s", CData.basicdat_dir, CData.globalparticles_dir, CData.money5_file);
  if (MAXPRTPIP == load_one_particle(CStringTmp1, 0, 0))
  {
    log_error("Data file was not found! (%s)\n", CStringTmp1);
  }

  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s/%s", CData.basicdat_dir, CData.globalparticles_dir, CData.money25_file);
  if (MAXPRTPIP == load_one_particle(CStringTmp1, 0, 0))
  {
    log_error("Data file was not found! (%s)\n", CStringTmp1);
  }

  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s/%s", CData.basicdat_dir, CData.globalparticles_dir, CData.money100_file);
  if (MAXPRTPIP == load_one_particle(CStringTmp1, 0, 0))
  {
    log_error("Data file was not found! (%s)\n", CStringTmp1);
  }

  //Load module specific information
  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s%s/%s", modname, CData.gamedat_dir, CData.weather4_file);
  if (MAXPRTPIP == load_one_particle(CStringTmp1, 0, 0))
  {
    log_error("Data file was not found! (%s)\n", CStringTmp1);
  }

  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s%s/%s", modname, CData.gamedat_dir, CData.weather5_file);
  if (MAXPRTPIP == load_one_particle(CStringTmp1, 0, 0))
  {
    log_error("Data file was not found! (%s)\n", CStringTmp1);
  }

  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s%s/%s", modname, CData.gamedat_dir, CData.splash_file);
  if (MAXPRTPIP == load_one_particle(CStringTmp1, 0, 0))
  {
    log_error("Data file was not found! (%s)\n", CStringTmp1);
  }

  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s%s/%s", modname, CData.gamedat_dir, CData.ripple_file);
  if (MAXPRTPIP == load_one_particle(CStringTmp1, 0, 0))
  {
    log_error("Data file was not found! (%s)\n", CStringTmp1);
  }

  //This is also global...
  snprintf(CStringTmp1, sizeof(CStringTmp1), "%s/%s/%s", CData.basicdat_dir, CData.globalparticles_dir, CData.defend_file);
  if (MAXPRTPIP == load_one_particle(CStringTmp1, 0, 0))
  {
    log_error("Data file was not found! (%s)\n", CStringTmp1);
  }


  // Now clear out the local pips
  
  for (object = 0; object < MAXMODEL; object++)
  {
    for (cnt = 0; cnt < PRTPIP_PEROBJECT_COUNT; cnt++)
    {
      madprtpip[object][cnt] = MAXPRTPIP;
    }  
  }
}

