/* Egoboo - proto.h
 * Function prototypes for a huge portion of the game code.
 */

/*
    This file is part of Egoboo.

    Egoboo is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Egoboo is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Egoboo.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _PROTO_H_
#define _PROTO_H_

#include "egobootypedef.h"
#include <SDL_mixer.h> // for Mix_* stuff.
#include <SDL_opengl.h>

typedef enum control_t CONTROL;

void load_graphics();

void empty_import_directory(void);
void insert_space(int position);
void copy_one_line(int write);
int load_one_line(int read);
int load_parsed_line(int read);
void surround_space(int position);
void parse_null_terminate_comments();
int get_indentation();
void fix_operators();
int starts_with_capital_letter();
Uint32 get_high_bits();
int tell_code(int read);
void add_code(Uint32 highbits);
void parse_line_by_line();
Uint32 jump_goto(int index);
void parse_jumps(int ainumber);
void log_code(int ainumber, char* savename);
int ai_goto_colon(int read);
void get_code(int read);
void load_ai_codes(char* loadname);
Uint32 load_ai_script(char *loadname);
void reset_ai_script();
int what_action(char cTmp);
int get_level(Uint8 x, Uint8 y, Uint32 fan, Uint8 waterwalk);
void release_all_textures();
void load_one_icon(char *szLoadName);
void prime_titleimage();
void prime_icons();
void release_all_icons();
void release_all_titleimages();
void release_all_models();
void reset_sounds();
void release_map();
void release_module(void);
void close_session();
bool_t memory_cleanUp();
void make_newloadname(char *modname, char *appendname, char *newloadname);
void export_one_character(Uint16 character, Uint16 owner, int number);
void export_all_local_players(void);
void quit_module(void);
void quit_game(void);
void goto_colon(FILE* fileread);
Uint8 goto_colon_yesno(FILE* fileread);
char get_first_letter(FILE* fileread);
void reset_tags();
int read_tag(FILE *fileread);
void read_all_tags(char *szFilename);
int tag_value(char *string);
void read_controls(char *szFilename);
bool_t control_key_is_pressed(CONTROL control);
bool_t control_mouse_is_pressed(CONTROL control);
bool_t control_joya_is_pressed(CONTROL control);
bool_t control_joyb_is_pressed(CONTROL control);
void free_all_enchants();
char * undo_idsz(Uint32 idsz);
IDSZ get_idsz(FILE* fileread);
void load_one_enchant_type(char* szLoadName, Uint16 profile);
Uint16 get_free_enchant();
void unset_enchant_value(Uint16 enchantindex, Uint8 valueindex);
void remove_enchant_value(Uint16 enchantindex, Uint8 valueindex);
int get_free_message(void);
void display_message(int message, Uint16 character);
void remove_enchant(Uint16 enchantindex);
Uint16 enchant_value_filled(Uint16 enchantindex, Uint8 valueindex);
void set_enchant_value(Uint16 enchantindex, Uint8 valueindex,
                       Uint16 enchanttype);
void getadd(int MIN, int value, int MAX, int* valuetoadd);
void fgetadd(float MIN, float value, float MAX, float* valuetoadd);
void add_enchant_value(Uint16 enchantindex, Uint8 valueindex,
                       Uint16 enchanttype);
Uint16 spawn_enchant(Uint16 owner, Uint16 target,
                             Uint16 spawner, Uint16 enchantindex, Uint16 modeloptional);
void load_action_names(char* loadname);
void get_name(FILE* fileread, char *szName);
void read_setup(char* filename);
void log_madused(char *savename);
void make_lightdirectionlookup();
float spek_global_lighting(int rotation, int normal, float lx, float ly, float lz);
void make_spektable(float lx, float ly, float lz);
void make_lighttospek(void);
int vertexconnected(int modelindex, int vertex);
void get_madtransvertices(int modelindex);
int rip_md2_header(void);
void fix_md2_normals(Uint16 modelindex);
void rip_md2_commands(Uint16 modelindex);
int rip_md2_frame_name(int frame);
void rip_md2_frames(Uint16 modelindex);
int load_one_md2(char* szLoadname, Uint16 modelindex);

//Sound and music stuff
void load_all_music_sounds();
void stop_music();
void load_global_waves(char *modname);
bool_t sdlmixer_initialize();
int play_sound(float xpos, float ypos, Mix_Chunk *loadedwave);
void stop_sound(int whichchannel);
void play_music(int songnumber, int fadetime, int loops);

//Graphical Functions
void render_particles();
void render_particle_reflections();
void render_mad_lit(Uint16 character);
void render_water_fan_lit(Uint32 fan, Uint8 layer, Uint8 mode);

//void dump_matrix(GLMATRIX a);
/*inline GLMATRIX IdentityMatrix();
inline GLMATRIX ZeroMatrix(void);  // initializes matrix to zero
inline GLMATRIX MatrixMult(const GLMATRIX a, const GLMATRIX b);
GLMATRIX Translate(const float dx, const float dy, const float dz);
GLMATRIX RotateX(const float rads);
GLMATRIX RotateY(const float rads);
GLMATRIX RotateZ(const float rads);
GLMATRIX ScaleXYZ(const float sizex, const float sizey, const float sizez);
GLMATRIX ScaleXYZRotateXYZTranslate(const float sizex, const float sizey, const float sizez,
   Uint16 turnz, Uint16 turnx, Uint16 turny,
   float tx, float ty, float tz);
GLMATRIX FourPoints(float orix, float oriy, float oriz,
                     float widx, float widy, float widz,
                     float forx, float fory, float forz,
                     float upx,  float upy,  float upz,
                     float scale);
inline GLMATRIX ViewMatrix(const GLVECTOR from,      // camera location
                            const GLVECTOR at,        // camera look-at target
                            const GLVECTOR world_up,  // worlds up, usually 0, 0, 1
                            const float roll);          // clockwise roll around
                                                       //    viewing direction,
                                                       //    in radians
inline GLMATRIX ProjectionMatrix(const float near_plane,     // distance to near clipping plane
                                  const float far_plane,      // distance to far clipping plane
                                  const float fov);            // field of view angle, in radians
*/
int open_passage(int passage);
void check_passage_music();
int break_passage(int passage, Uint16 starttile, Uint16 frames,
                  Uint16 become, Uint8 meshfxor);
void flash_passage(int passage, Uint8 color);
Uint8  find_tile_in_passage(int passage, int tiletype);
Uint16 who_is_blocking_passage(int passage);
Uint16 who_is_blocking_passage_ID(int passage, IDSZ idsz);
int close_passage(int passage);
void clear_passages();
void add_shop_passage(int owner, int passage);
void add_passage(int tlx, int tly, int brx, int bry, Uint8 open, Uint8 mask);
void flash_character_height(Uint16 character, Uint8 valuelow, Sint16 low,
                            Uint8 valuehigh, Sint16 high);
void flash_character(Uint16 character, Uint8 value);
void flash_select(float dUpdate);
void add_to_dolist(int cnt);
void order_dolist(void);
void make_dolist(void);
void keep_weapons_with_holders();
void make_enviro(void);
void make_prtlist(void);
void make_turntosin(void);
void make_one_character_matrix(Uint16 cnt);
void free_one_particle_no_sound(Uint16 particle);
void play_particle_sound(Uint16 particle, Sint8 sound);
void free_one_particle(Uint16 particle);
void free_one_character(Uint16 character);
void free_inventory(Uint16 character);
void attach_particle_to_character(Uint16 particle, Uint16 character, int grip);
void make_one_weapon_matrix(Uint16 cnt);
void make_character_matrices();
int get_free_particle(int force);
int get_free_character();
void debug_message(int time, const char *format, ...);
void reset_end_text();
void append_end_text(int message, Uint16 character);
Uint16 spawn_one_particle(float x, float y, float z,
                                  Uint16 facing, Uint16 model, Uint16 pip,
                                  Uint16 characterattach, Uint16 grip, Uint8 team,
                                  Uint16 characterorigin, Uint16 multispawn, Uint16 oldtarget);
Uint8 __prthitawall(Uint16 particle);
void disaffirm_attached_particles(Uint16 character);
Uint16 number_of_attached_particles(Uint16 character);
void reaffirm_attached_particles(Uint16 character);
void do_enchant_spawn(float dUpdate);
void move_particles(float dUpdate);
void attach_particles();
void free_all_particles();
void free_all_characters();
void show_stat(Uint16 statindex);
void check_stats();
void check_screenshot();
bool_t dump_screenshot();
void add_stat(Uint16 character);
void move_to_top(Uint16 character);
void sort_stat();
void setup_particles();
Uint16 terp_dir(Uint16 majordir, float dx, float dy, float dUpdate);
Uint16 terp_dir_fast(Uint16 majordir, float dx, float dy, float dUpdate);
Uint8 __chrhitawall(Uint16 character);
void move_water(float dUpdate);
void play_action(Uint16 character, Uint16 action, Uint8 actionready);
void set_frame(Uint16 character, Uint16 frame, Uint8 lip);
void reset_character_alpha(Uint16 character);
void reset_character_accel(Uint16 character);
void detach_character_from_mount(Uint16 character, bool_t ignorekurse,
                                 bool_t doshop);
void spawn_bump_particles(Uint16 character, Uint16 particle);
int generate_number(int numbase, int numrand);
void drop_money(Uint16 character, Uint16 money);
void call_for_help(Uint16 character);
void give_experience(Uint16 character, int amount, Uint8 xptype);
void give_team_experience(Uint8 team, int amount, Uint8 xptype);
void disenchant_character(Uint16 cnt);
void damage_character(Uint16 character, Uint16 direction,
                      int damagebase, int damagerand, Uint8 damagetype, Uint8 team,
                      Uint16 attacker, Uint16 effects);
void kill_character(Uint16 character, Uint16 killer);
void spawn_poof(Uint16 character, Uint16 profile);
void naming_names(int profile);
void read_naming(int profile, char *szLoadname);
void prime_names(void);
void tilt_characters_to_terrain();
int spawn_one_character(float x, float y, float z, int profile, Uint8 team,
                        Uint8 skin, Uint16 facing, char *name, int override);
void respawn_character(Uint16 character);
Uint16 change_armor(Uint16 character, Uint16 skin);
void change_character(Uint16 cnt, Uint16 profile, Uint8 skin,
                      Uint8 leavewhich);
bool_t cost_mana(Uint16 character, int amount, Uint16 killer);
void switch_team(Uint16 character, Uint8 team);

void issue_clean(Uint16 character);
int restock_ammo(Uint16 character, IDSZ idsz);
void issue_order(Uint16 character, Uint32 order);
void issue_special_order(Uint32 order, IDSZ idsz);
void set_alerts(Uint16 character);
bool_t module_reference_matches(char *szLoadName, IDSZ idsz);
void add_module_idsz(char *szLoadName, IDSZ idsz);

bool_t add_quest_idsz(char *whichplayer, IDSZ idsz);
bool_t beat_quest_idsz(char *whichplayer, IDSZ idsz);
int check_player_quest(char *whichplayer, IDSZ idsz);

Uint8 run_function(Uint32 value, Uint16 character);
void set_operand(Uint8 variable);
void run_operand(Uint32 value, Uint16 character);
void let_character_think(Uint16 character);
void let_ai_think();
void attach_character_to_mount(Uint16 character, Uint16 mount,
                               Uint16 grip);
Uint16 stack_in_pack(Uint16 item, Uint16 character);
void add_item_to_character_pack(Uint16 item, Uint16 character);
Uint16 get_item_from_character_pack(Uint16 character, Uint16 grip, bool_t ignorekurse);
void drop_keys(Uint16 character);
void drop_all_items(Uint16 character);
void character_grab_stuff(Uint16 chara, int grip, bool_t people);
void character_swipe(Uint16 cnt, Uint8 grip);
void move_characters(float dUpdate);
void make_textureoffset(void);
int add_player(Uint16 character, Uint16 player, Uint8 device);
void clear_messages();
void setup_characters(char *modname);
void setup_passage(char *modname);
void setup_alliances(char *modname);
void load_mesh_fans();
void make_fanstart();
void make_twist();
bool_t load_mesh(char *modname);
void read_mouse();
void read_key();
void read_joystick();
void reset_press();
void read_input();
void check_add(Uint8 key, char bigletter, char littleletter);
void create_szfpstext(float fps, float ups);
void camera_look_at(float x, float y);
void project_view();
void make_renderlist();
void make_camera_matrix();
void figure_out_what_to_draw();
void bound_camera();
void bound_camtrack();
void set_one_player_latch(Uint16 player);
void set_local_latches(void);
void adjust_camera_angle(int height);
void move_camera(float dUpdate);
void make_onwhichfan(void);
void bump_characters(float dUpdate);
bool_t prt_is_over_water(int cnt);
void do_weather_spawn(float dUpdate);
void animate_tiles(float dUpdate);
void stat_return(float dUpdate);
void pit_kill(float dUpdate);
void reset_players();
int find_module(char *smallname);
void resize_characters(float dUpdate);
void update_game(float dUpdate);
void update_timers();
void load_basic_textures(char *modname);
Uint16 action_number();
Uint16 action_frame();
Uint16 test_frame_name(char letter);
void action_copy_correct(int object, Uint16 actiona, Uint16 actionb);
void get_walk_frame(int object, int lip, int action);
void get_framefx(int frame);
void make_framelip(int object, int action);
void get_actions(int object);
void read_pair(FILE* fileread);
void undo_pair(int base, int rand);
void ftruthf(FILE* filewrite, char* text, Uint8 truth);
void fdamagf(FILE* filewrite, char* text, Uint8 damagetype);
void factiof(FILE* filewrite, char* text, Uint8 action);
void fgendef(FILE* filewrite, char* text, Uint8 gender);
void fpairof(FILE* filewrite, char* text, int base, int rand);
void funderf(FILE* filewrite, char* text, char* usename);
void export_one_character_name(char *szSaveName, Uint16 character);
void export_one_character_profile(char *szSaveName, Uint16 character);
void export_one_character_skin(char *szSaveName, Uint16 character);
int load_one_character_profile(char *szLoadName);
Uint32 load_one_particle(char *szLoadName, int object, int pip);
void reset_particles(char* modname);
void make_mad_equally_lit(int model);
void get_message(FILE* fileread);
void load_all_messages(char *loadname, int object);
void check_copy(char* loadname, int object);
int load_one_object(int skin, char* tmploadname);
void load_all_objects(char *modname);
bool_t load_bars(char* szBitmap);
void load_map(char* szModule);
bool_t load_font(char* szBitmap, char* szSpacing);
void make_water();
void read_wawalite(char *modname);
void reset_teams();
void reset_messages();
void make_randie();
void load_module(char *smallname);
void render_prt();
void render_shadow(Uint16 character);
void render_bad_shadow(Uint16 character);
void render_refprt();
void render_fan(Uint32 fan, char tex_loaded);
void render_water_fan(Uint32 fan, Uint8 layer, Uint8 mode);
void render_enviromad(Uint16 character, Uint8 trans);
void render_texmad(Uint16 character, Uint8 trans);
void render_mad(Uint16 character, Uint8 trans);
void render_refmad(int tnc, Uint8 trans);
void light_characters();
void light_particles();
void set_fan_light(int fanx, int fany, Uint16 particle);
void do_dynalight();
void render_water();
void draw_scene_sadreflection();
void draw_scene_zreflection();
bool_t get_mesh_memory();
void draw_blip(Uint8 color, int x, int y);
void draw_one_icon(int icontype, int x, int y, Uint8 sparkle);
void draw_one_font(int fonttype, int x, int y);
void draw_map(int x, int y);
int draw_one_bar(int bartype, int x, int y, int ticks, int maxticks);
void draw_string(char *szText, int x, int y);
int length_of_word(char *szText);
int draw_wrap_string(char *szText, int x, int y, int maxx);
int draw_status(Uint16 character, int x, int y);
void draw_text();
bool_t request_pageflip();
bool_t do_pageflip();
bool_t do_clear();
void draw_scene();
void draw_main(float);

Uint32 mesh_get_fan(float x, float y);
Uint32 mesh_get_block(float x, float y);

bool_t prt_find_target_in_block(int block_x, int block_y, float chrx, float chry, Uint16 facing,
                            bool_t onlyfriends, bool_t anyone, Uint8 team,
                            Uint16 donttarget, Uint16 oldtarget);
Uint16 prt_find_target(float prtx, float prty, Uint16 facing,
                   Uint16 targetangle, bool_t onlyfriends, bool_t anyone,
                   Uint8 team, Uint16 donttarget, Uint16 oldtarget);
Uint16 chr_find_distant_target(Uint16 character, int maxdistance, bool_t enemies);

void chr_get_nearest_in_block(int  block_x, int  block_y, Uint16 character, bool_t items,
                          bool_t friends, bool_t enemies, bool_t dead, bool_t seeinvisible, IDSZ idsz);
Uint16 chr_get_nearest_target(Uint16 character, bool_t items,
                                  bool_t friends, bool_t enemies, bool_t dead, IDSZ idsz);
Uint16 chr_get_wide_target(Uint16 character, bool_t items,
                               bool_t friends, bool_t enemies, bool_t dead, IDSZ idsz, bool_t excludeid);
Uint16 chr_get_target_in_block(int  block_x, int  block_y, Uint16 character, bool_t items,
                                   bool_t friends, bool_t enemies, bool_t dead, bool_t seeinvisible, IDSZ idsz,
                                   bool_t excludeid);
Uint16 chr_get_nearby_target(Uint16 character, bool_t items,
                                 bool_t friends, bool_t enemies, bool_t dead, IDSZ idsz);


//REMOVE?
//Uint16 build_select_target(float tlx, float tly, float brx, float bry, Uint8 team);
//void do_cursor_rts();
//REMOVE END

bool_t load_one_title_image(int titleimage, char *szLoadName);
bool_t get_module_data(int modnumber, char *szLoadName);
bool_t get_module_summary(char *szLoadName);
void load_all_menu_images();
void load_blip_bitmap(char * modname);
void draw_trimx(int x, int y, int length);
void draw_trimy(int x, int y, int length);
void draw_trim_box(int left, int top, int right, int bottom);
void draw_trim_box_opening(int left, int top, int right, int bottom, float amount);
void draw_titleimage(int image, int x, int y);
void do_cursor();
void draw_module_tag(int module, int y);
int get_skin(char *filename);
bool_t check_skills(int who, Uint32 whichskill);
void check_player_import(char *dirname);
void fiddle_with_menu();
void reset_timers();
void reset_camera();
void sdlinit(int argc, char **argv);
int  glinit(int argc, char **argv);
void gltitle();


//---------------------------------------------------------------------------------------------
// Filesystem functions

typedef enum priority_e
{
  PRI_NONE = 0,
  PRI_WARN,
  PRI_FAIL
} PRIORITY;

int DirGetAttrib(char *fromdir);
void fs_init();
FILE * fs_fileOpen(PRIORITY pri, const char * src, const char * fname, const char * mode);
void fs_fileClose(FILE * pfile);
const char *fs_getTempDirectory();
const char *fs_getImportDirectory();
const char *fs_getGameDirectory();
const char *fs_getSaveDirectory();
int fs_fileIsDirectory(const char *filename);
int fs_createDirectory(const char *dirname);
int fs_removeDirectory(const char *dirname);
void fs_deleteFile(const char *filename);
void fs_copyFile(const char *source, const char *dest);
void fs_removeDirectoryAndContents(const char *dirname);
void fs_copyDirectory(const char *sourceDir, const char *destDir);

// Enumerate directory contents
const char *fs_findFirstFile(const char *path, const char *extension);
const char *fs_findNextFile(void);
void fs_findClose();



void render_mad_lit(Uint16 character);
void render_particle_reflections();
void render_water_fan_lit(Uint32 fan, Uint8 layer, Uint8 mode);

void BeginText(void);
void EndText(void);

void Begin2DMode(void);
void End2DMode(void);

#endif //#ifndef _PROTO_H_
